(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Extraction Language Ocaml.

Require Import ExtrOcamlBasic.
Require Import NPeano.
Require Import lang_common lang_tgt_interp zwickel pilsner awkward.

Set Extraction AccessOpaque.    


(** * Standard library stuff.
    Are parts of this already included in ExtrOcamlBasic?
*)

Extract Constant negb => "not".

Extract Inductive option => "option"
  [ "Some" "None" ].

Extract Inductive prod => "( * )"
  [ "( , )" ].
Extract Constant fst => "fst".
Extract Constant snd => "snd".




Extract Inductive list => "list"
  [ "[]" "( :: )" ].











(** Mapping Coq's [nat] to Caml's [big_int]. *)

Extract Inductive nat => "Big_int.big_int"
  [ "Big_int.zero_big_int" "Big_int.succ_big_int" ]
  "(fun zero succ n ->
      if Big_int.eq_big_int n Big_int.zero_big_int then zero ()
      else succ (Big_int.sub_big_int n Big_int.unit_big_int))".
Extract Constant Peano_dec.eq_nat_dec => "Big_int.eq_big_int".
Extract Constant EqNat.eq_nat_decide => "Big_int.eq_big_int".
Extract Constant EqNat.beq_nat => "Big_int.eq_big_int".
Extract Constant plus => "Big_int.add_big_int".
Extract Constant mult => "Big_int.mult_big_int".
Extract Constant max => "Big_int.max_big_int".
Extract Constant min => "Big_int.min_big_int".

Extract Constant land => "Big_int.and_big_int".
Extract Constant lor => "Big_int.or_big_int".
Extract Constant shiftr =>
  "fun x n -> Big_int.shift_right_big_int x (Big_int.int_of_big_int n)".
Extract Constant shiftl =>
  "fun x n -> Big_int.shift_left_big_int x (Big_int.int_of_big_int n)".
Extract Constant Nat.ones =>
  "let one = Big_int.big_int_of_int 1 in
   fun n -> Big_int.sub_big_int
              (Big_int.shift_left_big_int one (Big_int.int_of_big_int n))
              one".










Extract Constant pilsner.le_exp =>
  "let order = ExpHashtbl.create 42 in
   let counter = ref 0 in
   fun x y ->
     if not (ExpHashtbl.mem order x) then (
       ExpHashtbl.add order x !counter;
       counter := !counter + 1
     );
     if not (ExpHashtbl.mem order y) then (
       ExpHashtbl.add order y !counter;
       counter := !counter + 1
     );
     ExpHashtbl.find order x <= ExpHashtbl.find order y".















Separate Extraction zwickel pilsner lang_tgt.link lang_tgt_interp mt lang_src.Elet lang_common.MAIN.
