(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

open Big_int
open Printf

open Lang_tgt_interp
open Pils_tgt


let rec execute (cp: computer) n : unit =
  if cphp cp = None
  then printf "\nCrashed at %s (%s steps).\n"
           (string_of_big_int (cppc cp)) (string_of_int n)
  else match step cp with
  | Coq_efc_halted ->
    printf "\nTerminated successfully (%s steps).\n" (string_of_int n)
  | Coq_efc_tau cp' ->
    execute cp' (n+1)
  | Coq_efc_in f ->
    print_string "in: ";
    let m = big_int_of_string (read_line ()) in
    execute (f m) (n+1)
  | Coq_efc_out (m, cp') ->
    printf "out: %s\n" (string_of_big_int m);
    execute cp' (n+1)
;;


let _ =
  let m = load_module Sys.argv.(1) in
  print_module m;
  match cload m with
  | Some cp ->
    printf "Executing function %s (code starts at %s).\n\n"
        (string_of_big_int Lang_common.coq_MAIN)
        (string_of_big_int (cppc cp));
    execute cp 0
  | None -> failwith "Loading failed."
;;
