(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List.
Require Import vgtac basis myclassical.
Require Import lang_common lang_spec lang_src_lsi lang_tgt.
Require Import model_common model model_properties.
Require Import gw_common gw ssr util ts.
Require Import zwickel zwickel_correct_aux.

Set Implicit Arguments. 
Global Set Bullet Behavior "Strict Subproofs".



Lemma tr_exp_correct_unpack: forall F G ty ty' e x e' aa
    (IH: ts_psim F G (Texists ty)
                 (fun n => map enc (tr_exp e n (map fst F) (map fst G))) e)
    (LC: lc_typ 1 ty)
    (FRF: ~ ftn_typ_l (ty :: ty' :: map (@snd _ _) F) aa)
    (FRG: ~ ftn_typ_l (ty :: ty' :: map (@snd _ _) G) aa)
    (IH': ts_psim F ((x, tsubst_bvar 0 (Tname aa) ty)::G) ty'
                  (fun n => map enc (tr_exp e' n (map fst F) (x::map fst G))) e'),
  ts_psim F G ty' (fun n => map enc (tr_exp (Eunpack x e e') n (map fst F) (map fst G))) (Eunpack x e e').
Proof.
  i; rdes IH; rdes IH'; econs; r; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.

  destruct (LEM (pmvalid (snd s4))); last first.
  { ts_prepare.
    exfalso; apply f; split_pmvalid HV; auto.
  }

  instantiate (1 := wfo_add _ _); rewrite <-wfo_add_assoc.
  eapply IH with (kb:= (SKunpack _ _)::kb); clear IH; eauto; [|].
  { rewrite Eladdr map_app pmcells_app; econs; symmetry.
    rewrite -pmplus_assoc; rhs rewrite -(pmplus_swap {_ - _ |=> _}).
    pmplus_eq'.
  }
  fold subst_exp; i; simpl app_cont.
  i; ts_prepare.

  invc VR REL; last by (exfalso; invc VR).
  rrdes CA; subst.
  rrdes CB; invc SVQ.

  eapply pesim_stepb.
  i; rrdes XTR RLb; clarify.
  rrdes RL; s in Ec; noemp; clarify; esplits; [|right; esplits].
  { econs 1; inst_pairs; eapply Rok; eauto. }
  { inst_pairs; s; noemp; rr; esplits; eauto. }

  eapply pesim_dance
  ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
  ; clear RLa0 SV HV FIN FIN0.
  +
    dance_init.
    { eapply Sok; [s; i; omega|eapply Sopr].
      - readpc EXT0 Eladdr laddr.
        unfold pop; eauto.
      - eauto.
      - eauto.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [s; i; omega|eapply Sld].
      - readpc EXT0 Eladdr laddr.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [s; i; omega|eapply Sld].
      - readpc EXT0 Eladdr laddr.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [s; i; omega|eapply Salc]; [|by eauto].
      readpc EXT0 Eladdr laddr.
    }
    { depmcplus1; eauto using alloc_valid. }
    s; setnewloc; tmext_uniq.

    dance_init.
    { eapply Sok; [s; i; omega|eapply Sld].
      - readpc EXT0 Eladdr laddr.
      - econs; eapply pmread_plus; eauto.
        s; unfold mrf; tmext_simpl.
        (rhs rewrite LST -pmplus_assoc pmplus_swap); auto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc EXT0 Eladdr laddr.
      - rewrite pmcells_cons -pmplus_assoc.
        prep_store; eauto.
    }
    { tmext_simpl; pmcells_fold.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc EXT0 Eladdr laddr.
      - rewrite 2!pmcells_cons -2!pmplus_assoc (pmplus_swap {_|->_}).
        prep_store; eauto.
    }
    { s; tmext_simpl; pmcells_order; pmcells_fold.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    eapply dance_O.
    instantiate (1 := fun _ _ _ cf _ _ _  => depmach1 cf); s; eauto.

  + i; rrdes XTRb RLb; clarify.
    s in Ec; clarify.
    esplits; first 1 last.
    { right; split; eauto using wfo_le_n_add. }
    { inst_pairs; rr; esplits; s; noemp; eauto. }

    i; frames_dep 3 RLa XTRa cfas; frames_nodep RLa cfas.
    destruct ea'; clear FIN.
    destruct_tgt_real RLa.
    simpl lang_tgt.chp in *.
    setcrf mrf'.
    destruct s' as [[[[]]]]; clean.
    setnewloc.
    set (s' := (mrf', P1, s5, s6, (fst s4, {newloc |=> [vam; r1 r_env]} ## snd s7))).

    assert (VLD': pmvalid ({newloc |=> [vam; r1 r_env]} ## snd s7)).
    { decomp_pmvalid HV; auto. }

    assert (s5 = s).
    { rrdes EXT0 EXT0 EXT0; ss; vauto. }
    subst s5.

    assert (EXT': ext (wlift mylw) s' (r1, P1, s, s6, s7)).
    { s; splits; tmext_simpl; auto.
      i; eapply pmread_plus; eauto.
      pmplus_eq'.
    }

    hexploit (@pmplus_from_read _ _ _ (snd s4) (snd s7)).
    { i; rrdes EXT0; apply EXT1; s; auto. }
    { split_pmvalid VLD'; eauto. }
    intro HH; des.

    exists s'; esplits; [by eauto|].

    rewrite <-subexp_tmext_l; last by auto.
    rewrite tmext_lift tgt_csplit src_csplit.
    eapply (IH' ([tyi//aa]>>tsub)).
    * ii; destruct (LEM (x0 = aa)); subst.
      { rewrite tmext_eq; eauto. }
      { rewrite tmext_neq; eauto. }
    * eauto.
    * eauto using ext_trans.
    * i; s in IN; des; vauto.
      dup IN; apply RSV in IN; des; esplits; eauto.
      erewrite tsubname_agree.
      { eapply vclos_mon; eauto using ext_trans.
        i; eapply gk_mon; eauto using ext_trans.
      }
      { i; clear -FRF IN0 IN3.
        assert (x0 <> aa).
        { ii; subst; induction F; vauto.
          destruct a; apply list_assoc_cons_or in IN0; des; subst
          ; [apply FRF|eapply IHF]; ss; i; des; eauto using FRF.
        }
        rewrite tmext_neq; eauto.
      }
    * eauto.
    * insts $ ((x, vam)::denva).
      i; apply list_assoc_cons_or in IN; des; subst.
      { rewrite list_assoc_cons_eq; esplits; eauto.
        tmext_simpl.
        rewrite tsub_name_bvar.
        - simpl tsubst_name; tmext_simpl.
          erewrite tsubname_agree.
          + eapply vclos_mon; eauto using ext_trans.
            i; eapply gk_mon; eauto using ext_trans.
          + i; rewrite tmext_neq; eauto.
            ii; subst; apply FRF; s; eauto.
        - ii; destruct (LEM (x0 = aa)); subst.
          + rewrite tmext_eq; eauto.
          + rewrite tmext_neq; eauto.
      }
      rewrite tmext_neq; [|by vauto].
      rewrite list_assoc_cons_neq; [|by ii; subst x0; vauto].
      dup IN0; apply RDV in IN0; des.
      esplits; eauto.
      erewrite tsubname_agree.
      { eapply vclos_mon; eauto using ext_trans.
        i; eapply gk_mon; eauto using ext_trans.
      }
      { i; clear -FRG IN1 IN3.
        assert (x1 <> aa).
        { ii; subst; induction G; vauto.
          destruct a; apply list_assoc_cons_or in IN1; des; subst
          ; [apply FRG|eapply IHG]; ss; i; des; eauto using FRG.
        }
        rewrite tmext_neq; eauto.
      }
    * unfold s', mrf'; s; tmext_simpl; esplits
      ; try (by eapply pmread_plus; eauto; rewrite !pmcells_cons; pmplus_eq').
      rewrite -ENV0 in ENV; eapply reprlist_mon; eauto.
      i; eapply pmread_plus; eauto.
      rewrite HH; pmplus_eq'.
    * s; rewrite DORD; eauto.
    * i; simpl map in IN.
      apply in_consD in IN; des; subst.
      { rewrite tmext_eq in FH; eauto using ts_gw_vclos_closed. }
      rewrite tmext_neq in FH; eauto.
    * rewrite <-tmext_lift; eapply weakclo_tmext; eauto.
      i; exfalso; eauto using ts_gw_vclos_closed.
    * vauto.
    * unfold s'; simpl_proj; econs; symmetry.
      rewrite HH Eladdr; simpl tr_exp.
      rewrite !pmcells_cons map_app pmcells_app.
      rewrite <-!pmplus_assoc.
      match goal with |- context[length ?x] => set (len := length x) end.
      match goal with |- context[length ?x] => change (length x) with len end. 
      instantiate (2 := 7 + length (map enc (tr_exp e n (map fst F) (map fst G))) + n).
      match goal with |- context[length ?x] => change (length x) with len end. 
      replace (_ - (7 + _ + n)) with (laddr - n); last omega.
      unfold len; s; rhs rewrite <-!(pmplus_swap {laddr - n |=> _}).
      rewrite !pmcells_cons map_app pmcells_app plus_comm !map_length; pmplus_eq'.
    * eauto.
    * 
      clear SV HV FIN.
      destruct s'0 as [[[[]]]]; clean.
      i; ts_prepare.

      hexploit (@pmplus_from_read _ _ _ ({newloc |=> [vam; r1 r_env]} ## snd s7) (snd s9)).
      { i; rrdes EXT0; apply EXT1; s; eauto. }
      { eauto. }
      intro HH'; des.

      assert (EXT'': ext (wlift mylw) (mrf, P2, s5, s8, s9) (r0, P0, s, s3, s4))
        by (repeat (eapply ext_trans; eauto)).

      eapply pesim_dance
      ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
      ; clear RLa0 SV HV FIN.

      { dance_init.
        { eapply Sok; [by vauto|eapply Sld].
          - readpc EXT'' Eladdr laddr.
          - econs; s; unfold mrf; tmext_simpl.
            rewrite ENV1; s; unfold mrf'; tmext_simpl.
            eapply pmread_plus; eauto.
            rhs rewrite HH' !pmcells_cons -!pmplus_assoc -!(pmplus_swap {S newloc |-> _}).
            eauto.
        }
        { depmcplus0. }

        eapply dance_O.
        instantiate (1 := depmach0); s; eauto.
      }

      
      i; rrdes XTRb RLb0; clarify.
      s in Ec; clarify.
      esplits; first 1 last.
      { right; split; eauto using wfo_le_n_add. }
      { inst_pairs; rr; esplits; s; noemp; eauto. }
    
      i; frames_nodep RLa cfas.
      destruct ea'; clear FIN.
      destruct_tgt_real RLa.
      simpl lang_tgt.chp in *.
      setcrf mrf''.
      set (s'' := (mrf'', P2, env, s8, s9)).

      assert (VLD'': pmvalid (snd s9)).
      { split_pmvalid HV; auto. }

      assert (EXT''': ext (wlift mylw) s'' (mrf, P2, env, s8, s9)).
      { unfold s'', mrf''; clear -HV VLD''.
        s; splits; tmext_simpl; auto.
      }

      exists s''; esplits; [by eauto|].

      hexploit (KSIM g'0); clear KSIM.
      { eauto. }
      { eauto. }
      { instantiate (1:=s'').
        repeat (eapply ext_trans; eauto).
      }
      { s; unfold mrf''; tmext_simpl; auto. }
      { s; unfold mrf''; tmext_simpl; rewrite SP0.
        s; unfold mrf'; tmext_simpl; rewrite SP; omega.
      }
      { s; unfold mrf''; tmext_simpl; rewrite RET0.
        s; unfold mrf'; tmext_simpl; rewrite RET; auto.
      }
      { s; unfold mrf''; tmext_simpl; rewrite LST0; s; auto. }
      { erewrite tsubname_agree.
        - eapply vclos_mon; eauto using ext_trans.
          i; eapply gk_mon; eauto using ext_trans.
        - i; rewrite tmext_neq; eauto.
          ii; subst; apply FRF; s; eauto.
      }
      intro X; simpl tr_exp in X.
      exploit X; clear X.
      { unfold s''; rr; esplits; eauto; [|by rr; esplits; eauto].
        rr; esplits; eauto.
        - rr; splits; [|by rr; esplits; eauto]; s.
          instantiate (1 := {| cpc := Nothing; crf := Just mrf''; cst := freest0; chp := hp0 |}).
          rrdes CPa0; ss; subst; rr; esplits; s; unfold mrf''; tmext_simpl; eauto.
        - dup EXT'''; rrdes EXT'''; simpl_proj.
          eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
          s; i; eapply gk_mon; eauto; split; eauto.
      }
      simpl cplus; noemp; intro ES.
      eapply pesim_rewrite; eauto.
      f_equal.
      { rewrite !map_length !app_length; s; f_equal.
        rewrite !app_length; s; omega.
      }
      { rewrite -pmplus_assoc; auto. }
    * unfold s'; rr; esplits.
      { rr; esplits; eauto.
        - s; rr; esplits; clean; eauto.
          eapply tgt_confpred_pop with (popped := [vam]); eauto.
          + unfold mrf'; tmext_simpl; s; auto.
          + rewrite LST in SV; unfold mrf'; tmext_simpl.
            rewrite -pmcells_sng pmplus_comm; decomp_pmvalid SV; auto.
        - dup EXT'; rrdes EXT'; simpl_proj.
          eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
          + s; i; eapply gk_mon; eauto; split; eauto.
          + s; i; eapply gk_mon; eauto; split; eauto.
      }
      { rr; esplits; eauto. }
      { s; noemp; rewrite pmplus_swap; f_equal.
        - rewrite LST pmplus_swap -pmplus_assoc -pmcells_sng.
          unfold mrf'; tmext_simpl; auto.
        - rewrite -pmplus_assoc (pmplus_swap refha); auto.
      }
      { s; noemp; eauto. }
Grab Existential Variables.
  exact i.
Qed.



Lemma tr_exp_correct_unroll: forall F G e ty
    (IH: ts_psim F G (Tmu ty) (fun n => map enc (tr_exp e n (map fst F) (map fst G))) e),
  ts_psim F G (tsubst_bvar 0 (Tmu ty) ty)
          (fun n => map enc (tr_exp (Eunroll e) n (map fst F) (map fst G))) (Eunroll e).
Proof.
  i; rdes IH; econs; red; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.

  destruct (LEM (pmvalid (snd s4))); last first.
  { ts_prepare.
    exfalso; apply f; split_pmvalid HV; auto.
  }

  eapply IH with (kb:= SKunroll::kb); eauto.
  i; ts_prepare.

  invc VR REL; [|by inv VR]; rrdes CB; invc SVQ.
  eapply pesim_stepb.
  i; rrdes XTR RLb; clarify.
  rrdes RL; s in Ec; noemp; clarify; esplits; [|right; esplits].
  { econs 1; inst_pairs; eapply Rok; eauto. }
  { inst_pairs; s; noemp; rr; esplits; eauto. }

  hexploit (KSIM g').
  - eauto.
  - eauto.
  - instantiate (1:=s').
    eapply ext_trans; [eapply extpub_in_ext|]; eauto.
  - s; unfold mrf; tmext_simpl; auto.
  - s; unfold mrf; tmext_simpl; auto.
  - s; unfold mrf; tmext_simpl; auto.
  - rewrite LST; auto.
  - rrdes CA; subst.
    rewrite tsub_name_bvar; eauto.
  - intro X; simpl (match _ with | _ => _ end) in X.
    exploit X.
    + rr; esplits; eauto; [|by rr; esplits; eauto].
      rr; esplits; eauto.
      rr; splits; [|by rr; esplits; eauto]; eauto.
    + simpl cplus; noemp.
      rewrite -!pmplus_assoc.
      esplits; last first; eauto.
Qed.



Lemma tr_exp_correct_roll: forall F G e ty
    (IH: ts_psim F G (tsubst_bvar 0 (Tmu ty) ty) (fun n => map enc (tr_exp e n (map fst F) (map fst G))) e),
  ts_psim F G (Tmu ty)
          (fun n => map enc (tr_exp (Eroll e) n (map fst F) (map fst G))) (Eroll e).
Proof.
  i; rdes IH; econs; red; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.

  destruct (LEM (pmvalid (snd s4))); last first.
  { ts_prepare.
    exfalso; apply f; split_pmvalid HV; auto.
  }

  eapply IH with (kb:= SKroll::kb); eauto.
  i; ts_prepare.

  hexploit (KSIM g').
  - eauto.
  - eauto.
  - instantiate (1:=s').
    eapply ext_trans; [eapply extpub_in_ext|]; eauto.
  - s; unfold mrf; tmext_simpl; auto.
  - s; unfold mrf; tmext_simpl; auto.
  - s; unfold mrf; tmext_simpl; auto.
  - rewrite LST; auto.
  - rewrite tsub_name_bvar in VR; eauto.
    insts $ (Vroll vb).
    econs; econs.
    + s; eauto.
    + s; rr; splits; [by econs|]; s; eauto using ts_gw_vclos_closed.
    + fold tsubst_name; rewrite plus_0_r; eauto.
  - intro X; simpl (match _ with | _ => _ end) in X.
    exploit X.
    + rr; esplits; eauto; [|by rr; esplits; eauto].
      rr; esplits; eauto.
      rr; splits; [|by rr; esplits; eauto]; eauto.
    + simpl cplus; noemp.
      rewrite -!pmplus_assoc.
      esplits; last first; eauto.
Qed.



Lemma tr_exp_correct_pack: forall F G e ty ty'
    (IH: ts_psim F G (tsubst_bvar 0 ty' ty) (fun n => map enc (tr_exp e n (map fst F) (map fst G))) e)
    (LC: lc_typ 0 ty'),
  ts_psim F G (Texists ty)
          (fun n => map enc (tr_exp (Epack e) n (map fst F) (map fst G))) (Epack e).
Proof.
  i; rdes IH; econs; red; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.

  destruct (LEM (pmvalid (snd s4))); last first.
  { ts_prepare.
    exfalso; apply f; split_pmvalid HV; auto.
  }

  eapply IH with (kb:= SKpack::kb); eauto.
  i; ts_prepare.

  hexploit (KSIM g').
  - eauto.
  - eauto.
  - instantiate (1:=s').
    eapply ext_trans; [eapply extpub_in_ext|]; eauto.
  - s; unfold mrf; tmext_simpl; auto.
  - s; unfold mrf; tmext_simpl; auto.
  - s; unfold mrf; tmext_simpl; auto.
  - rewrite LST; auto.
  - insts $ (Vpack vb).
    econs; econs; first 1 last.
    + s; eauto.
    + s; rr; splits; [by econs|]; s; eauto using ts_gw_vclos_closed.
    + fold tsubst_name; rewrite plus_0_r.
      rewrite tsub_name_bvar in VR; eauto.
    + eapply lc_tsubname; eauto.
  - intro X; simpl (match _ with | _ => _ end) in X.
    exploit X.
    + rr; esplits; eauto; [|by rr; esplits; eauto].
      rr; esplits; eauto.
      rr; splits; [|by rr; esplits; eauto]; eauto.
    + simpl cplus; noemp.
      rewrite -!pmplus_assoc.
      esplits; last first; eauto.
Qed.
