(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List.
Require Import vgtac basis myclassical.
Require Import lang_common lang_spec lang_src_lsi lang_tgt.
Require Import model_common model model_properties.
Require Import gw_common gw ssr util ts.
Require Import zwickel zwickel_correct_aux.

Unset Implicit Arguments. 
Global Set Bullet Behavior "Strict Subproofs".



Lemma tr_exp_correct_inst: forall F G ty ty' e
    (IH: ts_psim F G (Tforall ty) (fun n => map enc (tr_exp e n (map fst F) (map fst G))) e)
    (LC: lc_typ 0 ty'),
  ts_psim F G (tsubst_bvar 0 ty' ty) (fun n => map enc (tr_exp (Einst e) n (map fst F) (map fst G))) (Einst e).
Proof.
  i; rdes IH; econs; r; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.

  destruct (LEM (pmvalid (snd s4))); last first.
  { ts_prepare.
    exfalso; apply f; split_pmvalid HV; auto.
  }

  instantiate (1 := wfo_add _ _); rewrite <-wfo_add_assoc.
  eapply IH with (kb:= SKinst::kb); clear IH; eauto; [|].
  { rewrite Eladdr map_app pmcells_app; econs; symmetry.
    rewrite -pmplus_assoc; rhs rewrite -(pmplus_swap {_ - _ |=> _}).
    pmplus_eq'.
  }
  fold subst_exp; i; simpl app_cont.
  i; ts_prepare.

  
  invc VR REL; inv VR.
  dup CPa; rrdes CPa CA; clean.
  dup CA; apply HP in CA.
  s in CA; clear HP ST.

  eapply pesim_dance
  ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
  ; clear RLa0 SV HV FIN.
  - dance_init.
    { eapply Sok; [s; i; omega|eapply Sld].
      - readpc EXT0 Eladdr laddr.
      - econs; s; unfold mrf; tmext_simpl.
        eapply pmread_plus; eauto.
        symmetry; rewrite pmplus_swap LST SP.
        rewrite -!pmplus_assoc.
        repeat match goal with |- context[?a + 1 - 1] =>
          replace (a + 1 - 1) with a by omega
        end; eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc EXT0 Eladdr laddr.
      - unfold stack_top; rewrite pmplus_swap LST -!pmplus_assoc.
        prep_store; eauto.
    }
    { s; noemp; depmcplus0.
      rewrite LST -!pmplus_assoc pmplus_swap (pmplus_swap freest) in SV0.
      eauto using pmvalid_plus_cell.
    }

    dance_init.
    { eapply Sok; [by vauto|eapply Slpc].
      readpc EXT0 Eladdr laddr.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Sopr].
      - readpc EXT0 Eladdr laddr.
      - eauto.
      - eauto.
      - eauto.
    }
    { depmcplus0. }
    tmext_uniq.

    
    eapply dance_O.
    instantiate (1 := fun _ _ _ => depmach0); s; eauto.

  - destruct s' as [[[[]]]]; clean.
    i; rrdes XTRb RLb0; clarify; s in Ec; clarify.
    esplits; first 2 last; last first.
    + right; splits; eauto using wfo_le_n_add.
    + i; frames_nodep RLa cfas.
      destruct ea'; clear FIN; destruct_tgt_real RLa.
      clear RLa0.
      simpl lang_tgt.chp in *.
      setcrf mrf.
      set (s7' := ({r1 r_sp - 1 |-> r1 r_ret} ## fst s4, snd s7)).
      set (s' := (mrf, P1, env, s6, s7')).
      assert (EXT': ext (wlift mylw) s' (r1, P1, env, s6, s7)).
      { unfold s'; repeat (split; auto); unfold mrf; s; tmext_simpl; auto.
        - split_pmvalid HV; auto.
        - rrdes IN; eauto.
      }

      
      insts $ (refhb, Just env, Just (app_cont kb (Einst vb))).
      esplits; eauto.
      pfold; econs; i; esplits; eauto.
      destruct cfa; rrdes XTRa; s in XTRa; s in XTRa0; clarify.
      destruct cfb; rrdes XTRb; clarify.
      invc CNF; des; subst; noemp.
      rrdes MB; s in Ec; clarify. subst env0 e0 h2.
      destruct ma; clear SV HV FIN; simpl cplus in MA.
      destruct_tgt_real MA.
      rewrite <-!pmplus_assoc in *; clarify.

      rrdes CB; invc SVQ.

      assert (EXT'': ext (wlift mylw) s' (r0, P0, s, s3, s4)). 
      { eapply ext_trans; eauto. }

      eapply esim_step; first (i; rrdes FH; ss; omega).
      i; s in ST; eapply tgt_det in ST; first 1 last; [|by eauto|by eauto| |].
      { eapply Sok; [i; ss; omega|eapply Sjmp].
        - readpc EXT'' Eladdr laddr.
        - econs; tmext_simpl; s; eapply pmread_plus; eauto.
      }
      { ii; vauto. }
      des; subst.

      setmrf mrf'. fold mrf' in s'.
      esplits; first (by eauto); last (by eauto); [right; econs; eauto|]
      ; last first; last 1 first.
      * inst_pairs; apply Rok; apply red_ctx; auto.
      * econs; first 1 last; last first.
        { rr; esplits; last by eauto.
          - rr; esplits; [ | |by eauto|by eauto]; last first.
            + eapply ts_refconfrel_mon
              ; [eapply ts_refconfrel'_smon; s; eauto|]; inst
              ; s; i; eapply gk_mon; eauto; rrdes EXT'; rr; eauto.
            + unfold s'; rr; splits; last by eauto.
              s; instantiate (1 := {| cpc := Nothing; crf := Just mrf'; cst := freest; chp := hp |}).
              rrdes CPa0; ss; subst; rr; esplits; s; unfold mrf'; tmext_simpl; eauto.
          - rr; esplits; eauto.
          - s; noemp; eauto.
        }
        { instantiate (1:=(pmemp, Nothing, _)); s; noemp.
          rr; esplits; eauto.
        }
        { instantiate (1 := mkconf (Just _) Nothing pmemp pmemp); s; noemp.
          rewrite <-!pmplus_assoc. tmext_simpl.
          rr; esplits; s; noemp; eauto.
          rewrite 2!(pmplus_swap freest); auto.
        }
        { split; rr; s; eauto. }
      * eapply call_inst; last first; first 1 last.
        { instantiate (1:=tsubst_name tsub ty').
          eapply lc_tsubname; eauto.
        }
        { eapply vclf_gen.
          - econs; apply EXT'; eauto.
          - rr; esplits; [|by eauto].
            rewrite -H; eauto.
          - s in H0; clarify; eauto using (gk_mon GK').
        }
        { s; econs; eauto. }
        { s; replace va with (mrf' r_clo); last (by unfold mrf'; tmext_simpl; auto).
          econs; s; unfold mrf'; tmext_simpl; eauto.
        }
        
        
        econs; i; rrdes EXTP EXTP EXTP EXTP; rrdes EXTP0.
        rewrite clear_bote; eapply gres_incl_id, pesim_gknow; i.
        invc RETa RETb.
        rename ST into ST', RF into RF'.
        clear SV HV FIN FIN0 HP.
        destruct s'0 as [[[[]]]]; clean.
        apply pesim_acquire.
        ts_prepare.
        eapply pesim_dance
        ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
        ; clear RLa0 SV HV FIN.

        { assert (EXT2: @ext (wlift mylw) (mrf, P2, s5, s8, s9) (r0, P0, s, s3, s4))
            by (eapply ext_trans; eauto).

          dance_init.
          { eapply Sok; [s; i; omega|eapply Sld].
            - readpc EXT2 Eladdr laddr.
            - econs; s; unfold mrf; tmext_simpl.
              eapply pmread_plus; eauto.
              rewrite ST'; s; rewrite -RF'; s; unfold mrf'; tmext_simpl.
              (rhs rewrite -!pmplus_assoc pmplus_swap); eauto.
          }
          { depmcplus0. }

          dance_init.
          { eapply Sok; [by vauto|eapply Ssto].
            - readpc EXT2 Eladdr laddr.
            - unfold stack_top.
              rewrite ST'; s.
              rewrite -pmplus_assoc pmplus_swap.
              match goal with |- context[{?x - 1 |-> r1 r_ret}] =>
                replace (x - 1) with (mrf r_sp - 1) by
                  (unfold mrf; s; tmext_simpl; rewrite -RF'; s; unfold mrf'; tmext_simpl; auto)
              end.
              prep_store; eauto.
          }
          { s; tmext_simpl; depmcplus0.
            rewrite ST' -!pmplus_assoc pmplus_swap in SV0.
            rewrite -RF'; s; unfold mrf'; tmext_simpl.
            eauto using pmvalid_plus_cell.
          }

          eapply dance_O.
          instantiate (1 := fun _ => depmach0); s; eauto.
        }

        
        clear RLb0.
        i; rrdes XTRb RLb0; clarify.
        s in Ec; clarify.
        esplits; first 1 last.
        { right; split; eauto using wfo_le_n_add. }
        { inst_pairs; rr; esplits; s; noemp; eauto. }
    
        i; frames_nodep RLa cfas.
        destruct ea'; clear FIN.
        destruct_tgt_real RLa.
        simpl lang_tgt.chp in *.
        setcrf mrf''.
        set (s'' := (mrf'', P2, env0, s8, ({mrf r_sp - 1 |-> mrf r_arg} ## fst s4, snd s9))).

        assert (VLD'': pmvalid (snd s9)).
        { split_pmvalid HV; auto. }

        assert (EXT''': ext (wlift mylw) s'' (mrf, P2, env0, s8, s9)).
        { unfold s'', mrf''; clear -HV VLD''.
          s; splits; tmext_simpl; auto.
        }

        exists s''; esplits; [by eauto|].

        hexploit (KSIM g'0); clear KSIM.
        { eauto. }
        { eauto. }
        { instantiate (1:=s'').
          repeat (eapply ext_trans; eauto).
        }
        { s; unfold mrf''; tmext_simpl; rewrite -RF0; s; unfold mrf'; tmext_simpl; rewrite ENV0; auto. }
        { s; unfold mrf''; tmext_simpl; rewrite -RF'; s; unfold mrf'; tmext_simpl; rewrite SP; omega. }
        { s; unfold mrf''; tmext_simpl; rewrite RET; auto. }
        { s; unfold mrf''; tmext_simpl; auto. }
        { rewrite tsub_name_bvar; eauto.
          eapply vclos_mon; eauto.
          i; eapply gk_mon; eauto; repeat eapply ext_trans; eauto.
        }
        intro X; simpl tr_exp in X.
        exploit X; clear X.
        { unfold s''; rr; esplits; eauto; [|by rr; esplits; eauto].
          rr; esplits; eauto.
          - rr; splits; [|by rr; esplits; eauto]; s.
            instantiate (1 := {| cpc := Nothing; crf := Just mrf''; cst := freest0; chp := hp0 |}).
            rrdes CPa; ss; subst; rr; esplits; s; unfold mrf''; tmext_simpl; eauto.
          - dup EXT'''; rrdes EXT'''; simpl_proj.
            eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
            + s; i; eapply gk_mon; eauto; split; eauto.
            + s; i; eapply gk_mon; eauto; split; eauto.
        }

        simpl cplus; noemp; intro ES.
        eapply pesim_rewrite; eauto.
        f_equal.
        { rewrite !map_length !app_length; s; f_equal; omega. }
        { rewrite pmplus_swap; auto. }
        { rewrite -pmplus_assoc; auto. }
    + s; noemp; rr; esplits; eauto.
Qed.



Lemma tr_exp_correct_app: forall F G ty ty' e e'
    (IH: ts_psim F G (Tarrow ty' ty) (fun n => map enc (tr_exp e n (map fst F) (map fst G))) e)
    (IH': ts_psim F G ty' (fun n => map enc (tr_exp e' n (map fst F) (map fst G))) e'),
  ts_psim F G ty (fun n => map enc (tr_exp (Eapp e e') n (map fst F) (map fst G))) (Eapp e e').
Proof.
  i; rdes IH; rdes IH'; econs; r; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.

  destruct (LEM (pmvalid (snd s4))); last first.
  { ts_prepare.
    exfalso; apply f; split_pmvalid HV; auto.
  }

  instantiate (1 := wfo_add _ _); rewrite <-wfo_add_assoc.
  eapply IH with (kb:= SKappL _ ::kb); clear IH; eauto; [|].
  { rewrite Eladdr map_app pmcells_app; econs; symmetry.
    rewrite -pmplus_assoc; rhs rewrite -(pmplus_swap {_ - _ |=> _}).
    pmplus_eq'.
  }
  fold subst_exp; i; simpl app_cont.

  destruct (LEM (pmvalid (snd (snd s')))); last first.
  { ts_prepare.
    exfalso; apply f; split_pmvalid HV; auto.
  }

  instantiate (1 := wfo_add _ (wfo_nat 6)); rewrite <-wfo_add_assoc.
  eapply IH' with (kb:= SKappR _ ::kb); clear IH'; eauto using ext_trans.
  { i; apply RSV in IN; des; esplits; eauto.
    - rrdes EXT0 EXT0 EXT0 EXT0; rewrite EXT3; eauto.
    - eapply vclos_mon; eauto using (gk_mon GK).
  }
  { i; apply RDV in IN; des; esplits; eauto.
    eapply vclos_mon; eauto using (gk_mon GK).
  }
  { rewrite ENV0.
    rrdes EXT0 EXT0 EXT0 EXT0.
    rrdes EXT1; eapply reprlist_mon; eauto.
  }
  { omega. }
  { rrdes EXT0 EXT0 EXT0 EXT0; rrdes EXT1.
    hexploit (pmplus_from_read EXT1); eauto.
    instantiate (1 := length (map enc (tr_exp e n (map fst F) (map fst G))) + n).
    i; des; rewrite H Eladdr !map_app !pmcells_app plus_comm.
    econs; symmetry.
    replace (_ - _) with (laddr - n); last first.
    { clear; set (x := length _); omega. }
    rewrite <-!pmplus_assoc.
    rhs rewrite pmplus_swap. rewrite !map_length.
    rewrite ->pmplus_swap; pmplus_eq'.
  }
  
  i; ts_prepare.

  
  invc VR REL; inv VR.
  dup CPa; rrdes CPa CA; clean.
  dup CA; apply EXT1, HP in CA.
  s in CA; clear HP ST.

  eapply pesim_dance
  ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
  ; clear RLa0 SV HV FIN.
  - assert (EXT2: @ext (wlift mylw) s'0 (r0, P0, s, s3, s4))
      by (eapply ext_trans; eauto).

    dance_init.
    { eapply Sok; [s; i; omega|eapply Sopr].
      - readpc EXT2 Eladdr laddr.
        unfold pop; auto.
      - eauto.
      - eauto.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [s; i; omega|eapply Sld].
      - readpc EXT2 Eladdr laddr.
      - econs; s; unfold mrf; tmext_simpl.
        eapply pmread_plus; eauto.
        symmetry; rewrite pmplus_swap LST0 LST SP0 SP.
        rewrite -!pmplus_assoc pmplus_swap.
        repeat match goal with |- context[?a + 1 - 1] =>
          replace (a + 1 - 1) with a by omega
        end; eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc EXT2 Eladdr laddr.
      - unfold stack_top; rewrite pmplus_swap LST0 LST -!pmplus_assoc pmplus_swap.
        match goal with |- context[{?x - 1 |-> va}] =>
          
          replace x with (fst (fst (fst (fst s'0))) r_sp - 1) by (apply eq_minus; auto) 
        end.
        prep_store; eauto.
    }
    { s; noemp; pmcells_fold; depmcplus0.
      rewrite LST0 LST -!pmplus_assoc pmplus_swap in SV0.
      rewrite pmplus_swap.
      eapply pmvalid_plus_cell.
      rewrite pmplus_swap.
      rewrite SP0 in SV0 *.
      rewrite (pmplus_swap freest) pmplus_swap (pmplus_swap freest)in SV0.
      repeat match goal with
        | |- context[?x + 1 - 1] => replace (x + 1 - 1) with x by omega
        | H: context[?x + 1 - 1] |- _ => replace (x + 1 - 1) with x in H by omega
      end.
      eauto using pmvalid_plus_cell.
    }

    dance_init.
    { eapply Sok; [s; i; omega|eapply Sld].
      - readpc EXT2 Eladdr laddr.
      - econs; s; unfold mrf; tmext_simpl.
        eapply pmread_plus; eauto.
        (rhs rewrite pmplus_swap); eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Slpc].
      readpc EXT2 Eladdr laddr.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Sopr].
      - readpc EXT2 Eladdr laddr.
      - eauto.
      - eauto.
      - eauto.
    }
    { depmcplus0. }
    tmext_uniq.

    
    eapply dance_O.
    instantiate (1 := fun _ _ _ _ _ => depmach0); s; eauto.

  - destruct s' as [[[[]]]]; clean.
    i; rrdes XTRb RLb0; clarify; s in Ec; clarify.
    destruct s'0 as [[[[]]]].
    esplits; first 2 last; last first.
    + right; splits; eauto using wfo_le_n_add.
    + i; frames_nodep RLa cfas.
      destruct ea'; clear FIN; destruct_tgt_real RLa.
      clear RLa0.
      simpl lang_tgt.chp in *.
      setnewloc.
      setcrf mrf.
      set (s10' := ({r2 r_sp - 1 - 1|-> r2 r_ret} ## fst s4, snd s10)).
      set (s' := (mrf, P2, s8, s9, s10')).
      assert (EXT': ext (wlift mylw) s' (r2, P2, s8, s9, s10)).
      { unfold s'; repeat (split; auto); unfold mrf; s; tmext_simpl; auto.
        - split_pmvalid HV; auto.
        - rrdes IN; eauto.
      }

      
      insts $ (refhb, Just s8, Just (app_cont kb (Eapp vb vb0))).
      esplits; eauto.
      pfold; econs; i; esplits; eauto.
      destruct cfa; rrdes XTRa; s in XTRa; s in XTRa0; clarify.
      destruct cfb; rrdes XTRb; clarify.
      invc CNF; des; subst; noemp.
      rrdes MB; s in Ec; clarify. subst env e0 h2.
      destruct ma; clear SV HV FIN; simpl cplus in MA.
      destruct_tgt_real MA.
      rewrite <-!pmplus_assoc in *; clarify.

      rrdes CB; invc SVQ.

      assert (EXT'': ext (wlift mylw) s' (r0, P0, s, s3, s4)). 
      { eapply ext_trans; [eapply ext_trans|]; eauto. }

      eapply esim_step; first (i; rrdes FH; ss; omega).
      i; s in ST; eapply tgt_det in ST; first 1 last; [|by eauto|by eauto| |].
      { eapply Sok; [i; ss; omega|eapply Sjmp].
        - readpc EXT'' Eladdr laddr.
        - econs; tmext_simpl; s; eapply pmread_plus; eauto.
      }
      { ii; vauto. }
      des; subst.

      setmrf mrf'.
      esplits; first (by eauto); last (by eauto); [right; econs; eauto|]
      ; last first; last 1 first.
      * inst_pairs; apply Rok; apply red_ctx; auto.
      * econs; first 1 last; last first.
        { rr; esplits; last by eauto.
          - rr; esplits; [ | |by eauto|by eauto]; last first.
            + eapply ts_refconfrel_mon
              ; [eapply ts_refconfrel'_smon; s; eauto|]; inst
              ; s; i; eapply gk_mon; eauto; rrdes EXT'; rr; eauto.
            + rr; splits; last by eauto.
              eapply tgt_confpred_pop with (popped := [va0]); eauto.
              { tmext_simpl; s; auto. }
              { tmext_simpl; rewrite -pmcells_sng; decomp_pmvalid SV; auto. }
          - rr; esplits; eauto.
          - s; noemp; eauto.
        }
        { instantiate (1:=(pmemp, Nothing, _)); s; noemp.
          rr; esplits; eauto.
        }
        { instantiate (1 := mkconf (Just _) Nothing pmemp pmemp); s; noemp.
          rewrite <-!pmplus_assoc. tmext_simpl.
          rr; esplits; s; noemp; eauto.
          rewrite -pmcells_sng 2!(pmplus_swap freest) pmplus_swap; auto.
        }
        { split; rr; s; eauto. }
      * eapply call_app; last first; first 1 last.
        { move VR0 at bottom; eapply vclos_mon; eauto using (gk_mon GK'0). }
        { eapply vclf_fix.
          - econs; apply EXT', EXT1; eauto.
          - rr; esplits; [|by eauto].
            rewrite -H; eauto.
          - s in H0; clarify; eauto using (gk_mon GK'0).
        }
        { rewrite H; rr; econs; eauto. }
        { s; replace va with (mrf' r_clo); last (by unfold mrf'; tmext_simpl; auto).
          replace va0 with (mrf' r_arg); last (by unfold mrf'; tmext_simpl; auto).
          econs; s; unfold mrf'; tmext_simpl; eauto.
          apply EXT', EXT1; eauto.
        }
        
        
        econs; i; rrdes EXTP EXTP EXTP EXTP; rrdes EXTP0.
        rewrite clear_bote; eapply gres_incl_id, pesim_gknow; i.
        invc RETa RETb.
        rename ST into ST', RF into RF'.
        clear SV HV FIN FIN0 HP.
        destruct s'0 as [[[[]]]]; clean.
        apply pesim_acquire.
        ts_prepare.
        eapply pesim_dance
        ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
        ; clear RLa0 SV HV FIN.

        { assert (EXT3: @ext (wlift mylw) (mrf, P3, s11, s12, s13) (r0, P0, s, s3, s4))
            by (eapply ext_trans; eauto).

          dance_init.
          { eapply Sok; [s; i; omega|eapply Sld].
            - readpc EXT3 Eladdr laddr.
            - econs; s; unfold mrf; tmext_simpl.
              eapply pmread_plus; eauto.
              rewrite ST'; s; rewrite -RF'; s; tmext_simpl.
              (rhs rewrite -!pmplus_assoc pmplus_swap); eauto.
          }
          { depmcplus0. }

          dance_init.
          { eapply Sok; [by vauto|eapply Ssto].
            - readpc EXT3 Eladdr laddr.
            - unfold stack_top.
              rewrite ST'; s.
              rewrite -pmplus_assoc pmplus_swap.
              match goal with |- context[{?x - 1 - 1 |-> r2 r_ret}] =>
                replace (x - 1 - 1) with (mrf r_sp - 1)
                  by (unfold mrf; s; tmext_simpl; rewrite -RF'; s; tmext_simpl; auto)
              end.
              prep_store; eauto.
          }
          { s; tmext_simpl; depmcplus0.
            rewrite ST' -!pmplus_assoc pmplus_swap in SV0.
            rewrite -RF'; s; tmext_simpl.
            eauto using pmvalid_plus_cell.
          }

          eapply dance_O.
          instantiate (1 := fun _ => depmach0); s; eauto.
        }

        
        clear RLb0.
        i; rrdes XTRb RLb0; clarify.
        s in Ec; clarify.
        esplits; first 1 last.
        { right; split; eauto using wfo_le_n_add. }
        { inst_pairs; rr; esplits; s; noemp; eauto. }
    
        i; frames_nodep RLa cfas.
        destruct ea'; clear FIN.
        destruct_tgt_real RLa.
        simpl lang_tgt.chp in *.
        setcrf mrf''.
        set (s'' := (mrf'', P3, env, s12, ({mrf r_sp - 1 |-> mrf r_arg} ## fst s4, snd s13))).

        assert (VLD''': pmvalid (snd s13)).
        { split_pmvalid HV; auto. }

        assert (EXT''': ext (wlift mylw) s'' (mrf, P3, env, s12, s13)).
        { unfold s'', mrf''; clear -HV VLD'''.
          s; splits; tmext_simpl; auto.
        }

        exists s''; esplits; [by eauto|].

        hexploit (KSIM g'1); clear KSIM.
        { eauto. }
        { eauto. }
        { instantiate (1:=s'').
          repeat (eapply ext_trans; eauto).
        }
        { s; unfold mrf''; tmext_simpl; rewrite -RF0; s; tmext_simpl; rewrite ENV1 ENV0; auto. }
        { s; unfold mrf''; tmext_simpl; rewrite -RF'; s; tmext_simpl; rewrite SP0 SP; omega. }
        { s; unfold mrf''; tmext_simpl; rewrite RET0 RET; auto. }
        { s; unfold mrf''; tmext_simpl; auto. }
        { eapply vclos_mon; eauto.
          i; eapply gk_mon; eauto; repeat eapply ext_trans; eauto.
        }
        intro X; simpl tr_exp in X.
        exploit X; clear X.
        { unfold s''; rr; esplits; eauto; [|by rr; esplits; eauto].
          rr; esplits; eauto.
          - rr; splits; [|by rr; esplits; eauto]; s.
            instantiate (1 := {| cpc := Nothing; crf := Just mrf''; cst := freest0; chp := hp0 |}).
            rrdes CPa; ss; subst; rr; esplits; s; unfold mrf''; tmext_simpl; eauto.
          - dup EXT'''; rrdes EXT'''; simpl_proj.
            eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
            + s; i; eapply gk_mon; eauto; split; eauto.
            + s; i; eapply gk_mon; eauto; split; eauto.
        }

        simpl cplus; noemp; intro ES.
        eapply pesim_rewrite; eauto.
        f_equal.
        { rewrite !map_length !app_length; s; f_equal; omega. }
        { rewrite pmplus_swap; auto. }
        { rewrite -pmplus_assoc; auto. }
    + s; noemp; rr; esplits; eauto.
Qed.



(** This lemma statement is basically a copy&paste of the proof
   obligation arising at the heart of tr_exp_correct_fix.
*)
Lemma tr_exp_correct_fix_val: forall
    F G ty ty' f x e i tsub g P0 s3 s4 senva denva n h kb mpc mrf mb
    refha refhb hp env chp newloc mrf'
    (IH: ts_psim_aux i F ((x, ty) :: (f, Tarrow ty ty') :: G) ty'
                     (fun n : word => map enc (tr_exp e n (map fst F) (x::f::map fst G))) e)
    (HT: hastypepre F ((x, ty) :: (f, Tarrow ty ty') :: G) e ty')
    (LCT: lc_tsub tsub)
    (GK: gknow (wlift mylw) g)
    (SORD : map fst F = map fst senva)
    (denvb: termvar -> lang_spec.val src_lsi)
    (DORD: map fst G = map fst denva)
    (CL: forall x, In x (map fst G) -> forall y : termvar, fv_exp (denvb x) y -> False)
    (WCL: subst_weakly_closed (fun x : termvar => denvb x))
    (ENV: reprlist (snd s4) (mrf r_env) (map snd denva))
    (Eladdr: snd s4 =
             {mpc |=> map enc (tr_exp (Efix f x e) n (map fst F) (map fst G))}
               ## {mpc - n |=> map snd senva} ## h)
    (RSV: forall (l : label) (ty' : typ),
            lookup l F = Some ty' ->
            exists (va : lang_spec.val tgt_lsi) (vb : val),
              lookup l senva = Some va /\
              lookup l env = Some vb /\
              vclos (wlift mylw) (mrf, P0, env, s3, s4)
                    (g (mrf, P0, env, s3, s4)) (tsubst_name tsub ty') va vb)
    (RDV: forall (x : termvar) (ty' : typ),
            lookup x G = Some ty' ->
            exists va : lang_spec.val tgt_lsi,
              lookup x denva = Some va /\
              vclos (wlift mylw) (mrf, P0, env, s3, s4)
                    (g (mrf, P0, env, s3, s4)) (tsubst_name tsub ty') va
                    (denvb x))
    (CPb: src_confpred env (pmemp, Just env, Nothing))
    (RLb: real src_lsi
               (refhb, Just env,
                Just
                  (app_cont kb
                            (Efix f x
                                  (subst_exp
                                     [Evar x // x]>>
                                     [Evar f // f]>> (fun x : termvar => denvb x) e)))) mb)
    (Enl: newloc = alloc (hp ## refha ## snd s4 ## chp) 2)
    (Emrf': mrf' = [S (S (mpc + 8 + (length (tr_exp e
                  (S (S (S (S (S (S (S (S (S (S (S (S (S (S (S (S (S
                     (S (S (S (S (S n))))))))))))))))))))))
                  (map fst F) (x :: f :: map fst G)) + 16))) // r_arg]>>
     [mrf r_sp + 1 // r_sp]>> [newloc // r_clo]>> mrf)
    (s': state (wlift mylw))
    (Es': s' = (mrf', extend P0 (ts_fun (S (S (mpc + 8)))) newloc, env, s3,
                ({mrf r_sp |-> newloc} ## fst s4, {S newloc |-> mrf r_env} ## snd s4)))
    (VLD': pmvalid ({S newloc |-> mrf r_env} ## snd s4))
    (EXT': ext (wlift mylw) s' (mrf, P0, env, s3, s4))
    (CL': forall a : termvar,
            fv_exp (Vfix f x (subst_exp
                       [Evar x // x]>> [Evar f // f]>> (fun x : termvar => denvb x)
                       e)) a -> False),
  vclos (wlift mylw) s' (g s') (tsubst_name tsub (Tarrow ty ty')) newloc
        (Vfix f x (subst_exp [Evar x//x]>>[Evar f//f]>>(fun x0 : termvar => denvb x0) e)).
Proof.
  econs; apply vclgen_thunk, vclf_fix.
  { rewrite Es'; econs; rr; eauto. }
  { rr; splits; s; eauto using CL'. }

  eapply gk_lsim; eauto.
  evar (i0: wfo_set (wlift mylw)).
  econs; [econs|].
  { rewrite Es'; econs; rr; eauto. }
  { rr; splits; s; eauto using CL'. }

  insts $ i0.
  pcofix CIH; unfold i0 in *; clear i0.
  i; invc STK; [|by invc INSTb RF CB VR].
  eapply pesim_gknow; i; eapply pesim_gknow_; i.
  invc APPa APPb; invc RF CB VR; rrdes CA; s in VAL; clarify.
  destruct s'0 as [[[[]]]]; clean.
  ren PC .
  destruct s1.
  apply pesim_acquire.
  ts_prepare.
  setnewloc.

  assert (Epc: mpc0 = 10 + mpc).
  { rrdes EXT EXT EXT EXT.
    hexploit tgt_vquery_fun_uniq; [by eauto|by eauto| |].
    - eapply EXT; s; rr; eauto.
    - rewrite plus_comm; eauto.
  }
  subst mpc0.

  clear CA CPb.

  
  hexploit (tgt_stack CPa 1); i; des; s in FST; subst freest.
  des; repeat (destruct vs; [by vauto|]); destruct vs; [|by vauto].
  rewrite <-!pmplus_assoc, <-pmcells_sng in *; clear NST.

  hexploit (@pmplus_from_read _ _ _ (snd s4) h0).
  { i; rrdes EXT; apply EXT0; s.
    eapply pmread_plus; eauto; pmplus_eq'.
  }
  { split_pmvalid VLD'; eauto. }
  i; des; subst h0.

  eapply pesim_dance; first (by eauto); clear RLa0.
  - clear SV HV FIN.
    rename mrf into mrf1, mrf0 into mrf.

    dance_init.
    { eapply Sok; [by vauto|eapply Sopr].
      - readpc' mpc Eladdr.
        unfold push; eauto.
      - eauto.
      - eauto.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc' mpc Eladdr.
      - unfold stack_top; prep_store; eauto.
    }
    { s; tmext_simpl.
      repeat match goal with |- context[?a + 1 - 1] =>
        replace (a + 1 - 1) with a by omega
      end.
      depmcplus0; eauto using pmvalid_plus_cell.
    }

    dance_init.
    { eapply Sok; [by vauto|eapply Sld].
      - readpc' mpc Eladdr.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Salc]; [|by eauto].
      readpc' mpc Eladdr.
    }
    { depmcplus1; eauto using alloc_valid. }
    s; setnewloc.

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc' mpc Eladdr.
      - rewrite pmcells_cons -pmplus_assoc.
        prep_store; eauto.
    }
    { fold newloc0; tmext_simpl; s; pmcells_fold.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    dance_init.
    { eapply Sok; [by vauto|eapply Sld].
      - readpc' mpc Eladdr.
      - econs.
        rewrite (pmplus_assoc (snd s4)); set (tmp := (snd s4 ## _)).
        unfold mrf; s; tmext_simpl; rrdes EXT; rewrite -H4.
        eapply pmread_plus; first (eapply EXT0).
        + s; eapply pmread_plus; eauto; pmplus_eq'.
        + s; fold tmp; pmplus_eq'.
        + unfold tmp; rewrite -pmplus_assoc H4; eauto.
    }
    { depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells. }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc' mpc Eladdr.
      - rewrite 2!pmcells_cons -!pmplus_assoc (pmplus_swap {newloc0 |-> _}).
        prep_store; eauto.
    }
    { fold newloc0; tmext_simpl; s; pmcells_order; pmcells_fold.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    dance_init.
    { eapply Sok; [by vauto|eapply Sld].
      - readpc' mpc Eladdr.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Salc]; [|by eauto].
      readpc' mpc Eladdr.
    }
    { depmcplus1; eauto using alloc_valid. }
    s; setnewloc.

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc' mpc Eladdr.
      - rewrite pmcells_cons -pmplus_assoc.
        prep_store; eauto.
    }
    { fold newloc1; tmext_simpl; s; pmcells_fold.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc' mpc Eladdr.
      - rewrite 2!pmcells_cons -!pmplus_assoc (pmplus_swap {newloc1 |-> _}).
        prep_store; eauto.
    }
    { fold newloc1; tmext_simpl; s; pmcells_order; pmcells_fold; noemp.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    dance_init.
    { eapply Sok; [by vauto|eapply Sld].
      - readpc' mpc Eladdr.
      - eauto.
    }
    { depmcplus0. }

    eapply dance_O; instantiate (1 := fun _ _ _ cf _ _ _ _ cf' _ _ _ => depmach2 cf cf').
    s; tmext_simpl; eauto.

  - i; rrdes XTRb RLb0; clarify.
    esplits; first 1 last.
    { right; split; eauto using wfo_le_n_add. }
    { eauto. }
    
    i; frames_dep 3 RLa XTRa cfas; frames_dep 4 RLa XTRa cfas; frames_nodep RLa cfas.
    destruct ea'; clear SV HV FIN.
    destruct_tgt_real RLa.
    setnewloc; setcrf mrf'.
    set (s' := (mrf', P, env0, s0, ({mrf0 r_sp |-> mrf0 r_env} ## s1,
             {newloc1 |=> [mrf0 r_arg; newloc0]} ## {newloc0 |=> [mrf0 r_clo; mrf r_env]} ## snd s4 ## h''))).

    assert (VLD'': pmvalid ({newloc1 |=> [mrf0 r_arg; newloc0]} ##
                       {newloc0 |=> [mrf0 r_clo; mrf r_env]} ## snd s4 ## h'')).
    { clear -HV.
      rewrite -!(pmplus_swap hp0) -!(pmplus_swap refha0) in HV.
      eapply pmvalid_plusD, pmvalid_plusD; eauto.
    }

    assert (EXT'': ext (wlift mylw) s' (mrf0, P, env0, s0, (s1, snd s4 ## h''))).
    { clear -VLD''.
      s; splits; tmext_simpl; auto.
      i; eapply pmread_plus; eauto.
      set (tmp := snd s4 ## _); pmplus_eq'.
    }

    assert (env0 = env) by (by rrdes EXT EXT EXT EXT2).
    subst env.

    exists s'; esplits; [exact EXT''|].

    rewrite <-subexp_tmext_l2; last first.
    { eauto. }
    { eapply ts_gw_vclos_closed; eauto. }
    { eauto using subclo_weakly. }
    rewrite 2!tmext_lift.

    eapply paco10_mon.
    rewrite tgt_csplit src_csplit.
    eapply IH.
    + eauto.
    + eauto.
    + exact EXT''.
    + i; s in IN; des; vauto.
      apply RSV in IN. des. esplits; eauto.
      eapply vclos_mon; eauto using ext_trans.
      i; eapply gk_mon; eauto using ext_trans.
    + eauto.
    + instantiate (1 := (x0, mrf0 r_arg)::(f0, mrf0 r_clo)::denva).
      i; apply list_assoc_cons_or in IN; des; subst.
      { rewrite list_assoc_cons_eq; esplits; eauto.
        tmext_simpl.
        eapply vclos_mon; eauto using ext_trans.
        i; eapply gk_mon; eauto using ext_trans.
      }
      rewrite tmext_neq; [|by vauto].
      rewrite list_assoc_cons_neq; [|by ii; vauto].
      apply list_assoc_cons_or in IN0; des; subst.
      { rewrite list_assoc_cons_eq; esplits; eauto.
        tmext_simpl.
        
        econs; apply vclgen_thunk, vclf_fix.
        { unfold s'; econs; rr; eauto. }
        { rr; splits; s; eauto using CL'. }
        eapply gk_lsim; eauto.
        econs; [econs|].
        { unfold s'; econs; rr; eauto. }
        { rr; splits; s; eauto using CL'. }
        i; eapply gres_incl_id.
        right; eapply CIH; eauto using ext_trans.
        rewrite H4; eauto.
      }
      rewrite tmext_neq; [|by vauto].
      rewrite list_assoc_cons_neq; [|by ii; vauto].
      apply RDV in IN1; des.
      esplits; eauto.
      eapply vclos_mon; eauto using ext_trans.
      i; eapply gk_mon; eauto using ext_trans.
    + unfold s', mrf'; s; tmext_simpl; esplits
      ; try (by eapply pmread_plus; eauto; rewrite !pmcells_cons; pmplus_eq').
      eapply reprlist_mon; eauto.
      i; eapply pmread_plus; eauto; pmplus_eq'.
    + s; rewrite DORD; auto.
    + i; simpl map in IN.
      apply in_consD in IN; des; subst.
      { rewrite tmext_eq in FH; eauto using ts_gw_vclos_closed. }
      rewrite tmext_neq in FH; auto.
      apply in_consD in IN; des; subst.
      { rewrite tmext_eq in FH; eauto using ts_gw_vclos_closed. }
      rewrite tmext_neq in FH; eauto.
    + rewrite <-tmext_lift; eapply weakclo_tmext; last first.
      { i; exfalso; eauto using ts_gw_vclos_closed. }
      rewrite <-tmext_lift; eapply weakclo_tmext; last first.
      { i; exfalso; eauto using ts_gw_vclos_closed. }
      eauto.
    + vauto.
    + unfold s'; simpl_proj; econs; symmetry.
      rewrite Eladdr; simpl tr_exp.
      rewrite !pmcells_cons map_app pmcells_app.
      rewrite <-!pmplus_assoc.
      instantiate (2 := 22 + n).
      replace (_ - (22 + n)) with (mpc - n); [|omega].
      rhs rewrite <-!(pmplus_swap {mpc - n |=> _}).
      pmplus_eq'.
    + eauto.
    + 
      clear SV HV FIN FIN0.
      destruct s'0 as [[[[]]]]; clean.
      i; ts_prepare.

      hexploit (@pmplus_from_read _ _ _ (snd s4) (snd s5)).
      { i; rrdes EXT0; apply EXT1; s.
        eapply pmread_plus; eauto; pmplus_eq'.
      }
      { split_pmvalid VLD'; eauto. }
      i; des; rewrite ->H1 in *.

      eapply pesim_dance
      ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
      ; clear RLa0 SV HV FIN.

      * s in Eladdr.            

        dance_init.
        { eapply Sok; [s; i; omega|eapply Sld].
          - readpc' mpc Eladdr.
          - econs; s.
            eapply pmread_plus; eauto; rewrite LST; pmplus_eq'.
        }
        { depmcplus0. }

        dance_init.
        { eapply Sok; [by vauto|eapply Sopr].
          - readpc' mpc Eladdr.
            unfold pop; eauto.
          - eauto.
          - eauto.
          - eauto.
        }
        { depmcplus0. }

        dance_init.
        { eapply Sok; [s; i; omega|eapply Sld].
          - readpc' mpc Eladdr.
          - econs; s; tmext_simpl.
            eapply pmread_plus; eauto.
            rewrite LST SP; unfold s'; clean.
            unfold mrf'; tmext_simpl.
            repeat match goal with |- context[?a + 1 + 1 - 2] =>
              replace (a + 1 + 1 - 2) with a by omega
            end.                
            pmplus_eq'.
        }
        { depmcplus0. }

        dance_init.
        { eapply Sok; [by vauto|eapply Sjmp].
          - readpc' mpc Eladdr.
          - eauto.
        }
        { depmcplus0. }

        eapply dance_O; instantiate (1 := fun _ _ _ => depmach0).
        s; tmext_simpl; eauto.

      * clear RLb0.
        i; rrdes XTRb RLb0; clarify.
        s in Ec; clarify.
        esplits; first 1 last.
        { right; split; eauto using wfo_le_n_add. }
        { inst_pairs; rr; esplits; s; noemp; eauto. }
    
        i; frames_nodep RLa cfas.
        destruct ea'; clear FIN.
        destruct_tgt_real RLa.
        simpl lang_tgt.chp in *.
        setcrf mrf''.
        set (s'' := (mrf'', P1, env, s2, (s1, snd s5))).

        assert (VLD''': pmvalid (snd s5)).
        { rewrite H1; decomp_pmvalid HV; auto. }

        assert (EXT''': ext (wlift mylw) s'' (mrf1, P1, env, s2, s5)).
        { unfold s'', mrf''; clear -HV VLD'''.
          s; splits; tmext_simpl; auto.
        }

        exists s''; esplits; [by eauto|].
        rewrite tgt_csplit src_csplit.
        eapply pesim_release; last first; try by rr; s; eauto.
        { unfold s''; rr; esplits.
          - rr; esplits; eauto.
            + s; rr; esplits; clean; eauto.
              eapply tgt_confpred_pop; eauto; last first.
              * rewrite LST SP in SV; unfold s' in SV; unfold mrf' in SV.
                s in SV; tmext_simplH SV.
                repeat match goal with H: context[?a + 1 - 1] |- _ =>
                  replace (a + 1 - 1) with a in H by omega
                end.
                unfold mrf''; rewrite pmplus_comm; tmext_simpl.
                rewrite SP; s; unfold mrf'; tmext_simpl.
                repeat match goal with |- context[?a + 1 + 1 - 2] =>
                  replace (a + 1 + 1 - 2) with a by omega
                end.
                rewrite (pmplus_swap {_|->_}) in SV.
                rewrite plus_comm in SV; simpl plus in SV.
                rewrite (pmcells_sng va) in SV; pmcells_foldH SV.
                
                decomp_pmvalid SV; eauto.
              * unfold mrf''; tmext_simpl; s; auto.
            + dup EXT'''; rrdes EXT'''; simpl_proj.
              eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
              * s; i; eapply gk_mon; eauto; split; eauto.
              * s; i; eapply gk_mon; eauto; split; eauto.
          - rr; esplits; eauto.
          - s; rewrite LST SP H1 -!pmplus_assoc; noemp; s; f_equal.
            unfold mrf''; tmext_simpl; rewrite SP; s.
            unfold mrf'; tmext_simpl.
            rewrite -!(pmplus_swap freest).
            repeat match goal with |- context[?a + 1 + 1 - 2] =>
              replace (a + 1 + 1 - 2) with a by omega
            end.
            repeat match goal with |- context[?a + 1 - 1] =>
              replace (a + 1 - 1) with a by omega
            end.
            
            rewrite plus_comm (pmplus_swap {_ |-> _}); auto.
          - s; noemp; eauto.
        }

        invc RK; fold tsubst_name.
        rewrite RET; s; tmext_simpl; unfold mrf'; tmext_simpl.
        eapply ts_pesim_return.
        { s; unfold mrf''; tmext_simpl.
          eapply vclos_mon; eauto; i; eapply gk_mon; eauto.
        }
        { split; split; s.
          - rrdes EXT0 EXT0 EXT0.
            splits; auto.
            + unfold mrf''; tmext_simpl.
              rewrite SP; s; unfold mrf'; tmext_simpl.
              omega.
            + unfold mrf''; tmext_simpl; auto.
          - rrdes EXT0 EXT0. apply EXT2.
          - rrdes EXT0 EXT0. s in EXT1.
            rrdes EXT'' EXT''. s in EXT''0.
            unnw. eauto.
          - auto.
        }
    + unfold s'; rr; esplits.
      * rr; esplits; eauto.
        { s; rr; esplits; clean; eauto.
          rewrite pmcells_sng in CPa.
          eapply tgt_confpred_push; eauto.
          unfold mrf'; tmext_simpl; s; auto.
        }
        { dup EXT''; rrdes EXT''; simpl_proj.
          eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
          - s; i; eapply gk_mon; eauto; split; eauto.
          - s; i; eapply gk_mon; eauto; split; eauto.
        }
      * rr; esplits; eauto.
      * s. noemp; rewrite pmplus_swap; f_equal.
        rewrite 2!(pmplus_swap (hp0 ## _)) -!pmplus_assoc; eauto.
      * s; noemp; eauto.
    + i; vauto.
Grab Existential Variables.
  exact i.
Qed.


Lemma tr_exp_correct_fix: forall F G ty ty' f x e
    (IH: ts_psim F ((x, ty)::(f, Tarrow ty ty')::G) ty' (fun n => map enc (tr_exp e n (map fst F) (x::f::map fst G))) e)
    (HT: hastypepre  F ((x, ty)::(f, Tarrow ty ty')::G) e ty'), 
  ts_psim F G (Tarrow ty ty') (fun n => map enc (tr_exp (Efix f x e) n (map fst F) (map fst G))) (Efix f x e).
  
Proof.
  i; rdes IH; econs; r; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.

  i; ts_prepare.

  
  hexploit (tgt_stack CPa 1); i; des; s in FST; subst freest.
  des; repeat (destruct vs; [by vauto|]); destruct vs; [|by vauto].
  rewrite <-!pmplus_assoc, <-pmcells_sng in *; clear NST.

  eapply pesim_dance
  ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
  ; clear RLa0 SV HV FIN.

  - dance_init.
    { eapply Sok; [s; i; omega|eapply Sld].
      - readpc' mpc0 Eladdr.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [s; i; omega|eapply Salc]; [|by eauto].
      readpc' mpc0 Eladdr.
    }
    { depmcplus1; eauto using alloc_valid. }
    s; setnewloc.

    dance_init.
    { eapply Sok; [by vauto|eapply Slpc].
      readpc' mpc0 Eladdr.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Sopr].
      - readpc' mpc0 Eladdr.
      - eauto.
      - eauto.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc' mpc0 Eladdr.
      - rewrite pmcells_cons -2!pmplus_assoc (pmplus_swap {_|->_}).
        prep_store; eauto.
    }
    { s; tmext_simpl; noemp; pmcells_order; pmcells_fold.
      rewrite (pmplus_swap freest0).
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    fold newloc.                

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc' mpc0 Eladdr.
      - rewrite 2!pmcells_cons -!pmplus_assoc (pmplus_swap {newloc |-> _}).
        prep_store; eauto.
    }
    { fold newloc; tmext_simpl; s; pmcells_order; pmcells_fold.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    dance_init.
    { eapply Sok; [by vauto|eapply Sopr].
      - readpc' mpc0 Eladdr.
        unfold push; eauto.
      - eauto.
      - eauto.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc' mpc0 Eladdr.
      - unfold stack_top; prep_store; eauto.
    }
    { s; tmext_simpl.
      repeat match goal with |- context[?a + 1 - 1] =>
        replace (a + 1 - 1) with a by omega
      end.
      depmcplus0; eauto using pmvalid_plus_cell.
    }

    dance_init.
    { eapply Sok; [by vauto|eapply Sopr].
      - readpc' mpc0 Eladdr.
      - eauto.
      - eauto.
      - eauto.
    }
    { depmcplus0. }
    tmext_uniq.

    dance_init.
    { eapply Sok; [by vauto|eapply Sjmp].
      - readpc' mpc0 Eladdr.
      - eauto.
    }
    { depmcplus0. }

    eapply dance_O; instantiate (1 := fun _ cf _ _ _ _ _ _ _ _ => depmach1 cf).
    s; tmext_simpl; eauto.

  - i; rrdes XTRb RLb0; clarify.
    s in Ec; clarify.
    esplits; first 1 last.
    { right; split; eauto using wfo_le_n_add. }
    { inst_pairs; rr; esplits; s; noemp; eauto. }
    
    i; frames_dep 1 RLa XTRa cfas; frames_nodep RLa cfas.
    destruct ea'; clear FIN.
    destruct_tgt_real RLa.
    simpl lang_tgt.chp in *.
    setnewloc.
    setcrf mrf'.
    set (s' := (mrf', extend P0 (ts_fun (S (S (mpc + 8)))) newloc, env, s3,
                ({mrf r_sp |-> newloc} ## fst s4, {S newloc |-> mrf r_env} ## snd s4))).

    assert (VLD': pmvalid ({S newloc |-> mrf r_env} ## snd s4)).
    { clear -HV.
      rewrite pmcells_cons -pmplus_assoc (pmplus_swap {newloc |-> _}) -pmcells_sng in HV.
      rewrite <-(pmplus_comm (snd s4)), <-2(pmplus_swap (snd s4)) in HV.
      do 2 eapply pmvalid_plus_l; rewrite <-3pmplus_assoc; eauto.
    }

    assert (EXT': ext (wlift mylw) s' (mrf, P0, env, s3, s4)).
    { unfold s', mrf'; clear -HV VLD'.
      s; splits; tmext_simpl; auto; [by i; right; eauto|].
      i; eapply pmread_plus; eauto; pmplus_eq'.
    }

    exists s'; esplits; [by eauto|].


    replace (S _) with (length (map enc (tr_exp (Efix f x e) n (map fst F) (map fst G))) + mpc)
    ; last by (rewrite map_length; s; rewrite app_length; s; omega).

    assert (CL': forall a,
                  ~ fv_exp (Vfix f x (subst_exp [Evar x//x]>>[Evar f//f]>>denvb e)) a).
    { simpl fv_exp; i; des.
      eapply subst_exp_closed; eauto; i.
      destruct (LEM (x0 = x)) eqn:Ex.
      { subst; rewrite tmext_eq in FH2; s in FH2; subst; vauto. }
      rewrite tmext_neq in FH2; auto.
      destruct (LEM (x0 = f)) eqn:Ef.
      { subst; rewrite tmext_eq in FH2; subst; vauto. }
      rewrite tmext_neq in FH2; auto.
      eapply hastype_fv in FV; last first; eauto.
      s in FV; des; eauto. }
    

    hexploit KSIM; clear KSIM.
    { eauto. }
    { eauto. }
    { eauto. }
    { s; unfold mrf'; tmext_simpl; auto. }
    { s; unfold mrf'; tmext_simpl; auto. }
    { s; unfold mrf'; tmext_simpl; auto. }
    { s; unfold mrf'; tmext_simpl.
      repeat match goal with |- context[?a + 1 - 1] =>
        replace (a + 1 - 1) with a by omega
      end; eauto.             
    }
    { eapply tr_exp_correct_fix_val; s; eauto.
      unfold newloc; eauto.
    }

    intro X; simpl tr_exp in X.
    exploit X; clear X.
    { unfold s'; rr; esplits; eauto; [|by rr; esplits; eauto].
      rr; esplits; eauto.
      - rr; splits; [|by rr; esplits; eauto]; s.
        eapply tgt_confpred_extend; first eapply (tgt_confpred_push mrf').
        + rewrite <-pmcells_sng; eauto.
        + unfold mrf'; s; tmext_simpl; auto.
        + s; eauto.
        + eauto.
        + rewrite pmcells_cons -pmplus_assoc in HV.
          decomp_pmvalid HV; eauto.
      - dup EXT'; rrdes EXT'; simpl_proj.
        eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
        + s; i; eapply gk_mon; eauto; split; eauto.
        + s; i; eapply gk_mon; eauto; split; eauto.
    }

    simpl cplus; noemp; intro ES.
    eapply pesim_rewrite; eauto.
    rewrite pmplus_swap; f_equal.
    rewrite <-!pmplus_assoc, <-2!(pmplus_swap {_|->_}); pmcells_fold; auto.
Qed.


Lemma tr_exp_correct_gen_val: forall
    F G ty e i tsub g P0 s3 s4 senva denva n h kb mpc mrf mb aa
    refha refhb hp env chp newloc mrf'
    (IH: ts_psim_aux i F G (tsubst_bvar 0 (Tname aa) ty)
                     (fun n : word => map enc (tr_exp e n (map fst F) (map fst G))) e)
    (HT: hastypepre F G e (tsubst_bvar 0 (Tname aa) ty))
    (LCT: lc_tsub tsub)
    (GK: gknow (wlift mylw) g)
    (SORD : map fst F = map fst senva)
    (denvb: termvar -> lang_spec.val src_lsi)
    (DORD: map fst G = map fst denva)
    (CL: forall x, In x (map fst G) -> forall y : termvar, fv_exp (denvb x) y -> False)
    (WCL: subst_weakly_closed (fun x : termvar => denvb x))
    (ENV: reprlist (snd s4) (mrf r_env) (map snd denva))
    (Eladdr: snd s4 =
             {mpc |=> map enc (tr_exp (Egen e) n (map fst F) (map fst G))}
               ## {mpc - n |=> map snd senva} ## h)
    (RSV: forall (l : label) (ty' : typ),
            lookup l F = Some ty' ->
            exists (va : lang_spec.val tgt_lsi) (vb : val),
              lookup l senva = Some va /\
              lookup l env = Some vb /\
              vclos (wlift mylw) (mrf, P0, env, s3, s4)
                    (g (mrf, P0, env, s3, s4)) (tsubst_name tsub ty') va vb)
    (RDV: forall (x : termvar) (ty' : typ),
            lookup x G = Some ty' ->
            exists va : lang_spec.val tgt_lsi,
              lookup x denva = Some va /\
              vclos (wlift mylw) (mrf, P0, env, s3, s4)
                    (g (mrf, P0, env, s3, s4)) (tsubst_name tsub ty') va
                    (denvb x))
    (CPb: src_confpred env (pmemp, Just env, Nothing))
    (RLb: real src_lsi (refhb, Just env,
                        Just (app_cont kb (Egen (subst_exp denvb e)))) mb)
    (Enl: newloc = alloc (hp ## refha ## snd s4 ## chp) 2)
    (Emrf': mrf' = [S (S (mpc + 8 + (length (tr_exp e
                       (S (S (S (S (S (S (S (S (S (S (S (S (S n)))))))))))))
                       (map fst F) (map fst G)) + 7))) // r_arg]>>
     [mrf r_sp + 1 // r_sp]>> [newloc // r_clo]>> mrf)
    (s': state (wlift mylw))
    (Es': s' = (mrf', extend P0 (ts_fun (S (S (mpc + 8)))) newloc, env, s3,
                ({mrf r_sp |-> newloc} ## fst s4, {S newloc |-> mrf r_env} ## snd s4)))
    (VLD': pmvalid ({S newloc |-> mrf r_env} ## snd s4))
    (EXT': ext (wlift mylw) s' (mrf, P0, env, s3, s4))
    (CL': forall a : termvar, fv_exp (Vgen (subst_exp denvb e)) a -> False)
    (FTF: ~ ftn_typ_l (ty :: map snd F) aa)
    (FTG: ~ ftn_typ_l (ty :: map snd G) aa),
  vclos (wlift mylw) s' (g s') (tsubst_name tsub (Tforall ty)) newloc
        (Vgen (subst_exp denvb e)).
Proof.
  econs; apply vclgen_thunk, vclf_gen.
  { rewrite Es'; econs; rr; eauto. }
  { rr; splits; s; eauto using CL'. }

  eapply gk_lsim; eauto.
  evar (i0: wfo_set (wlift mylw)).
  econs; [econs|].
  { rewrite Es'; econs; rr; eauto. }
  { rr; splits; s; eauto using CL'. }

  insts $ i0.
  pcofix CIH; unfold i0 in *; clear i0.
  i; invc STK; [by invc APPb RF CB VR|].
  eapply pesim_gknow; i; eapply pesim_gknow_; i.
  invc INSTa INSTb; invc RF CB VR; rrdes CA; s in VAL; clarify.
  destruct s'0 as [[[[]]]]; clean.
  ren PC.
  destruct s1.
  apply pesim_acquire.
  ts_prepare.
  setnewloc.

  assert (Epc: mpc0 = 10 + mpc).
  { rrdes EXT EXT EXT EXT.
    hexploit tgt_vquery_fun_uniq; [by eauto|by eauto| |].
    - eapply EXT; s; rr; eauto.
    - rewrite plus_comm; eauto.
  }
  subst mpc0.

  clear CA CPb.

  
  hexploit (tgt_stack CPa 1); i; des; s in FST; subst freest.
  des; repeat (destruct vs; [by vauto|]); destruct vs; [|by vauto].
  rewrite <-!pmplus_assoc, <-pmcells_sng in *; clear NST.

  hexploit (@pmplus_from_read _ _ _ (snd s4) h0).
  { i; rrdes EXT; apply EXT0; s.
    eapply pmread_plus; eauto; pmplus_eq'.
  }
  { split_pmvalid VLD'; eauto. }
  i; des; subst h0.

  eapply pesim_dance; first (by eauto); clear RLa0.
  - clear SV HV FIN.
    rename mrf into mrf1, mrf0 into mrf.

    dance_init.
    { eapply Sok; [by vauto|eapply Sopr].
      - readpc' mpc Eladdr.
        unfold push; eauto.
      - eauto.
      - eauto.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc' mpc Eladdr.
      - unfold stack_top; prep_store; eauto.
    }
    { s; tmext_simpl.
      repeat match goal with |- context[?a + 1 - 1] =>
        replace (a + 1 - 1) with a by omega
      end.
      depmcplus0; eauto using pmvalid_plus_cell.
    }

    dance_init.
    { eapply Sok; [by vauto|eapply Sld].
      - readpc' mpc Eladdr.
      - econs.
        rewrite (pmplus_assoc (snd s4)); set (tmp := (snd s4 ## _)).
        unfold mrf; s; tmext_simpl; rrdes EXT; rewrite -H3.
        eapply pmread_plus; first (eapply EXT0).
        + s; eapply pmread_plus; eauto; pmplus_eq'.
        + s; fold tmp; pmplus_eq'.
        + unfold tmp; rewrite -pmplus_assoc; eauto.
    }
    { depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells. }

    eapply dance_O; instantiate (1 := fun _ _ => depmach0).
    s; tmext_simpl; eauto.

  - i; rrdes XTRb RLb0; clarify.
    esplits; first 1 last.
    { right; split; eauto using wfo_le_n_add. }
    { eauto. }
    
    i; frames_nodep RLa cfas.
    destruct ea'; clear SV HV FIN.
    destruct_tgt_real RLa.
    setnewloc; setcrf mrf'.
    set (s' := (mrf', P, env0, s0, ({mrf0 r_sp |-> mrf0 r_env} ## s1, snd s4 ## h''))).

    assert (VLD'': pmvalid (snd s4 ## h'')).
    { clear -HV.
      eapply pmvalid_plusD, pmvalid_plusD; eauto.
    }

    assert (EXT'': ext (wlift mylw) s' (mrf0, P, env0, s0, (s1, snd s4 ## h''))).
    { clear -VLD''.
      s; splits; tmext_simpl; auto.
    }

    assert (env0 = env) by (by rrdes EXT EXT EXT EXT2).
    subst env.

    exists s'; esplits; [exact EXT''|].

    eapply paco10_mon.
    rewrite tgt_csplit src_csplit.
    eapply (IH ([tyi//aa]>>tsub)).
    + ii; destruct (LEM (x = aa)); subst.
      * rewrite tmext_eq; eauto.
      * rewrite tmext_neq; eauto.
    + eauto.
    + eauto.
    + i; s in IN; des; vauto.
      dup IN; apply RSV in IN; des; esplits; eauto.
      erewrite tsubname_agree.
      * eapply vclos_mon; eauto using ext_trans.
        i; eapply gk_mon; eauto using ext_trans.
      * i; clear -FTF IN0 IN3.
        assert (x <> aa).
        { ii; subst; induction F; vauto.
          destruct a; apply list_assoc_cons_or in IN0; des; subst
          ; [apply FTF|eapply IHF]; ss; i; des; eauto using FTF.
        }
        rewrite tmext_neq; eauto. 
    + eauto.
    + instantiate (1 := denva).
      i; dup IN; apply RDV in IN; des.
      esplits; eauto.
      erewrite tsubname_agree.
      * eapply vclos_mon; eauto using ext_trans.
        i; eapply gk_mon; eauto using ext_trans.
      * i; clear -FTG IN0 IN2.
        assert (x0 <> aa).
        { ii; subst; induction G; vauto.
          destruct a; apply list_assoc_cons_or in IN0; des; subst
          ; [apply FTG|eapply IHG]; ss; i; des; eauto using FTG.
        }
        rewrite tmext_neq; eauto. 
    + unfold s', mrf'; s; tmext_simpl; esplits
      ; try (by eapply pmread_plus; eauto; rewrite !pmcells_cons; pmplus_eq').
      eapply reprlist_mon; eauto.
      i; eapply pmread_plus; eauto; pmplus_eq'.
    + s; rewrite DORD; auto.
    + eauto.
    + eauto.
    + vauto.
    + unfold s'; simpl_proj; econs; symmetry.
      rewrite Eladdr; simpl tr_exp.
      rewrite !pmcells_cons map_app pmcells_app.
      rewrite <-!pmplus_assoc.
      instantiate (2 := 13 + n).
      replace (_ - (13 + n)) with (mpc - n); [|omega].
      rhs rewrite <-!(pmplus_swap {mpc - n |=> _}).
      pmplus_eq'.
    + eauto.
    + 
      clear SV HV FIN FIN0.
      destruct s'0 as [[[[]]]]; clean.
      i; ts_prepare.

      hexploit (@pmplus_from_read _ _ _ (snd s4) (snd s5)).
      { i; rrdes EXT0; apply EXT1; s.
        eapply pmread_plus; eauto; pmplus_eq'.
      }
      { split_pmvalid VLD'; eauto. }
      i; des; rewrite ->H1 in *.

      eapply pesim_dance
      ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
      ; clear RLa0 SV HV FIN.

      * s in Eladdr.            

        dance_init.
        { eapply Sok; [s; i; omega|eapply Sld].
          - readpc' mpc Eladdr.
          - econs; s.
            eapply pmread_plus; eauto; rewrite LST; pmplus_eq'.
        }
        { depmcplus0. }

        dance_init.
        { eapply Sok; [by vauto|eapply Sopr].
          - readpc' mpc Eladdr.
            unfold pop; eauto.
          - eauto.
          - eauto.
          - eauto.
        }
        { depmcplus0. }

        dance_init.
        { eapply Sok; [s; i; omega|eapply Sld].
          - readpc' mpc Eladdr.
          - econs; s; tmext_simpl.
            eapply pmread_plus; eauto.
            rewrite LST SP; unfold s'; clean.
            unfold mrf'; tmext_simpl.
            repeat match goal with |- context[?a + 1 + 1 - 2] =>
              replace (a + 1 + 1 - 2) with a by omega
            end.                
            pmplus_eq'.
        }
        { depmcplus0. }

        dance_init.
        { eapply Sok; [by vauto|eapply Sjmp].
          - readpc' mpc Eladdr.
          - eauto.
        }
        { depmcplus0. }

        eapply dance_O; instantiate (1 := fun _ _ _ => depmach0).
        s; tmext_simpl; eauto.

      * clear RLb0.
        i; rrdes XTRb RLb0; clarify.
        s in Ec; clarify.
        esplits; first 1 last.
        { right; split; eauto using wfo_le_n_add. }
        { inst_pairs; rr; esplits; s; noemp; eauto. }
    
        i; frames_nodep RLa cfas.
        destruct ea'; clear FIN.
        destruct_tgt_real RLa.
        simpl lang_tgt.chp in *.
        setcrf mrf''.
        set (s'' := (mrf'', P1, env, s2, (s1, snd s5))).

        assert (VLD''': pmvalid (snd s5)).
        { rewrite H1; decomp_pmvalid HV; auto. }

        assert (EXT''': ext (wlift mylw) s'' (mrf1, P1, env, s2, s5)).
        { unfold s'', mrf''; clear -HV VLD'''.
          s; splits; tmext_simpl; auto.
        }

        exists s''; esplits; [by eauto|].
        rewrite tgt_csplit src_csplit.
        eapply pesim_release; last first; try by rr; s; eauto.
        { unfold s''; rr; esplits.
          - rr; esplits; eauto.
            + s; rr; esplits; clean; eauto.
              eapply tgt_confpred_pop; eauto; last first.
              * rewrite LST SP in SV; unfold s' in SV; unfold mrf' in SV.
                s in SV; tmext_simplH SV.
                repeat match goal with H: context[?a + 1 - 1] |- _ =>
                  replace (a + 1 - 1) with a in H by omega
                end.
                unfold mrf''; rewrite pmplus_comm; tmext_simpl.
                rewrite SP; s; unfold mrf'; tmext_simpl.
                repeat match goal with |- context[?a + 1 + 1 - 2] =>
                  replace (a + 1 + 1 - 2) with a by omega
                end.
                rewrite (pmplus_swap {_|->_}) in SV.
                rewrite plus_comm in SV; simpl plus in SV.
                rewrite (pmcells_sng va) in SV; pmcells_foldH SV.
                
                decomp_pmvalid SV; eauto.
              * unfold mrf''; tmext_simpl; s; auto.
            + dup EXT'''; rrdes EXT'''; simpl_proj.
              eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
              * s; i; eapply gk_mon; eauto; split; eauto.
              * s; i; eapply gk_mon; eauto; split; eauto.
          - rr; esplits; eauto.
          - s; rewrite LST SP H1 -!pmplus_assoc; noemp; s; f_equal.
            unfold mrf''; tmext_simpl; rewrite SP; s.
            unfold mrf'; tmext_simpl.
            rewrite -!(pmplus_swap freest).
            repeat match goal with |- context[?a + 1 + 1 - 2] =>
              replace (a + 1 + 1 - 2) with a by omega
            end.
            repeat match goal with |- context[?a + 1 - 1] =>
              replace (a + 1 - 1) with a by omega
            end.
            
            rewrite plus_comm (pmplus_swap {_ |-> _}); auto.
          - s; noemp; eauto.
        }

        invc RK; fold tsubst_name.
        rewrite RET; s; tmext_simpl; unfold mrf'; tmext_simpl.
        eapply ts_pesim_return.
        { s; unfold mrf''; tmext_simpl.
          rewrite tsub_name_bvar in VR.
          - s in VR. rewrite tmext_eq in VR.
            erewrite tsubname_agree in VR.
            + eapply vclos_mon; eauto; i; eapply gk_mon; eauto.
            + i; rewrite tmext_neq; eauto.
              ii; subst; apply FTF; s; eauto.
          - ii; destruct (LEM (x = aa)); subst.
            + rewrite tmext_eq; eauto.
            + rewrite tmext_neq; eauto. 
        }
        { split; split; s.
          - rrdes EXT0 EXT0 EXT0.
            splits; auto.
            + unfold mrf''; tmext_simpl.
              rewrite SP; s; unfold mrf'; tmext_simpl.
              omega.
            + unfold mrf''; tmext_simpl; auto.
          - rrdes EXT0 EXT0. apply EXT2.
          - rrdes EXT0 EXT0. s in EXT1.
            rrdes EXT'' EXT''. s in EXT''0.
            unnw. eauto.
          - auto.
        }
    + unfold s'; rr; esplits.
      * rr; esplits; eauto.
        { s; rr; esplits; clean; eauto.
          rewrite pmcells_sng in CPa.
          eapply tgt_confpred_push; eauto.
          unfold mrf'; tmext_simpl; s; auto.
        }
        { dup EXT''; rrdes EXT''; simpl_proj.
          eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
          - s; i; eapply gk_mon; eauto; split; eauto.
          - s; i; eapply gk_mon; eauto; split; eauto.
        }
      * rr; esplits; eauto.
      * s. noemp; rewrite pmplus_swap; f_equal.
        rewrite -!pmplus_assoc; eauto.
      * s; noemp; eauto.
    + i; vauto.
Grab Existential Variables.
  exact i.
Qed.


Lemma tr_exp_correct_gen: forall F G ty e aa
    (IH: ts_psim F G (tsubst_bvar 0 (Tname aa) ty) (fun n => map enc (tr_exp e n (map fst F) (map fst G))) e)
    (HT: hastypepre F G e (tsubst_bvar 0 (Tname aa) ty))
    (FTF: ~ ftn_typ_l (ty :: map snd F) aa)
    (FTG: ~ ftn_typ_l (ty :: map snd G) aa),
  ts_psim F G (Tforall ty) (fun n => map enc (tr_exp (Egen e) n (map fst F) (map fst G))) (Egen e).
Proof.
  i; rdes IH; econs; r; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.

  i; ts_prepare.

  
  hexploit (tgt_stack CPa 1); i; des; s in FST; subst freest.
  des; repeat (destruct vs; [by vauto|]); destruct vs; [|by vauto].
  rewrite <-!pmplus_assoc, <-pmcells_sng in *; clear NST.

  eapply pesim_dance
  ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
  ; clear RLa0 SV HV FIN.

  - dance_init.
    { eapply Sok; [s; i; omega|eapply Sld].
      - readpc' mpc0 Eladdr.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [s; i; omega|eapply Salc]; [|by eauto].
      readpc' mpc0 Eladdr.
    }
    { depmcplus1; eauto using alloc_valid. }
    s; setnewloc.

    dance_init.
    { eapply Sok; [by vauto|eapply Slpc].
      readpc' mpc0 Eladdr.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Sopr].
      - readpc' mpc0 Eladdr.
      - eauto.
      - eauto.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc' mpc0 Eladdr.
      - rewrite pmcells_cons -2!pmplus_assoc (pmplus_swap {_|->_}).
        prep_store; eauto.
    }
    { s; tmext_simpl; noemp; pmcells_order; pmcells_fold.
      rewrite (pmplus_swap freest0).
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    fold newloc.                

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc' mpc0 Eladdr.
      - rewrite 2!pmcells_cons -!pmplus_assoc (pmplus_swap {newloc |-> _}).
        prep_store; eauto.
    }
    { fold newloc; tmext_simpl; s; pmcells_order; pmcells_fold.
      depmcplus0; eauto using pmvalid_plus_cells, pmfinite_plus_cells.
    }

    dance_init.
    { eapply Sok; [by vauto|eapply Sopr].
      - readpc' mpc0 Eladdr.
        unfold push; eauto.
      - eauto.
      - eauto.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc' mpc0 Eladdr.
      - unfold stack_top; prep_store; eauto.
    }
    { s; tmext_simpl.
      repeat match goal with |- context[?a + 1 - 1] =>
        replace (a + 1 - 1) with a by omega
      end.
      depmcplus0; eauto using pmvalid_plus_cell.
    }

    dance_init.
    { eapply Sok; [by vauto|eapply Sopr].
      - readpc' mpc0 Eladdr.
      - eauto.
      - eauto.
      - eauto.
    }
    { depmcplus0. }
    tmext_uniq.

    dance_init.
    { eapply Sok; [by vauto|eapply Sjmp].
      - readpc' mpc0 Eladdr.
      - eauto.
    }
    { depmcplus0. }

    eapply dance_O; instantiate (1 := fun _ cf _ _ _ _ _ _ _ _ => depmach1 cf).
    s; tmext_simpl; eauto.

  - i; rrdes XTRb RLb0; clarify.
    s in Ec; clarify.
    esplits; first 1 last.
    { right; split; eauto using wfo_le_n_add. }
    { inst_pairs; rr; esplits; s; noemp; eauto. }
    
    i; frames_dep 1 RLa XTRa cfas; frames_nodep RLa cfas.
    destruct ea'; clear FIN.
    destruct_tgt_real RLa.
    simpl lang_tgt.chp in *.
    setnewloc.
    setcrf mrf'.
    set (s' := (mrf', extend P0 (ts_fun (S (S (mpc + 8)))) newloc, env, s3,
                ({mrf r_sp |-> newloc} ## fst s4, {S newloc |-> mrf r_env} ## snd s4))).

    assert (VLD': pmvalid ({S newloc |-> mrf r_env} ## snd s4)).
    { clear -HV.
      rewrite pmcells_cons -pmplus_assoc (pmplus_swap {newloc |-> _}) -pmcells_sng in HV.
      rewrite <-(pmplus_comm (snd s4)), <-2(pmplus_swap (snd s4)) in HV.
      do 2 eapply pmvalid_plus_l; rewrite <-3pmplus_assoc; eauto.
    }

    assert (EXT': ext (wlift mylw) s' (mrf, P0, env, s3, s4)).
    { unfold s', mrf'; clear -HV VLD'.
      s; splits; tmext_simpl; auto; [by i; right; eauto|].
      i; eapply pmread_plus; eauto; pmplus_eq'.
    }

    exists s'; esplits; [by eauto|].

    replace (S _) with (length (map enc (tr_exp (Egen e) n (map fst F) (map fst G))) + mpc)
    ; last (rewrite map_length; s; rewrite app_length; s; omega).

    assert (CL': forall a, ~ fv_exp (Vgen (subst_exp denvb e)) a).
    { simpl fv_exp; i; des.
      eapply subst_exp_closed; eauto; i.
      eapply hastype_fv in FV; eauto.
    }
    

    hexploit KSIM; clear KSIM.
    { eauto. }
    { eauto. }
    { eauto. }
    { s; unfold mrf'; tmext_simpl; auto. }
    { s; unfold mrf'; tmext_simpl; auto. }
    { s; unfold mrf'; tmext_simpl; auto. }
    { s; unfold mrf'; tmext_simpl.
      repeat match goal with |- context[?a + 1 - 1] =>
        replace (a + 1 - 1) with a by omega
      end; eauto.             
    }
    { eapply tr_exp_correct_gen_val; s; eauto.
      unfold newloc; eauto.
    }

    intro X; simpl tr_exp in X.
    exploit X; clear X.
    { unfold s'; rr; esplits; eauto; [|by rr; esplits; eauto].
      rr; esplits; eauto.
      - rr; splits; [|by rr; esplits; eauto]; s.
        eapply tgt_confpred_extend; first eapply (tgt_confpred_push mrf').
        + rewrite <-pmcells_sng; eauto.
        + unfold mrf'; s; tmext_simpl; auto.
        + s; eauto.
        + eauto.
        + rewrite pmcells_cons -pmplus_assoc in HV.
          decomp_pmvalid HV; eauto.
      - dup EXT'; rrdes EXT'; simpl_proj.
        eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
        + s; i; eapply gk_mon; eauto; split; eauto.
        + s; i; eapply gk_mon; eauto; split; eauto.
    }

    simpl cplus; noemp; intro ES.
    eapply pesim_rewrite; eauto.
    rewrite pmplus_swap; f_equal.
    rewrite <-!pmplus_assoc, <-2!(pmplus_swap {_|->_}); pmcells_fold; auto.
Qed.
