(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List Program.
Require Import ssr countable vgtac basis lang_common lang_spec lang_mid lang_mid_renaming.
Require Import model_common model_untyped model_properties_untyped gw_untyped.
Require Import hcomp_esim_untyped.

Set Implicit Arguments.
Global Set Bullet Behavior "Strict Subproofs".


(** * Properties about [mm_gw]. *)


Lemma mm_confrel_shape: forall g (s: state mm_gwsts) ha hb
    (WH: confrel mm_gw g s ha hb),
  exists ha' hb',
    ha = (ha', Nothing) /\
    hb = (hb', Nothing).
Proof.
  i; rrdes WH WHX WHY; rrdes CPa CPb; subst; noemp; eauto.
Qed.


(** Inversion for untyped value closure. *)

Lemma mm_vclosDb: forall (w: world_loca mm_gw) s r va vb
    (VCL: vclos (wlift w) s r va vb),
  match vb with
    | IVunit =>
        va = IVunit
    | IVnat n =>
        va = IVnat n
    | IVpair vb' vb'' => exists va' va'',
        va = IVpair va' va'' /\
        <<VCL': vclos (wlift w) s r va' vb'>> /\
        <<VCL'': vclos (wlift w) s r va'' vb''>>
    | IVinl vb' => exists va',
        va = IVinl va' /\
        <<VCL': vclos (wlift w) s r va' vb'>>
    | IVinr vb' => exists va',
        va = IVinr va' /\
        <<VCL': vclos (wlift w) s r va' vb'>>
    | IVloc _ => exists va',
        va = IVloc va' /\
        <<QR: mm_gw.(rquery) (fst s) va vb>>
    | IVclo envb (IEfix xfb xvb xkb eb) => exists enva xfa xva xka ea,
        va = IVclo enva (IEfix xfa xva xka ea) /\
        <<RV: r va vb>>
    | IVclo envb (IEgen xkb eb) => exists enva xka ea,
        va = IVclo enva (IEgen xka ea) /\
        <<RV: r va vb>>
    | _ => True
  end.
Proof.
  i; destruct vb eqn:E; rewrite <-E in VCL
  ; move VCL at top; revert_until VCL; pattern va, vb
  ; try (by apply (vclos_induction_strong VCL); i
         ; invc IH; des; try inv CA CB; eauto
         ; [rrdes VR; esplits; rr|inv CB|inv VR CA CB]; vauto).
  apply (vclos_induction_strong VCL); i.
  subst; invc IH; des; try invc CA CB; auto.
    + specialize (VR _ _ eq_refl); destruct e; vauto.
    + specialize (VR _ _ eq_refl); destruct e; vauto.
    + rrdes VR; vauto.
    + inv CB.
    + invc VR CA CB; eauto 10.
Grab Existential Variables.
exact IVunit. exact IVunit. exact IVunit. exact IVunit.
Qed.


(** Convenvience lemma for reasoning about [IEread]. *)

Lemma mm_refconfrel_read: forall g s ha hb la lb
    (RCR: mm_refconfrel g s (ha, Nothing) (hb, Nothing))
    (REF: snd s la lb),
  exists va vb,
    <<RDa: pmread ha la va>> /\
    <<RDb: pmread hb lb vb>> /\
    <<VCL: vclos mm_gw s (g s) va vb>>.
Proof.
  i; rrdes RCR; clarify; eauto.
Qed.

Lemma mm_confrel_read: forall (lw: world_loca mm_gw) g s ha hb la lb
    (RH: confrel (wlift lw) g s (ha, Nothing) (hb, Nothing))
    (VLDa: pmvalid ha)
    (VLDb: pmvalid hb)
    (DB: snd (fst s) la lb),
  exists va vb,
    <<RDa: pmread ha la va>> /\
    <<RDb: pmread hb lb vb>> /\
    <<VCL: vclos (wlift lw) s (g s) va vb>>.
Proof.
  i; dup RH; rrdes RH.
  hexploit mm_confrel_shape; eauto; []; i; des.
  rrdes WHX; dup WHY0; rrdes WHY0; subst.
  hexploit mm_refconfrel_read; eauto; []; i; des.
  hexploit lwconf_extra; [exact WHY|]; intros [XTRa XTRb].
  inversion XTRa; inversion XTRb; clear XTRa XTRb; subst.
  rrdes RDa RDb; rrdes WHX0 CPa CPb; s in HPA; s in HPB.
  subst; noemp; clarify.        

  hexploit mm_refconfrel_read; eauto; []; i; des; esplits.
  - eapply pmread_plus; eauto; rewrite pmplus_swap; eauto.
  - eapply pmread_plus; eauto; rewrite pmplus_swap; eauto.
  - destruct s; eapply vclos_lift; eauto.
Qed.


(** Convenvience lemmas for reasoning about [IEwrite]. *)

Lemma mm_refconfrel_write: forall g s la lb va vb ha hb
    (RCR: mm_refconfrel g s ({la |-> va} ## ha, Nothing)
                            ({lb |-> vb} ## hb, Nothing))
    (RL: snd s la lb)
    va' vb' (VCL: vclos mm_gw s (g s) va' vb'),
  mm_refconfrel g s ({la |-> va'} ## ha, Nothing)
                    ({lb |-> vb'} ## hb, Nothing).
Proof.
  i; rrdes RCR; clarify.
  rr; esplits; eauto using pmvalid_plus_cell.
  - i; eapply pmread_plusD in ALC; des.
    + eapply pmread_cellD in ALC; des; clarify.
      eauto using pmread_plus, pmread_cell.
    + eapply DOMa; eauto using pmread_plus, pmplus_comm.
  - i; eapply pmread_plusD in ALC; des.
    + eapply pmread_cellD in ALC; des; clarify.
      eauto using pmread_plus, pmread_cell.
    + eapply DOMb; eauto using pmread_plus, pmplus_comm.
  - i; dup REG; apply RELV in REG; des.
    apply pmread_plusD in RDa; apply pmread_plusD in RDb; des.
    + apply pmread_cellD in RDa; apply pmread_cellD in RDb; des; subst.
      destruct s; esplits; last (exact VCL)
      ; eauto using pmread_plus, pmread_cell, pmvalid_plus_cell.
    + exfalso.
      apply pmread_cellD in RDb; des; subst.
      hexploit (refst_bij REG0 RL); i; des.
      specialize (X0 eq_refl). subst.
      rrdes RDa; subst.
      apply RELV in RL; des.
      rewrite pmplus_assoc in VLDa.
      apply pmvalid_plus_l in VLDa.
      apply pmvalid_cell_cell in VLDa; vauto.
    + exfalso.
      apply pmread_cellD in RDa; des; subst.
      hexploit (refst_bij REG0 RL); i; des.
      specialize (X eq_refl). subst.
      rrdes RDb; subst.
      apply RELV in RL; des.
      rewrite pmplus_assoc in VLDb.
      apply pmvalid_plus_l in VLDb.
      apply pmvalid_cell_cell in VLDb; vauto.
    + eauto 10 using pmread_plus, pmplus_comm, pmvalid_plus_cell.
Qed.

Lemma mm_confrel_write: forall (lw: world_loca mm_gw) g s la lb va vb ha hb
    (CR: confrel (wlift lw) g s ({la |-> va} ## ha, Nothing)
                                ({lb |-> vb} ## hb, Nothing))
    (RL: snd (fst s) la lb)
    va' vb' (VCL: vclos (wlift lw) s (g s) va' vb')
    (VLDa: pmvalid ({la |-> va} ## ha))
    (VLDb: pmvalid ({lb |-> vb} ## hb)),
  confrel (wlift lw) g s ({la |-> va'} ## ha, Nothing)
                         ({lb |-> vb'} ## hb, Nothing).
Proof.
  i; dup CR; rrdes CR.
  hexploit mm_confrel_shape; eauto; []; i; des.
  rrdes WHX; dup WHY0; rrdes WHY0; subst.
  hexploit mm_refconfrel_read; eauto; []; i; des.
  hexploit lwconf_extra; [exact WHY|]; intros [XTRa XTRb].
  inversion XTRa; inversion XTRb; clear XTRa XTRb; subst.
  rrdes RDa RDb; rrdes WHX0 CPa CPb; s in HPA; s in HPB.
  subst; noemp; clarify.

  rewrite -!pmplus_assoc in H0; rewrite -H0 in VLDa.
  apply pmcell_uniq in H0; [|by auto]; des; subst.
  rewrite -!pmplus_assoc in H1; rewrite -H1 in VLDb.
  apply pmcell_uniq in H1; [|by auto]; des; subst.

  rr; esplits; [|by eauto| |].
  - rr; esplits; [by eauto| | |].
    + clear VCL0; destruct s.
      inst_pairs; eapply mm_refconfrel_write; eauto using vclos_unlift.
    + eauto.
    + eauto.
  - noemp; s; rewrite -pmplus_assoc; auto.
  - noemp; s; rewrite -pmplus_assoc; auto.
Qed.


(** Convenvience lemmas for reasoning about [IEnew]. *)

Lemma confrel_newa: forall {lw: world_loca mm_gw} {g s s0 s1 s2 ha hb la va}
    (RH: confrel (wlift lw) g (s, s0, s1, s2) (ha, Nothing) (hb, Nothing))
    (VLD: pmvalid ({la |-> va} ## ha)),
  forall lb', ~ s1 la lb'.
Proof.
  i; rrdes RH.
  hexploit mm_confrel_shape; eauto; []; i; des.
  hexploit lwconf_extra; [exact WHY|]; intros [XTRa XTRb].
  inversion XTRa; inversion XTRb; clear XTRa XTRb.
  rrdes WHX WHY0 WHX0; rrdes CPa CPb.
  subst; noemp; simpl cplus in *; clarify.
  apply RELV in FH; des; rrdes RDa; subst.
  rewrite -pmplus_assoc pmplus_assoc in VLD; decomp_pmvalid VLD.
  apply pmvalid_cell_cell in VLD4; vauto.
Qed.

Lemma confrel_newb: forall {lw: world_loca mm_gw} {g s s0 s1 s2 ha hb lb vb}
    (RH: confrel (wlift lw) g (s, s0, s1, s2) (ha, Nothing) (hb, Nothing))
    (VLD: pmvalid ({lb |-> vb} ## hb)),
  forall la', ~ s1 la' lb.
Proof.
  i; rrdes RH.
  hexploit mm_confrel_shape; eauto; []; i; des.
  hexploit lwconf_extra; [exact WHY|]; intros [XTRa XTRb].
  inversion XTRa; inversion XTRb; clear XTRa XTRb.
  rrdes WHX WHY0 WHX0; rrdes CPa CPb.
  subst; noemp; simpl cplus in *; clarify.
  apply RELV in FH; des; rrdes RDb; subst.
  rewrite -pmplus_assoc pmplus_assoc in VLD; decomp_pmvalid VLD.
  apply pmvalid_cell_cell in VLD4; vauto.
Qed.

Lemma confrel_alloc: forall (lw: world_loca mm_gw) g s s0 s1 s2 va vb ha hb la lb
    (GK: gknow (wlift lw) g)
    (SIM: vclos (wlift lw) (s, s0, s1, s2) (g (s, s0, s1, s2)) va vb)
    (RH: confrel (wlift lw) g (s, s0, s1, s2) (ha, Nothing) (hb, Nothing))
    (VLDa: pmvalid ({la |-> va} ## ha))
    (VLDb: pmvalid ({lb |-> vb} ## hb))
    (FINa: pmfinite ({la |-> va} ## ha))
    (FINb: pmfinite ({lb |-> vb} ## hb))
    (s1': refst) (S1: refst_alloc s1 la lb = s1')
    (STA: stable lw),
  exists s2',
    extpub _ s2' s2 /\
    confrel (wlift lw) g (s, s0, s1', s2')
            ({la |-> va} ## ha, Nothing) ({lb |-> vb} ## hb, Nothing).
Proof.
  i; rrdes RH.
  hexploit mm_confrel_shape; eauto; i; des; clarify.
  destruct hay, hby; vauto; s in HPB; s in HPA; clarify.
  hexploit lwconf_extra; [exact WHY|]; intros [XTRa XTRb].
  inversion XTRa; inversion XTRb; clear XTRa XTRb; clarify.

  assert (CR: forall s2' (EXT: extpub _ s2' s2),
                confrel mm_gw (fun sg : state mm_gw => g (sg, s2')) (s, s0, s1')
                        ({la |-> va} ## ha', Nothing) ({lb |-> vb} ## hb', Nothing)).
  {
    i; rrdes WHX; destruct hax, hbx, hay, hby; vauto.
    destruct e, e1; vauto; destruct e0, e2; vauto.
    s in HPA; s in HPB; clarify.

    rr; esplits; eauto; last first.
    - inst_pairs; s; rewrite <-(pmplus_swap h0); eauto.
    - inst_pairs; s; rewrite <-(pmplus_swap h); eauto.
    - rrdes WHY0; clarify.
      assert (pmvalid ({la |-> va} ## ha')).
      { rewrite (pmplus_comm h) 2!pmplus_assoc in VLDa.
        decomp_pmvalid VLDa; eauto.
      }
      assert (pmvalid ({lb |-> vb} ## hb')).
      { rewrite (pmplus_comm h0) 2!pmplus_assoc in VLDb.
        decomp_pmvalid VLDb; eauto.
      }
      unfold refst_alloc in *.
      rr; esplits; eauto.
      + i; apply pmread_plusD in ALC; des.
        { s; rewrite <-S1; apply pmread_cellD in ALC; des; s; eauto. }
        { s; rewrite <-S1; hexploit DOMa; eauto; i; des; s; eauto. }
      + i; apply pmread_plusD in ALC; des.
        { s; rewrite <-S1; apply pmread_cellD in ALC; des; s; eauto. }
        { s; rewrite <-S1; hexploit DOMb; eauto; i; des; s; eauto. }
      + clean; rewrite <-S1.
        s; i; des; subst.
        { apply RELV in REG; des; esplits.
          - eapply pmread_plus; eauto.
            rewrite <-pmplus_comm; eauto.
          - eapply pmread_plus; eauto.
            rewrite <-pmplus_comm; eauto.
          - eapply vclos_mon; s; try rewrite <-S1; eauto.
            i; eapply (gk_mon GK); rr; s; esplits; eauto.
            i; rewrite -S1; eauto.
        }
        { esplits.
          - eapply pmread_plus; eauto using pmread_cell.
          - eapply pmread_plus; eauto using pmread_cell.
          - eapply vclos_unlift in SIM; eapply vclos_mon
            ; s; try rewrite <-S1; eauto.
            i; eapply (gk_mon GK); rr; s; esplits; eauto.
            i; rewrite -S1; eauto.
        }
  }

  exploit STA.
  { eauto. }
  { eauto. }
  { split; last first.
    - inst_pairs; eapply refst_alloc_ext; eauto.
    - s; auto.
  }
  { insts $ s0 s.
    set (dummy := (pmemp, Just (nil, IEunit)) : conf mid_lsi).
    rr; esplits; econs.
    - eauto.
    - insts $ dummy dummy; split; unfold dummy; s; eauto.
    - instantiate (2:=cemp); inst_pairs; s; noemp.
      unfold dummy; rewrite -pmplus_assoc; rr; esplits; eauto.
    - instantiate (2:=cemp); inst_pairs; s; noemp.
      unfold dummy; rewrite -pmplus_assoc; rr; esplits; eauto.
  }
  
  i; des; esplits; eauto.
  rr; esplits; clean; eauto
  ; try (by rewrite pmplus_assoc; inst_pairs; s; noemp; eauto).
Qed.



(** Tactic for destructing some basic world-related assumptions that show up
    in the compatibility proofs.
*)

Ltac mm_destruct XTRa XTRb CNF :=
  inversion XTRa; inversion XTRb; clear XTRa XTRb 
  ; inversion CNF as [RH CO MA MB]; clear CNF
  ; dup RH; destruct RH as (? & ? & ? & ? & WHX & WHY & ? & ?)
  ; apply mm_confrel_shape in WHX; des
  ; destruct MA as (? & ? & Eca & Ema & PVa & FINa)
  ; destruct MB as (? & ? & Ecb & Emb & PVb & FINb)
  ; unnw; subst; hexploit lwconf_extra; [exact WHY|]; intros [XTRa XTRb]
  ; inversion XTRa; inversion XTRb; clear XTRa XTRb; subst
  ; s in Eca; s in Ecb; clarify; noemp.


(** Convenience lemma for reasoning via esim_step.

    This basically says that when applying the esim_step case of
    esim_main, it is okay to assume that the given step on the target
    side was successful (not an error step), IF we can separately show
    that if the source can make a successful step, then so can the
    target (in other words: if the target makes an error step, then so
    does the source).

    TODO: To make the lemma more useful, we should add an assumption
    to the step case of mm_ok_esim_main saying that the source can
    take a successful step.
*)

Inductive mm_ok_esim_main (w: world mid_lsi mid_lsi) esimstep
  kia kib g si s cfa ma cfb mb : Prop :=
| mm_ok_esim_err
    mb' (STP: rtc (step _ etau) mb mb')
    (ER: mid_lsi.(error) mb')
| mm_ok_esim_step
    (NT: ~ mid_lsi.(halted) ma)
    (EST: forall evt ma' (STa: reduce_ok evt (fst ma) (fst (snd ma)) (snd (snd ma))
                                             (fst ma') (fst (snd ma')) (snd (snd ma'))),
          exists ea' csa' eb' csb' mb' mbt s' sync,
          <<STP: rtc (step _ etau) mb mb'>> /\
          <<ES: esimstep s' sync evt mb' mbt ea' eb'>> /\
          <<CNF: configure w g s' sync ea' eb' csa' csb' cfa cfb ma' mbt>> /\
          <<EXT: w.(ext) s' s>>)
| mm_ok_esim_ret s'
    mb' (STP: rtc (step _ etau) mb mb')
    va ea' (RETa: w.(cquerya) s' (ret va kia) ea')
    vb eb' (RETb: w.(cqueryb) s' (ret vb kib) eb')
    (RV: vclos w s' (g s') va vb)
    csa' csb' (CNF: configure w g s' true ea' eb' csa' csb' cfa cfb ma mb')
    (EXT: w.(ext) s' s) (EXTP: w.(extpub) s' si)
.

Lemma mm_esim_ok: forall w ha hb cfa cfb ea eb enva envb r kia kib g si s
    (NH: ~ exp_halted eb)
    (ST: forall evt hb' envb' eb'
                (STb: reduce_ok evt hb envb eb hb' envb' eb'),
         exists ha' enva' ea',
           reduce_ok evt ha enva ea ha' enva' ea')
    (SIM: forall evt hb' envb' eb'
                 (STb: reduce_ok evt hb envb eb hb' envb' eb'),
            mm_ok_esim_main w r kia kib g si s cfa (ha, (enva, ea)) cfb (hb, (envb, eb))),
     esim_main r kia kib g si s cfa (ha, (enva, ea)) cfb (hb, (envb, eb)).
Proof.
  i.
  destruct (LEM (forall evt hb' envb' eb',
                   ~ reduce_ok evt hb envb eb hb' envb' eb')).
  { econs; [econs; eapply IRerr|]; eauto. }
  do 4 (apply not_all_ex_not in f; des); apply NNPP in f.

  destruct (LEM (forall evt ha' enva' ea',
                   ~ reduce_ok evt ha enva ea ha' enva' ea')).
  - econs.
    + econs; eapply IRerr; [by eauto|].
      i; hexploit ST; eauto; i; des; eauto.
    + rr; i; rrdes FH; clarify; rrdes PV; vauto.
  - hexploit SIM; eauto; []; i; des.
    esplits; eauto; inv X; eauto.
    eapply esim_step; [by eauto|].
    i; inv ST0; eapply EST; eauto.
Qed.



(** * Program similarity.

    TODO: this should perhaps use a full world parameter, i.e. it may
    make also sense for other global worlds.
*)




Definition good_exp e := forall n, e <> IEnat n.

Definition mm_psim (lw: world_loca mm_gw) G ra ea eb : Prop :=
  <<GEa: good_exp ea>> /\
  <<GEb: good_exp eb>> /\
  exists i', forall g s si enva envb i kia kib
    (GK: gknow (wlift lw) g)
    (EXT: extpub _ s si)
    (RV: forall x (IN: In (inl x) G), exists va vb,
           <<INa: lookup (inl (rename_itvar ra x)) enva = Some va>> /\
           <<INb: lookup (inl x) envb = Some vb>> /\
           <<SIM: vclos (wlift lw) s (g s) va vb>>)
    (RK: forall x (IN: In (inr x) G), exists ka kb,
           <<INa: lookup (inr (rename_contvar ra x)) enva = Some ka>> /\
           <<INb: lookup (inr x) envb = Some kb>> /\
           <<SIM: ksim (wlift lw) (wfo_nat i) kia kib g si s s ka kb>>),
  esim (wlift lw) (wfo_nat (i' + i)) kia kib g si s true
        (pmemp, Just (enva, ea))
        (pmemp, Just (envb, eb)).


Definition mm_asim (lw: world_loca mm_gw) (G: list var) ra ea eb : Prop :=
  forall g s enva envb
    (GK: gknow (wlift lw) g)
    (RV: forall x (IN: In (inl x) G), exists va vb,
           <<INa: lookup (inl (rename_itvar ra x)) enva = Some va>> /\
           <<INb: lookup (inl x) envb = Some vb>> /\
           <<SIM: vclos (wlift lw) s (g s) va vb>>)
    vb (EV: eval envb eb = Some vb),
  exists va, eval enva ea = Some va /\
             vclos (wlift lw) s (g s) va vb.


Definition mm_ksim (lw: world_loca mm_gw) (G: list var) ra ea eb : Prop :=
  forall g s si i j kia kib enva envb bxb eb'
      (GK: gknow (wlift lw) g)
      (EXT: extpub _ s si)
      (RV: forall x (IN: In (inl x) G), exists va vb,
             <<INa: lookup (inl (rename_itvar ra x)) enva = Some va>> /\
             <<INb: lookup (inl x) envb = Some vb>> /\
             <<SIM: vclos (wlift lw) s (g s) va vb>>)
      (RK: forall x (IN: In (inr x) G), exists ka kb,
             <<INa: lookup (inr (rename_contvar ra x)) enva = Some ka>> /\
             <<INb: lookup (inr x) envb = Some kb>> /\
             <<SIM: ksim (wlift lw) (wfo_nat i) kia kib g si s s ka kb>>)
      (Eb: eb = IEcont bxb eb'),
    exists bxa ea',
      <<Ea: ea = IEcont bxa ea'>> /\
      <<SIM: ksim (wlift lw) j kia kib g si s s (IVclo enva ea) (IVclo envb eb)>>.

                  
Definition mm_psim' (lw: world_loca mm_gw) G ea eb : Prop :=
  <<GEa: good_exp ea>> /\
  <<GEb: good_exp eb>> /\
  exists i', forall g s si enva envb i kia kib
    (GK: gknow (wlift lw) g)
    (EXT: extpub _ s si)
    (RV: forall x (IN: In (inl x) G), exists va vb,
           <<INa: lookup (inl x) enva = Some va>> /\
           <<INb: lookup (inl x) envb = Some vb>> /\
           <<SIM: vclos (wlift lw) s (g s) va vb>>)
    (RK: forall x (IN: In (inr x) G), exists ka kb,
           <<INa: lookup (inr x) enva = Some ka>> /\
           <<INb: lookup (inr x) envb = Some kb>> /\
           <<SIM: ksim (wlift lw) (wfo_nat i) kia kib g si s s ka kb>>),
  esim (wlift lw) (wfo_nat (i' + i)) kia kib g si s true
        (pmemp, Just (enva, ea))
        (pmemp, Just (envb, eb)).


Definition mm_asim' (lw: world_loca mm_gw) (G: list var) ea eb : Prop :=
  forall g s enva envb
    (GK: gknow (wlift lw) g)
    (RV: forall x (IN: In (inl x) G), exists va vb,
           <<INa: lookup (inl x) enva = Some va>> /\
           <<INb: lookup (inl x) envb = Some vb>> /\
           <<SIM: vclos (wlift lw) s (g s) va vb>>)
    vb (EV: eval envb eb = Some vb),
  exists va, eval enva ea = Some va /\
             vclos (wlift lw) s (g s) va vb.


Definition mm_ksim' (lw: world_loca mm_gw) (G: list var) ea eb : Prop :=
  forall g s si i j kia kib enva envb bx eb'
      (GK: gknow (wlift lw) g)
      (EXT: extpub _ s si)
      (RV: forall x (IN: In (inl x) G), exists va vb,
             <<INa: lookup (inl x) enva = Some va>> /\
             <<INb: lookup (inl x) envb = Some vb>> /\
             <<SIM: vclos (wlift lw) s (g s) va vb>>)
      (RK: forall x (IN: In (inr x) G), exists ka kb,
             <<INa: lookup (inr x) enva = Some ka>> /\
             <<INb: lookup (inr x) envb = Some kb>> /\
             <<SIM: ksim_ (esim (wlift lw)) (wfo_nat i) kia kib g si s s ka kb>>)
      (Eb: eb = IEcont bx eb'),
    exists bxa ea',
      <<Ea: ea = IEcont bxa ea'>> /\
      <<SIM: ksim (wlift lw) j kia kib g si s s (IVclo enva ea) (IVclo envb eb)>>.

                  

Lemma mm_ksim_old lw G ea eb:
  mm_ksim lw G ren_id ea eb <-> mm_ksim' lw G ea eb.
Proof.
  split; ii; eapply H; eauto; ii; eapply RV in IN; des; esplits; eauto; revert INa;
  by rewrite -/(rename_var _ (inl _)) ren_var_id.
Qed.

Lemma mm_asim_old lw G ea eb:
  mm_asim lw G ren_id ea eb <-> mm_asim' lw G ea eb.
Proof.
  split; ii; eapply H; eauto; ii; eapply RV in IN; des; esplits; eauto; revert INa;
  by rewrite -/(rename_var _ (inl _)) ren_var_id.
Qed.

Lemma mm_psim_old lw G ea eb:
  mm_psim lw G ren_id ea eb <-> mm_psim' lw G ea eb.
Proof.
  split; ii; rrdes H; rr; splits; eauto; eexists; ii; eapply H1; eauto; ii;
  eapply RV in IN; des; esplits; eauto; revert IN0 IN; 
  by rewrite -/(rename_var _ (inl _)) ren_var_id.
Qed.

Hint Resolve mm_ksim_old mm_asim_old mm_psim_old.

                  


Lemma foo': forall (X : Type) (M : EqDec X) (Y Z : Type) 
         (f : Z -> X) (g : Z -> Z) (xs : list X) (xysa xysb : list (X * Y))
         (R : Y -> Y -> Prop) (xa' xb': X) (ya' yb' : Y)
         x (IN: In (f x) (xb' :: xs))
         (RR: f x = xb' -> f (g x) = xa' /\ R ya' yb')
         (KA: f (g x) = xa' -> f x = xb')
         (RST: forall (x : Z) (IN: In (f x) xs) (NEb: f x <> xb') (NEa: f (g x) <> xa'),
                 exists ya yb : Y,
                 lookup (f (g x)) xysa = Some ya /\
                 lookup (f x) xysb = Some yb /\ R ya yb),
  exists ya yb : Y,
    lookup (f (g x)) ((xa', ya') :: xysa) = Some ya /\
    lookup (f x) ((xb', yb') :: xysb) = Some yb /\ R ya yb.
Proof.
  i; apply in_consD in IN; des.
  - subst; rewrite !list_assoc_cons_eq.
    specialize (RR eq_refl); des; rewrite RR. eauto.
  - rewrite !list_assoc_cons_neq; eauto.
Qed.

Definition extendable_t (G: list var) ren xa xb :=
  forall xb'
      (IN: In (inl (inr xb')) G)
      (E: rename_termvar ren xb' = inr xa),
    xb' = xb.
    

Definition extendable_c (G: list var) ren xa xb :=
  forall xb'
      (IN: In (inr xb') G)
      (E: rename_contvar ren xb' = xa),
    xb' = xb.
    

Lemma bla: forall G ra bxa bxb bya byb
    (EX: extendable_t G ra bxa bxb)
    (EE: bya = bxa -> byb = bxb),
  extendable_t (inl (inr byb) :: G) (ra)[inr bya/byb] bxa bxb.
Proof.
  ii; destruct (LEM (xb' = byb)).
  - subst; s in E; tmext_simplH E; clarify; eauto.
  - s in E; rewrite tmext_neq in E; eauto.
    exploit EX; eauto; []; i; des; subst; eauto.
    apply in_consD in IN; des; vauto.
Qed.


Lemma mm_compat_input: forall lw G ra bxa bxb ea eb
    (EX: extendable_t G ra bxa bxb)
    (IH: mm_psim lw (inl (inr bxb) :: G) ra[inr bxa/bxb] ea eb),
  mm_psim lw G ra (IEinput bxa ea) (IEinput bxb eb).
Proof.
  i; rr; splits; vauto; []; exists 0; i; s.
  rrdes IH; pfold; econs; i.
  mm_destruct XTRa XTRb CNF.
  apply mm_esim_ok; [by eauto| |].
  - i; invc STb; eauto.
  - esplits; eauto; eapply mm_ok_esim_step; eauto; i.
    destruct ma' as [? []]; s in STa; invc STa.
    do 7 eexists; exists false; splits; first (by eauto)
    ; last (by s; eauto); first 1 last.
    + econs; eauto; noemp; inst_pairs; s; noemp; rr; eauto 20.
    + left; econs; eauto.
      * left; rr; eauto.
      * { eapply gres_incl_id; left.
          rewrite 2!mid_csplit.
          eapply pesim_release; eauto. 
          eapply IH1; [by eauto| by s; eauto| |]; eauto. 
          - i; eapply foo'; eauto.
            { i; clarify; s; tmext_simpl; split; eauto.
              econs; eapply vclgen_nat; s; eauto.
            }
            { i; clarify.
              destruct (LEM (x2 = inr bxb)); vauto.
              rewrite ren_ext_it_neq in H; [|by ii; vauto].
              destruct x2; vauto.
              exploit (EX t); s in H; i; des; vauto.
              apply in_consD in IN; des; vauto.
            } 
            clear IN; i; rewrite ren_ext_it_neq; [by eauto|by ii; vauto].
          - i; apply in_consD in IN; des; clarify.
            rewrite !list_assoc_cons_neq; auto; ii; clarify.
            rewrite ren_ext_c_neq; auto.
          - s; eauto.
          - s; eauto.
        }
Qed.


Lemma extendable_t_same: forall G ren x
    (E: forall y (IN: In y G) (E: rename_var ren y = inl (inr x)), rename_var ren y = y),
  extendable_t G ren x x.
Proof.
  s; ii; clarify; hexploit (E (inl (inr xb'))); [|s; rewrite E0|]; eauto.
  s; i; clarify; rewrite H in E0; clarify.
Qed.


Lemma mm_compat_input': forall lw G bx ea eb
    (IH: mm_psim' lw (inl (inr bx) :: G) ea eb),
  mm_psim' lw G (IEinput bx ea) (IEinput bx eb).
Proof.
  i; eapply mm_psim_old; eapply mm_compat_input; eauto
  ; try by eapply extendable_t_same; i; rewrite ren_var_id; eauto.
  rewrite ?ren_id_self_t ?ren_id_self_c; auto. by eapply mm_psim_old.
Qed.


Lemma mm_compat_output: forall lw G ra xa xb ea eb
    (WF: In (inl xb) G)
    (E: xa = rename_itvar ra xb)
    (IH: mm_psim lw G ra ea eb),
  mm_psim lw G ra (IEoutput xa ea) (IEoutput xb eb).
Proof.
  i; rr; splits; vauto; []; exists 0; i; s.
  apply RV in WF.
  rrdes IH; pfold; econs; i.
  mm_destruct XTRa XTRb CNF.
  eapply mm_esim_ok; [by eauto| |].
  - i; invc STb. rewrite WF0 in GET; clarify.
    eapply mm_vclosDb in WF1; subst; eauto; des; clarify; eauto.
  - esplits; eauto; eapply mm_ok_esim_step; eauto; i.
    destruct ma' as [? []]; s in STa; invc STa.
    do 7 eexists; exists false; splits; first (by eauto)
    ; last (by s; eauto); first 1 last.
    + econs; eauto; noemp; inst_pairs; s; noemp; rr; eauto 10.
    + left; econs; eauto.
      * invc STb; rewrite WF0 in GET0; rewrite WF in GET; clarify.
        left; eapply mm_vclosDb in WF1; eauto; des; clarify.
        eapply IRok; eauto.
      * eapply gres_incl_id; left.
        rewrite 2!mid_csplit.
        eapply pesim_release; s; eauto; eapply IH1; eauto.
Qed.

Lemma mm_compat_output': forall lw G x ea eb
    (WF: In (inl x) G)
    (IH: mm_psim' lw G ea eb),
  mm_psim' lw G (IEoutput x ea) (IEoutput x eb).
Proof.
  i; eapply mm_psim_old; eapply mm_compat_output; eauto
  ; try by rewrite ren_itvar_id; eauto.
  rewrite ?ren_itvar_id ?ren_id_self_t ?ren_id_self_c; auto; by eapply mm_psim_old.
Qed.


Lemma mm_compat_app: forall lw G ra xfa xfb xva xvb xka xkb
    (INf: In (inl xfb) G)
    (INv: In (inl xvb) G)
    (INk: In (inr xkb) G)
    (Ef: xfa = rename_itvar ra xfb)
    (Ev: xva = rename_itvar ra xvb)
    (Ek: xka = rename_contvar ra xkb),
  mm_psim lw G ra (IEapp xfa xva xka) (IEapp xfb xvb xkb).
Proof.
  i; rr; splits; vauto; []; exists 0; i; s.
  apply RV in INf; apply RV in INv; apply RK in INk; des.
  pfold; econs; i.
  mm_destruct XTRa XTRb CNF.
  eapply mm_esim_ok; [by eauto| |].
  - i; invc STb.
    rewrite INf0 in GET; rewrite INv0 in GETV; rewrite INk0 in GETK; clarify.
    eapply mm_vclosDb in INf1; des; clarify.
    esplits; eauto using IRapp.
  - esplits; eauto; eapply mm_ok_esim_step; eauto; i.
    destruct ma' as [? []]; s in STa; invc STa STb.
    do 7 eexists; exists true; esplits; first (by eauto); first 1 last; eauto.
    + econs; first (by eauto); first 1 last.
      * inst_pairs; s; noemp; rr; instantiate (2 := pmemp); s; noemp; eauto 10.
      * inst_pairs; s; noemp; rr; instantiate (5 := pmemp); s; noemp; eauto 10.
      * clear CO CO0; s; eauto 10.
    + rewrite INv0 in GETV0; rewrite INk0 in GETK0; rewrite INf0 in GET0; clarify.
      rewrite INv in GETV; rewrite INk in GETK; rewrite INf in GET; clarify.
      right; econs; eauto.
      * econs. 
        hexploit (IRapp ((hb' ## x3) ## x0) envb); [|exact INv0|exact INk0|]; eauto.
      * { econs 1.
          - rr; eauto.
          - rr; eauto.
          - apply mm_vclosDb in INf1; des.
            econs 1; rr; eauto.
          - eauto.
          - eapply ksim_mon; eauto using gres_incl_id.
        }
Qed.

Lemma mm_compat_app': forall lw G xf xv xk
    (INf: In (inl xf) G)
    (INv: In (inl xv) G)
    (INk: In (inr xk) G),
  mm_psim' lw G (IEapp xf xv xk) (IEapp xf xv xk).
Proof.
  i; eapply mm_psim_old; eapply mm_compat_app; eauto
  ; try by eapply extendable_t_same; i; rewrite ?ren_it_id; eauto.
  - rewrite ?ren_itvar_id ?ren_id_self_t ?ren_id_self_c; auto.
  - rewrite ?ren_itvar_id ?ren_id_self_t ?ren_id_self_c; auto.
Qed.



Lemma mm_compat_read: forall lw G ra xa xb bxa bxb ea eb
    (WF: In (inl xb) G)
    (E: xa = rename_itvar ra xb)
    (EX: extendable_t G ra bxa bxb)
    (IH: mm_psim lw (inl (inr bxb) :: G) ra[inr bxa/bxb] ea eb),
  mm_psim lw G ra (IEread xa bxa ea) (IEread xb bxb eb).
Proof.
  i; rr; splits; vauto; []; exists 0; i; s.
  apply RV in WF; des.
  rrdes IH; pfold; econs; i.
  mm_destruct XTRa XTRb CNF.
  eapply mm_esim_ok; [by eauto| |].
  - i; invc STb.
    rewrite WF0 in GET; clarify.
    eapply mm_vclosDb in WF1; des; clarify.
    rrdes QR; clarify.
    dup PVa; apply pmvalid_plusD in PVa; des.
    dup PVb; apply pmvalid_plusD in PVb; des.
simpl cplus in RH0.
    eapply mm_confrel_read in RH0; eauto; []; des.
    esplits; econs; eauto using pmread_plus.
  - esplits; eauto; eapply mm_ok_esim_step; eauto; i.
    destruct ma' as [? []]; s in STa; invc STa STb.
    rewrite WF0 in GET0; rewrite WF in GET; clarify.
    dup WF1; eapply mm_vclosDb in WF1; []; des; clarify.
    rrdes QR; clarify.
    dup PVa; apply pmvalid_plusD in PVa; des.
    dup PVb; apply pmvalid_plusD in PVb; des.
simpl cplus in RH0.
    dup RH0; eapply mm_confrel_read in RH0; eauto; []; des.
    eapply pmread_plus in RDa; eauto.
    eapply pmread_plus in RDb; eauto.
    hexploit (pmread_uniq RDa LOAD); []; i; inst; subst.
    hexploit (pmread_uniq RDb LOAD0); []; i; inst; subst.
    do 7 eexists; exists false; splits; first (by eauto); last (by s; eauto)
    ; last (econs; [by eauto|by eauto| |]); last first; first 1 last.
    + inst_pairs; s; noemp; rr; esplits; eauto.
    + left; econs; eauto.
      * left; inst_pairs; eapply IRok; econs; eauto.
      * eapply gres_incl_id; left.
        rewrite mid_csplit; instantiate (1:=_#_).
        clear CO0 CO.
        eapply pesim_release; eauto; try (by rr; eauto).
        { hexploit IH1; [M| | | |M]; (Mdo by eauto).
        - eapply extpub_trans; [|by eauto]; s; eauto.
        - i; eapply foo'; eauto.
          { i; clarify; s; tmext_simpl; eauto. }
          { i; clarify.
            destruct (LEM (x2 = inr bxb)); vauto.
            rewrite ren_ext_it_neq in H; [|by ii; vauto].
            destruct x2; vauto; exploit (EX t); s in H; i; vauto.
            apply in_consD in IN; des; vauto.
          } 
          clear IN; i; rewrite ren_ext_it_neq; [by eauto|by ii; vauto].
        - i; apply in_consD in IN; des; clarify.
          rewrite !list_assoc_cons_neq; auto; ii; clarify.
          rewrite ren_ext_c_neq; auto.
        }
    + s; noemp; rr; esplits; eauto 10.
Qed.

Lemma mm_compat_read': forall lw G x bx ea eb
    (WF: In (inl x) G)
    (IH: mm_psim' lw (inl (inr bx) :: G) ea eb),
  mm_psim' lw G (IEread x bx ea) (IEread x bx eb).
Proof.
  i; eapply mm_psim_old; eapply mm_compat_read; eauto
  ; try by eapply extendable_t_same; i; rewrite ?ren_var_id; eauto.
  - rewrite ?ren_itvar_id ?ren_id_self_t ?ren_id_self_c; auto.
  - rewrite ?ren_itvar_id ?ren_id_self_t ?ren_id_self_c; auto. by eapply mm_psim_old.
Qed.


Lemma mm_compat_write: forall lw G ra xa xb ya yb ea eb
    (WFx: In (inl xb) G)
    (WFy: In (inl yb) G)
    (Ex: xa = rename_itvar ra xb)
    (Ey: ya = rename_itvar ra yb)
    (IH: mm_psim lw G ra ea eb),
  mm_psim lw G ra (IEwrite xa ya ea) (IEwrite xb yb eb).
Proof.
  i; rr; splits; vauto; []; exists 0; i; s.
  apply RV in WFx; apply RV in WFy; des.
  rrdes IH; pfold; econs; i; invc CNF XTRa XTRb.
  dup RH; rrdes RH.
  hexploit mm_confrel_shape; eauto; []; i; des.
  rrdes WHX; dup WHY0; rrdes WHY0; subst.
  hexploit lwconf_extra; [exact WHY|]; intros [XTRa XTRb].
  inversion XTRa; inversion XTRb; clear XTRa XTRb; subst.
  rrdes MA MB; subst; s in Ec; s in Ec0; clarify; noemp.
  eapply mm_esim_ok; [by eauto| |].
  - i; invc STb.
    rewrite WFy0 in GET0; rewrite WFx0 in GET; clarify.
    eapply mm_vclosDb in WFx1; des; clarify.
    dup PV; apply pmvalid_plusD in PV; des.
    dup PV0; apply pmvalid_plusD in PV0; des.
    simpl cplus in RH0; rrdes QR; clarify.
    eapply mm_confrel_read in REG; eauto; des.
    eapply pmread_plus in RDa; eauto; rrdes RDa; rewrite RD.
    esplits; econs; eauto.
  - esplits; eauto; eapply mm_ok_esim_step; eauto; i.
    destruct ma' as [? []]; s in STa; invc STa STb.
    rewrite WFx0 in GET1; rewrite WFy0 in GET2; clarify.
    rewrite WFy in GET0; rewrite WFx in GET; clarify.
    dup WFx1; eapply mm_vclosDb in WFx1; eauto; []; des; clarify.
    dup PV; apply pmvalid_plusD in PV; des.
    dup PV0; apply pmvalid_plusD in PV0; des.
    simpl cplus in RH0; rrdes QR; clarify.
    dup REG; eapply mm_confrel_read in REG; eauto; des.
    dup RDa; rrdes RDa; subst; rewrite RD in H1, PV1.
    rewrite -pmplus_assoc in H1, PV1; rewrite -H1 in PV1.
    apply pmcell_uniq in H1; [|by auto]; des; subst.
    dup RDb; rrdes RDb; subst; rewrite RD0 in H2, PV3.
    rewrite -pmplus_assoc in H2, PV3; rewrite -H2 in PV3.
    apply pmcell_uniq in H2; [|by auto]; des; subst.
    rrdes WHX0 CPa CPb; clarify; noemp; clarify.
    do 7 eexists; exists false; splits; first (by eauto); last (by s; eauto)
    ; last (econs; [by eauto|by eauto| |]); last first; first 1 last.
    + rewrite pmplus_assoc; inst_pairs; s; noemp; rr; esplits; eauto.
      * rewrite <-pmplus_assoc; eauto using pmvalid_plus_cell.
      * rewrite RD -pmplus_assoc in FIN.
        rewrite <-pmplus_assoc; eauto using pmfinite_plus_cell.
    + left; econs; eauto.
      * left; inst_pairs; eapply IRok; econs; eauto.
      * eapply gres_incl_id; left.
        rewrite mid_csplit; instantiate (1:=_#_).
        clear CO CO0.
        eapply pesim_release; eauto; try (by rr; eauto).
        { hexploit IH1; eauto using extpub_trans. }
        { instantiate (1:= ({lb |-> v0} ## h'0, Nothing)).
          rewrite RD RD0 in RH0, VLD, VLD0.
          eapply mm_confrel_write; eauto.
        }
    + s; noemp.
      rr; rewrite <-pmplus_assoc; esplits; eauto using pmvalid_plus_cell.
      rewrite RD0 -pmplus_assoc in FIN0; eauto using pmfinite_plus_cell.
Qed.


Lemma mm_compat_write': forall lw G x y ea eb
    (WFx: In (inl x) G)
    (WFy: In (inl y) G)
    (IH: mm_psim' lw G ea eb),
  mm_psim' lw G (IEwrite x y ea) (IEwrite x y eb).
Proof.
  i; eapply mm_psim_old; eapply mm_compat_write; eauto
  ; try by eapply extendable_t_same; i; rewrite ?ren_var_id; eauto.
  - rewrite ?ren_itvar_id ?ren_id_self_t ?ren_id_self_c; auto.
  - rewrite ?ren_itvar_id ?ren_id_self_t ?ren_id_self_c; auto.
  - by eapply mm_psim_old.
Qed.



Lemma mm_compat_new: forall lw G ra xa xb bxa bxb ea eb
    (WF: In (inl xb) G)
    (E: xa = rename_itvar ra xb)
    (EX: extendable_t G ra bxa bxb)
    (IH: mm_psim lw (inl (inr bxb) :: G) ra[inr bxa/bxb] ea eb)
    (STA: stable lw),
  mm_psim lw G ra (IEnew xa bxa ea) (IEnew xb bxb eb).
Proof.
  i; rr; splits; vauto; []; exists 0; i; s.
  apply RV in WF; des.
  rrdes IH; pfold; econs; i.
  mm_destruct XTRa XTRb CNF.
  eapply mm_esim_ok; [by eauto| |].
  - i; invc STb.
    rewrite WF0 in GET; clarify.
    hexploit (@pmfinite_alloc _ _ _ ((ha' ## x1) ## x)); eauto. 
    i; des; esplits; econs; eauto.
  - esplits; eauto; eapply mm_ok_esim_step; eauto; i.
    destruct ma' as [? []]; s in STa; invc STa.
    hexploit (@pmfinite_alloc _ _ _ ((hb' ## x3) ## x0)); eauto.
    i; des; destruct s as [[[]]].
    dup H; rewrite pmplus_assoc in H0; apply pmvalid_plus_l in H0.
    dup PV; rewrite pmplus_assoc in PV0; apply pmvalid_plus_l in PV0.
    simpl cplus in RH0.
    rewrite WF in GET; clarify.
    hexploit confrel_alloc.
    { eauto. }
    { eapply vclos_mon; eauto. }
    { eapply confrel_mon; eauto. }
    { eauto. }
    { eauto. }
    { split_pmfinite FINa PVa; eauto. }
    { split_pmfinite FINb PVb; eauto. }
    { instantiate (1 := refst_allocP s1 loc x2 (confrel_newa RH0 PV0) (confrel_newb RH0 H0)).
      eauto.
    }
    { eauto. }
    i; des.
    do 6 econs
    ; exists (s, s0, refst_allocP s1 loc x2 (confrel_newa RH0 PV0) (confrel_newb RH0 H0), s2')
    ; exists false.
    splits; first (by eauto); last (by s; unfold refst_alloc; eauto)
    ; last (econs; [by eauto|by eauto| |]); last first; first 1 last.
    + inst_pairs; s; noemp; rewrite pmplus_assoc.
      rr; esplits; eauto; inst; rewrite <-pmplus_assoc; eauto.
    + left; econs; auto; [left; inst_pairs; econs 1|].
      eapply IRnew; eauto; econs; eauto.
      eapply gres_incl_id; left.
      rewrite mid_csplit; instantiate (1:=_#_).
      clear CO CO0.
      eapply pesim_release; eauto; try (by rr; eauto).
      { hexploit IH1; [M| | | |M]; (Mdo by eauto).
        - eapply extpub_trans; [|by eauto]; s; unfold refst_alloc; eauto.
        - i; eapply foo'; eauto.
          { i; clarify; s; tmext_simpl; split; eauto.
            econs; eapply vclgen_ref.
            rr; esplits; s; unfold refst_alloc; eauto.
          }
          { i; clarify.
            destruct (LEM (x4 = inr bxb)); vauto.
            rewrite ren_ext_it_neq in H3; [|by ii; vauto].
            destruct x4; vauto.
            exploit (EX t); s in H; i; des; vauto.
            apply in_consD in IN; des; vauto.
          } 
          clear IN; i; rewrite ren_ext_it_neq; [|by ii; vauto].
          apply RV in IN; des; esplits; eauto.
          eapply vclos_mon; s; unfold refst_alloc; eauto.
          i; eapply gk_mon; s; unfold refst_alloc; eauto.
        - i; apply in_consD in IN; des; clarify.
          rewrite ren_ext_c_neq !list_assoc_cons_neq; try by auto.
          apply RK in IN; des; esplits; eauto;
          eapply ksim_mon; s; unfold refst_alloc; eauto.
      }
    + s; unfold mid_val; noemp.
      rr; rewrite <-pmplus_assoc; esplits; eauto.
Qed.


Lemma mm_compat_new': forall lw G x bx ea eb
    (WF: In (inl x) G)
    (IH: mm_psim' lw (inl (inr bx) :: G) ea eb)
    (STA: stable lw),
  mm_psim' lw G (IEnew x bx ea) (IEnew x bx eb).
Proof.
  i; eapply mm_psim_old; eapply mm_compat_new; eauto
  ; try by eapply extendable_t_same; i; rewrite ?ren_var_id; eauto.
  - rewrite ?ren_itvar_id ?ren_id_self_t ?ren_id_self_c; auto.
  - rewrite ?ren_itvar_id ?ren_id_self_t ?ren_id_self_c; auto. by eapply mm_psim_old.
Qed.






Lemma mm_compat_ret: forall lw G ra xka xkb xva xvb
    (INk: In (inr xkb) G)
    (INv: In (inl xvb) G)
    (Ev: xva = rename_itvar ra xvb)
    (Ek: xka = rename_contvar ra xkb),
  mm_psim lw G ra (IEret xka xva) (IEret xkb xvb).
Proof.
  i; rr; splits; vauto; []; exists 0; i; s.
  apply RK in INk; apply RV in INv; des.
  inv INk1; eapply KR; s; eauto.
Qed.



Lemma mm_compat_ret': forall lw G xk x
    (INk: In (inr xk) G)
    (INv: In (inl x) G),
  mm_psim' lw G (IEret xk x) (IEret xk x).
Proof.
  i; eapply mm_psim_old; eapply mm_compat_ret; eauto
  ; try by eapply extendable_t_same; i; rewrite ?ren_var_id; eauto.
  rewrite ?ren_itvar_id ?ren_id_self_t ?ren_id_self_c; auto.
Qed.




Lemma mm_compat_ifnz: forall lw G ra xa xb ea ea' eb eb'
    (WF: In (inl xb) G)
    (E: xa = rename_itvar ra xb)
    (IH: mm_psim lw G ra ea eb)
    (IH': mm_psim lw G ra ea' eb'),
  mm_psim lw G ra (IEifnz xa ea ea') (IEifnz xb eb eb').
Proof.
  i; rr; splits; vauto; []; exists 0; i; s.
  apply RV in WF; des.
  rrdes IH IH'; pfold; econs; i.
  mm_destruct XTRa XTRb CNF.
  eapply mm_esim_ok; [by eauto| |].
  - i; inv STb
    ; rewrite WF0 in GET; clarify
    ; eapply mm_vclosDb in WF1; eauto; des; clarify; eauto.
  - esplits; eauto; eapply mm_ok_esim_step; eauto; i.
    destruct ma' as [? []]; s in STa; invc STb.
    { rewrite WF0 in GET; clarify.
      apply mm_vclosDb in WF1; []; subst.
      invc STa; [|by rewrite WF in GET; clarify].
      do 7 eexists; exists false; splits; first (by eauto); last (by s; eauto)
      ; first 1 last.
      + econs; eauto; noemp; inst_pairs; s; noemp; rr; eauto 10.
      + left; econs; eauto.
        * left; eapply IRok; eauto.
        * eapply gres_incl_id; left.
          rewrite 2!mid_csplit.
          eapply pesim_release; eauto; try (by rr; eauto).
          eapply IH'1; eauto.
    }
    { rewrite WF0 in GET; clarify.
      apply mm_vclosDb in WF1; []; subst.
      invc STa; [by rewrite WF in GET; clarify|].
      do 7 eexists; exists false; splits; first (by eauto); last (by s; eauto)
      ; first 1 last.
      + econs; eauto; noemp; inst_pairs; s; noemp; rr; eauto 10.
      + left; econs; eauto.
        * left; eapply IRok; eauto.
        * eapply gres_incl_id; left.
          rewrite 2!mid_csplit.
          eapply pesim_release; eauto; try (by rr; eauto).
          eapply IH1; eauto.
    }
Qed.


Lemma mm_compat_ifnz': forall lw G  x ea ea' eb eb'
    (WF: In (inl x) G)
    (IH: mm_psim' lw G ea eb)
    (IH': mm_psim' lw G ea' eb'),
  mm_psim' lw G (IEifnz x ea ea') (IEifnz x eb eb').
Proof.
  i; eapply mm_psim_old; eapply mm_compat_ifnz; eauto
  ; try by eapply extendable_t_same; i; rewrite ?ren_var_id.
  rewrite ?ren_itvar_id ?ren_id_self_t ?ren_id_self_c; auto.
  by eapply mm_psim_old.
  by eapply mm_psim_old.
Qed.


Lemma mm_compat_case: forall lw G ra xa xb bxa bxb ea ea' eb eb'
    (WF: In (inl xb) G)
    (E: xa = rename_itvar ra xb)
    (EX: extendable_t G ra bxa bxb)
    (IH: mm_psim lw (inl (inr bxb) :: G) ra[inr bxa/bxb] ea eb)
    (IH': mm_psim lw (inl (inr bxb) :: G) ra[inr bxa/bxb] ea' eb'),
  mm_psim lw G ra (IEcase xa bxa ea ea') (IEcase xb bxb eb eb').
Proof.
  i; rr; splits; vauto; []; exists 0; i; s.
  apply RV in WF; des.
  rrdes IH IH'; pfold; econs; i.
  mm_destruct XTRa XTRb CNF.
  eapply mm_esim_ok; [by eauto| |].
  - i; inv STb
    ; rewrite WF0 in GET; clarify
    ; eapply mm_vclosDb in WF1; eauto; des; clarify
    ; eauto using IRcasel, IRcaser.
  - esplits; eauto; eapply mm_ok_esim_step; eauto; i.
    destruct ma' as [? []]; s in STa; invc STb.
    { rewrite WF0 in GET; clarify.
      apply mm_vclosDb in WF1; []; des; subst.
      invc STa; [|by rewrite WF in GET; clarify].
      do 7 eexists; exists false; splits; first (by eauto); last (by s; eauto)
      ; first 1 last.
      + econs; eauto; noemp; inst_pairs; s; noemp; rr; eauto 10.
      + left; econs; eauto.
        * left; eapply IRok; eauto using IRcasel.
        * eapply gres_incl_id; left.
          rewrite 2!mid_csplit.
          eapply pesim_release; eauto; try (by rr; eauto).
          rewrite WF in GET; clarify.
          eapply IH1; eauto; [|by i; eapply foo'; eauto; vauto].
          i; eapply foo'; eauto; vauto.
          { i; clarify; s; tmext_simpl; eauto. }
          { i; clarify.
            destruct (LEM (x2 = inr bxb)); vauto.
            rewrite ren_ext_it_neq in H; [|by ii; vauto].
            destruct x2; vauto; exploit (EX t); s in H; i; vauto.
            apply in_consD in IN; des; vauto.
          }
          clear IN; i; rewrite ren_ext_it_neq; [eauto|ii; vauto].
    }
    { rewrite WF0 in GET; clarify.
      apply mm_vclosDb in WF1; []; des; subst.
      invc STa; [by rewrite WF in GET; clarify|].
      do 7 eexists; exists false; splits; first (by eauto); last (by s; eauto)
      ; first 1 last.
      + econs; eauto; noemp; inst_pairs; s; noemp; rr; eauto 10.
      + left; econs; eauto.
        * left; eapply IRok; eauto using IRcaser.
        * eapply gres_incl_id; left.
          rewrite 2!mid_csplit.
          eapply pesim_release; eauto; try (by rr; eauto).
          rewrite WF in GET; clarify.
          eapply IH'1; eauto; [|by i; eapply foo'; eauto; vauto].
          i; eapply foo'; eauto; vauto. 
          { i; clarify; s; tmext_simpl; eauto. }
          { i; clarify.
            destruct (LEM (x2 = inr bxb)); vauto.
            rewrite ren_ext_it_neq in H; [|by ii; vauto].
            destruct x2; vauto; exploit (EX t); s in H; i; vauto.
            apply in_consD in IN; des; vauto.
          }
          clear IN; i; rewrite ren_ext_it_neq; [eauto|ii; vauto].
    }
Qed.


Lemma mm_compat_case': forall lw G x bx ea ea' eb eb'
    (WF: In (inl x) G)
    (IH: mm_psim' lw (inl (inr bx) :: G) ea eb)
    (IH': mm_psim' lw (inl (inr bx) :: G) ea' eb'),
  mm_psim' lw G (IEcase x bx ea ea') (IEcase x bx eb eb').
Proof.
  i; eapply mm_psim_old; eapply mm_compat_case; eauto
  ; try by eapply extendable_t_same; i; rewrite ?ren_var_id; eauto.
  - rewrite ?ren_itvar_id ?ren_id_self_t ?ren_id_self_c; auto.
  - rewrite ?ren_itvar_id ?ren_id_self_t ?ren_id_self_c; auto. by eapply mm_psim_old.
  - rewrite ?ren_itvar_id ?ren_id_self_t ?ren_id_self_c; auto. by eapply mm_psim_old.
Qed.



Lemma mm_compat_inst: forall lw G ra xfa xfb xka xkb
    (INf: In (inl xfb) G)
    (INk: In (inr xkb) G)
    (Ef: xfa = rename_itvar ra xfb)
    (Ek: xka = rename_contvar ra xkb),
  mm_psim lw G ra (IEinst xfa xka) (IEinst xfb xkb).
Proof.
  i; rr; splits; vauto; []; exists 0; i; s.
  apply RV in INf; apply RK in INk; des.
  pfold; econs; i.
  mm_destruct XTRa XTRb CNF.
  eapply mm_esim_ok; [by eauto| |].
  - i; invc STb.
    rewrite INf0 in GET; rewrite INk0 in GETK; clarify.
    eapply mm_vclosDb in INf1; des; clarify.
    esplits; eauto using IRinst.
  - esplits; eauto; eapply mm_ok_esim_step; eauto; i.
    destruct ma' as [? []]; s in STa; invc STa STb.
    rewrite INf0 in GET0; rewrite INf in GET; clarify.
    do 7 eexists; exists true; esplits; first (by eauto); first 1 last; eauto.
    + econs; [by eauto| | |]; first last.
      * inst_pairs; s; noemp; rr; instantiate (5 := pmemp); s; noemp; eauto 10.
      * inst_pairs; s; noemp; rr; instantiate (2 := pmemp); s; noemp; eauto 10.
      * split; rr; eauto.
    + rewrite INk in GETK; rewrite INk0 in GETK0; clarify.
      right; econs; eauto.
      * econs; eauto using IRinst.
      * econs 2.
        { rr; eauto. }
        { rr; eauto. }
        { eapply mm_vclosDb in INf1; des; clarify.
          econs 2; rr; eauto. }
        { eapply ksim_mon; eauto using gres_incl_id. }
Qed.

Lemma mm_compat_inst': forall lw G xf xk
    (INf: In (inl xf) G)
    (INk: In (inr xk) G),
  mm_psim' lw G (IEinst xf xk) (IEinst xf xk).
Proof.
  i; eapply mm_psim_old; eapply mm_compat_inst; eauto
  ; try by eapply extendable_t_same; i; rewrite ?ren_var_id; eauto.
  rewrite ?ren_itvar_id ?ren_id_self_t ?ren_id_self_c; auto.
Qed.




Lemma mm_compat_bop: forall lw G ra op xa xb ya yb
    (WFx: Ini xb G)
    (WFy: Ini yb G)
    (Ex: xa = match xb with inl xb' => inl (rename_itvar ra xb') | _ => xb end)
    (Ey: ya = match yb with inl yb' => inl (rename_itvar ra yb') | _ => yb end),
  mm_asim lw G ra (IEbop op xa ya) (IEbop op xb yb).
Proof.
  ii; destruct yb; [apply RV in WFy|]; (destruct xb; [apply RV in WFx|])
  ; (des; s in EV; clarify
     ; try (destruct (lookup (inl i0) envb) eqn:Ei0; rewrite Ei0 in EV; vauto
            ; destruct v; vauto; rewrite Ei0 in INb0; clarify
            ; apply mm_vclosDb in SIM0; des; clarify)
     ; try (destruct (lookup (inl i) envb) eqn:Ei; rewrite Ei in EV; vauto
            ; destruct v; vauto; rewrite Ei in INb; clarify
            ; apply mm_vclosDb in SIM; des; clarify)
     ; esplits; [by s; rewrite ?INa0 ?INa; eauto|]
     ; econs; eapply vclgen_nat; s; eauto).
Qed.


Lemma mm_compat_bop': forall lw G op x y
    (WFx: Ini x G)
    (WFy: Ini y G),
  mm_asim' lw G (IEbop op x y) (IEbop op x y).
Proof.
  i; eapply mm_asim_old; eapply mm_compat_bop; eauto
  ; try by eapply extendable_t_same; i; rewrite ?ren_var_id; eauto.
  - destruct x; rewrite ?ren_itvar_id ?ren_id_self_t ?ren_id_self_c; auto.
  - destruct y; rewrite ?ren_itvar_id ?ren_id_self_t ?ren_id_self_c; auto.
Qed.


Lemma mm_compat_refeq: forall lw G ra xa xb ya yb
    (WFx: In (inl xb) G)
    (WFy: In (inl yb) G)
    (Ex: xa = rename_itvar ra xb)
    (Ey: ya = rename_itvar ra yb),
  mm_asim lw G ra (IErefeq xa ya) (IErefeq xb yb).
Proof.
  ii.
  destruct (RV _ WFx) as [lxa [lxb [Hxa [Hxb Vxab]]]].
  destruct (RV _ WFy) as [lya [lyb [Hya [Hyb Vyab]]]].
  des.
  case_eq lya; move: (mm_vclosDb Vyab); case_eq lyb; ii; subst; des; clarify; try done
  ; rewrite /eval Hyb Hxb in EV; destruct lxb; try done.
  case_eq lxa; move: (mm_vclosDb Vxab); ii; subst; des; clarify.
  simpl; rewrite ?Hxa ?Hya. 
  eexists,,; destruct QR, QR0; des. 
  move: (refst_bij REG0 REG).
  ins; desf; vauto; try by exfalso; eauto.
Qed.

Lemma mm_compat_refeq': forall lw G x y
    (WFx: In (inl x) G)
    (WFy: In (inl y) G),
  mm_asim' lw G (IErefeq x y) (IErefeq x y).
Proof.
  i; eapply mm_asim_old; eapply mm_compat_refeq; eauto
  ; try by eapply extendable_t_same; i; rewrite ?ren_var_id; eauto.
  - rewrite ?ren_itvar_id ?ren_id_self_t ?ren_id_self_c; auto.
  - rewrite ?ren_itvar_id ?ren_id_self_t ?ren_id_self_c; auto.
Qed.



Lemma mm_compat_fst: forall lw G ra xa xb
    (WF: In (inl xb) G)
    (E: xa = rename_itvar ra xb),
  mm_asim lw G ra (IEfst xa) (IEfst xb).
Proof.
  ii; apply RV in WF; des.
  destruct (lookup (inl xb) envb) eqn:E'; s in EV; rewrite E' in EV; vauto.
  destruct v; vauto.
  rewrite INb in E'; clarify.
  apply mm_vclosDb in SIM; des; clarify.
  esplits; s; try rewrite INa; eauto.
Qed.


Lemma mm_compat_fst': forall lw G x
    (WF: In (inl x) G),
  mm_asim' lw G (IEfst x) (IEfst x).
Proof.
  i; eapply mm_asim_old; eapply mm_compat_fst; eauto
  ; try by eapply extendable_t_same; i; rewrite ?ren_var_id; eauto.
  rewrite ?ren_itvar_id ?ren_id_self_t ?ren_id_self_c; auto.
Qed.


Lemma mm_compat_snd: forall lw G ra xa xb
    (WF: In (inl xb) G)
    (E: xa = rename_itvar ra xb),
  mm_asim lw G ra (IEsnd xa) (IEsnd xb).
Proof.
  ii; apply RV in WF; des.
  destruct (lookup (inl xb) envb) eqn:E'; s in EV; rewrite E' in EV; vauto.
  destruct v; vauto.
  rewrite INb in E'; clarify.
  apply mm_vclosDb in SIM; des; clarify.
  esplits; s; try rewrite INa; eauto.
Qed.


Lemma mm_compat_snd': forall lw G x
    (WF: In (inl x) G),
  mm_asim' lw G (IEsnd x) (IEsnd x).
Proof.
  i; eapply mm_asim_old; eapply mm_compat_snd; eauto
  ; try by eapply extendable_t_same; i; rewrite ?ren_var_id; eauto.
  rewrite ?ren_itvar_id ?ren_id_self_t ?ren_id_self_c; auto.
Qed.



Lemma mm_compat_unit: forall lw G ra,
  mm_asim lw G ra IEunit IEunit.
Proof.
  ii; s in EV; clarify.
  esplits; s; eauto.
  econs; eapply vclgen_unit; s; eauto.
Qed.

Lemma mm_compat_unit': forall lw G,
  mm_asim' lw G IEunit IEunit.
Proof.
  ii; s in EV; clarify.
  esplits; s; eauto.
  econs; eapply vclgen_unit; s; eauto.
Qed.


Lemma mm_compat_nat: forall lw G n ra,
  mm_asim lw G ra (IEnat n) (IEnat n).
Proof.
  ii; s in EV; clarify.
  esplits; s; eauto.
  econs; eapply vclgen_nat; s; eauto.
Qed.

Lemma mm_compat_nat': forall lw G n,
  mm_asim' lw G (IEnat n) (IEnat n).
Proof.
  ii; s in EV; clarify.
  esplits; s; eauto.
  econs; eapply vclgen_nat; s; eauto.
Qed.


Lemma mm_compat_pair: forall lw G ra xa xb ya yb
    (WFx: In (inl xb) G)
    (WFy: In (inl yb) G)
    (Ex: xa = rename_itvar ra xb)
    (Ey: ya = rename_itvar ra yb),
  mm_asim lw G ra (IEpair xa ya) (IEpair xb yb).
Proof.
  ii; apply RV in WFx; apply RV in WFy; des.
  destruct (lookup (inl xb) envb) eqn:Ex'; s in EV; rewrite Ex' in EV; vauto.
  destruct (lookup (inl yb) envb) eqn:Ey'; s in EV; rewrite Ey' in EV; vauto.
  esplits; s; [by rewrite INa INa0; s; eauto|].
  rewrite Ey' in INb; rewrite Ex' in INb0; clarify.
  econs; eapply vclgen_pair; s; eauto.
Qed.

Lemma mm_compat_pair': forall lw G x y
    (WFx: In (inl x) G)
    (WFy: In (inl y) G),
  mm_asim' lw G (IEpair x y) (IEpair x y).
Proof.
  i; eapply mm_asim_old; eapply mm_compat_pair; eauto
  ; try by eapply extendable_t_same; i; rewrite ?ren_var_id; eauto.
  - rewrite ?ren_itvar_id ?ren_id_self_t ?ren_id_self_c; auto.
  - rewrite ?ren_itvar_id ?ren_id_self_t ?ren_id_self_c; auto.
Qed.


Lemma mm_compat_inl: forall lw G ra xa xb
    (WF: In (inl xb) G)
    (E: xa = rename_itvar ra xb),
  mm_asim lw G ra (IEinl xa) (IEinl xb).
Proof.
  ii; apply RV in WF; des.
  destruct (lookup (inl xb) envb) eqn:Ex; s in EV; rewrite Ex in EV; vauto.
  s in EV; clarify.
  esplits; s; eauto; [by rewrite INa; s; eauto|].
  rewrite Ex in INb; clarify.
  econs; eapply vclgen_inl; s; eauto.
Qed.

Lemma mm_compat_inl': forall lw G x
    (WF: In (inl x) G),
  mm_asim' lw G (IEinl x) (IEinl x).
Proof.
  i; eapply mm_asim_old; eapply mm_compat_inl; eauto
  ; try by eapply extendable_t_same; i; rewrite ?ren_var_id; eauto.
  rewrite ?ren_itvar_id ?ren_id_self_t ?ren_id_self_c; auto.
Qed.


Lemma mm_compat_inr: forall lw G ra xa xb
    (WF: In (inl xb) G)
    (E: xa = rename_itvar ra xb),
  mm_asim lw G ra (IEinr xa) (IEinr xb).
Proof.
  ii; apply RV in WF; des.
  destruct (lookup (inl xb) envb) eqn:Ex; s in EV; rewrite Ex in EV; vauto.
  s in EV; clarify.
  esplits; s; eauto; [by rewrite INa; s; eauto|].
  rewrite Ex in INb; clarify.
  econs; eapply vclgen_inr; s; eauto.
Qed.


Lemma mm_compat_inr': forall lw G x
    (WF: In (inl x) G),
  mm_asim' lw G (IEinr x) (IEinr x).
Proof.
  i; eapply mm_asim_old; eapply mm_compat_inr; eauto
  ; try by eapply extendable_t_same; i; rewrite ?ren_var_id; eauto.
  rewrite ?ren_itvar_id ?ren_id_self_t ?ren_id_self_c; auto.
Qed.


Lemma mm_good_fix: forall (lw: world_loca mm_gw)
    bfa bfb bxa bxb bka bkb ea eb s enva envb
    (GDa: is_good_thunk (IEfix bfa bxa bka ea))
    (GDb: is_good_thunk (IEfix bfb bxb bkb eb)),
  good_thunk (wlift lw) s
             (IVclo enva (IEfix bfa bxa bka ea))
             (IVclo envb (IEfix bfb bxb bkb eb)).
Proof.
  econs 1; s; econs; i; inv FH.
Qed.


Lemma mm_compat_fix: forall lw G ra bfa bfb bxa bxb bka bkb ea eb
    (EXf: extendable_t G ra bfa bfb)
    (EXx: extendable_t G ra bxa bxb)
    (X: bfa = bxa -> bfb = bxb)
    (IH: mm_psim lw (inr bkb::inl (inr bxb)::inl (inr bfb)::purge G)
                 (ra)[inr bfa/bfb][inr bxa/bxb][cbka/bkb] ea eb),
  mm_asim lw G ra (IEfix bfa bxa bka ea) (IEfix bfb bxb bkb eb).
Proof.
  ii; rrdes IH; s in EV; clarify; esplits; [by s; eauto|].
  econs; eapply vclgen_thunk; econs; s; eauto.
  eapply gk_lsim; eauto; econs; [apply mm_good_fix|].
  { ii; inv FH; eapply GEa; eauto. }
  { ii; inv FH; eapply GEb; eauto. }
  instantiate (1:=wfo_nat (i'+0)).
  pcofix CIH; i.
  apply pesim_gknow; i; apply pesim_gknow_; i.
  ii; invc STK; [|by invc INSTa RF VR].
  invc APPa APPb RF; invc CA CB VR RK.
  eapply pesim_mon; [eapply IH1|by eauto]; eauto; i.
  - eapply bla in EXx; eauto; [].
    eapply foo'; vauto; []; clear IN; i.
    rewrite ren_ext_it_neq_c.
    i; eapply foo'; eauto; vauto.
    { i; clarify; s; tmext_simpl; eauto. }
    { i; clarify.
      destruct (LEM (x0 = inr bx0)); vauto.
      rewrite ren_ext_it_neq in H; [|by ii; vauto].
      destruct x0; vauto.
      exploit (EXx t); s in H; i; des; vauto.
      apply in_consD in IN; des; vauto.
      eauto using purge_inD.
    }
    clear IN; i; rewrite ren_ext_it_neq; [eauto|by ii; vauto].
    eapply foo'; eauto; vauto.
    { i; clarify; s; tmext_simpl; splits; eauto.
      econs; eapply vclgen_thunk; econs 1; s; eauto.
      eapply gk_lsim; eauto; econs; [eapply mm_good_fix|].
      { ii; inv FH; eapply GEa; eauto. }
      { ii; inv FH; eapply GEb; eauto. }
      i; eapply gres_incl_id; right; eapply CIH; esplits; eauto using ext_trans.
    }
    { i; clarify.
      destruct (LEM (x1 = inr bf0)); vauto.
      rewrite ren_ext_it_neq in H; [|by ii; vauto].
      destruct x1; vauto.
      exploit (EXf t); s in H; i; des; vauto.
      apply in_consD in IN; des; vauto.
      eauto using purge_inD.
    }
    clear IN; i; rewrite ren_ext_it_neq; [eauto|by ii; vauto].
    hexploit RV; [by eapply purge_inD; eauto|]; i; des; esplits; eauto.
    eapply vclos_mon; eauto; i; eapply gk_mon; eauto.
  - eapply foo'; vauto.
    { i; clarify; s; tmext_simpl; eauto. }
    { i; clarify.
      destruct (LEM (x = bk0)); eauto.
      exfalso.
      do 3 (apply in_consD in IN; des; vauto).
      eauto using purge_inDr.
    }
    { clear IN; i; do 2 (apply in_consD in IN; des; vauto).
      exfalso; eauto using purge_inDr.
    }
Qed.

Lemma mm_compat_fix': forall lw G bf bx bk ea eb
    (IH: mm_psim' lw (inr bk::inl (inr bx)::inl (inr bf)::purge G)
                 ea eb),
  mm_asim' lw G (IEfix bf bx bk ea) (IEfix bf bx bk eb).
Proof.
  i; eapply mm_asim_old; eapply mm_compat_fix; eauto
  ; try by eapply extendable_t_same; i; rewrite ?ren_var_id; eauto.
  rewrite ?ren_itvar_id ?ren_id_self_t ?ren_id_self_c; auto. by eapply mm_psim_old.
Qed.


Lemma mm_good_gen: forall (lw: world_loca mm_gw) bka bkb ea eb s enva envb
    (GDa: is_good_thunk (IEgen bka ea))
    (GDb: is_good_thunk (IEgen bkb eb)),
  good_thunk (wlift lw) s
             (IVclo enva (IEgen bka ea)) (IVclo envb (IEgen bkb eb)).
Proof.
  econs 2; s; econs; i; inv FH.
Qed.

Lemma mm_compat_gen: forall lw G ra bka bkb ea eb
    (IH: mm_psim lw (inr bkb :: purge G) (ra)[cbka/bkb] ea eb),
  mm_asim lw G ra (IEgen bka ea) (IEgen bkb eb).
Proof.
  ii; rrdes IH; s in EV; clarify; esplits; [by s; eauto|].
  econs; eapply vclgen_thunk; econs 2; s; eauto.
  eapply gk_lsim; eauto; econs; [apply mm_good_gen|].
  { ii; inv FH; eapply GEa; eauto. }
  { ii; inv FH; eapply GEb; eauto. }
  instantiate (1:=wfo_nat (i'+0)).
  i; apply pesim_gknow; i; apply pesim_gknow_; i.
  ii; invc STK; [by invc APPa RF VR|].
  invc INSTa INSTb RF; invc CA CB VR RK.
  eapply pesim_mon; [eapply IH1|by eauto]; eauto; i.
  - eapply foo'; eauto; i; vauto.
    hexploit RV; [by eapply purge_inD; eauto|]; i; des; esplits; eauto.
    eapply vclos_mon; eauto; i; eapply gk_mon; eauto.
  - eapply foo'; eauto; vauto. 
    { i; clarify; s; tmext_simpl; eauto. }
    { i; clarify.
      destruct (LEM (x = bk0)); eauto.
      exfalso.
      apply in_consD in IN; des; vauto.
      eauto using purge_inDr.
    }
    { clear IN; i.
      exfalso; eauto using purge_inDr.
    }
Qed.

Lemma mm_compat_gen': forall lw G bk ea eb
    (IH: mm_psim' lw (inr bk :: purge G) ea eb),
  mm_asim' lw G (IEgen bk ea) (IEgen bk eb).
Proof.
  i; eapply mm_asim_old; eapply mm_compat_gen; eauto
  ; try by eapply extendable_t_same; i; rewrite ren_var_id; eauto.
  rewrite ?ren_itvar_id ?ren_id_self_t ?ren_id_self_c; auto. by eapply mm_psim_old.
Qed.


Lemma mm_compat_let: forall lw G ra ea ea' bxa bxb eb eb'
    (EX: extendable_t G ra bxa bxb)
    (IH: mm_asim lw G ra ea eb)
    (IH': mm_psim lw (inl (inr bxb) :: G) ra[inr bxa/bxb] ea' eb'),
  mm_psim lw G ra (IElet ea (inl bxa) ea') (IElet eb (inl bxb) eb').
Proof.
  i; rr; splits; vauto; []; exists 0; i; s.
  rrdes IH IH'; pfold; econs; i.
  mm_destruct XTRa XTRb CNF.
  eapply mm_esim_ok; [by eauto| |].
  - i; invc STb.
    hexploit IH; eauto; i; des; eauto using IRlet.
  - destruct (LEM (exists vb, eval envb eb = Some vb)); last first.
    { i; eapply mm_ok_esim_err.
        by econs; eapply IRerr; eauto; i; inv FH; vauto; apply f; vauto.
      s; rr; i; inv FH; des; clarify; rrdes PV; vauto. }
    des; esplits; eauto; eapply mm_ok_esim_step; eauto; i.
    destruct ma' as [? []]; s in STa; invc STa.
    do 7 eexists; exists false; splits; first (by eauto); last (by s; eauto)
    ; first 1 last.
    + econs; eauto; noemp; inst_pairs; s; noemp; rr; eauto 10.
    + left; econs; eauto.
      * left; eapply IRok; econs; eauto.
      * eapply gres_incl_id; left.
        rewrite 2!mid_csplit.
        eapply pesim_release; eauto; try (by rr; eauto).
        eapply IH'1; eauto; i.
        { eapply foo'; eauto.
          { i; clarify; s rename_itvar; rewrite tmext_eq; splits; eauto.
            exploit IH; eauto; i; des.
            rewrite x4 in EVAL; clarify.
          }
          { i; clarify.
            destruct (LEM (x2 = inr bxb)); vauto.
            rewrite ren_ext_it_neq in H; [|by ii; vauto].
            destruct x2; vauto.
            exploit (EX t); s in H; i; des; vauto.
            apply in_consD in IN; des; vauto.
          } 
          clear IN; i; rewrite ren_ext_it_neq; [|by ii; vauto].
          apply RV in IN; des; esplits; eauto.
        }
        { apply in_consD in IN; des; vauto.
          rewrite ren_ext_c_neq.
          rewrite !list_assoc_cons_neq; eauto; vauto.
        }
Qed.

Lemma mm_compat_let': forall lw G ea ea' bx eb eb'
    (IH: mm_asim' lw G ea eb)
    (IH': mm_psim' lw (inl (inr bx) :: G) ea' eb'),
  mm_psim' lw G (IElet ea (inl bx) ea') (IElet eb (inl bx) eb').
Proof.
  i; eapply mm_psim_old; eapply mm_compat_let; eauto
  ; try by eapply extendable_t_same; i; rewrite ren_var_id; eauto. by eapply mm_asim_old.
  rewrite ?ren_itvar_id ?ren_id_self_t ?ren_id_self_c; auto. by eapply mm_psim_old.
Qed.


Lemma mm_compat_cont: forall lw G ra bxa bxb ea eb
    (EX: extendable_t G ra bxa bxb)
    (IH: mm_psim lw (inl (inr bxb) :: G) ra[inr bxa/bxb] ea eb),
  mm_ksim lw G ra (IEcont bxa ea) (IEcont bxb eb).
Proof.
  ii; clarify; rrdes IH; esplits; eauto.
  econs; i; invc RETa RETb.
  eapply pesim_gknow; i.
  pfold; econs; i.
  mm_destruct XTRa XTRb CNF.
  esplits; eauto; eapply esim_step; eauto; i.
  invc ST; [invc STEP|by exfalso; eapply STK; econs; eauto].
  do 7 eexists; exists false; splits; first (by eauto); last (by s; eauto)
  ; last (econs; [by eauto|by eauto| |]); last first; first 1 last.
  - inst_pairs; s; noemp; rr; eauto 10.
  - left; econs; eauto.
    + left; inst_pairs; eapply IRok; econs; eauto.
    + rewrite GET in GETk; rewrite GETV in GETv; clarify.
      eapply gres_incl_id; left.
      rewrite mid_csplit; instantiate (1:=_#_).
      clear CO CO0.
      eapply pesim_release; eauto; try (by rr; eauto).
      inst_pairs; eapply IH1; eauto using extpub_trans.
      * i; eapply foo'; eauto.
        { i; clarify; s rename_itvar; rewrite tmext_eq; splits; eauto. }
        { i; clarify.
          destruct (LEM (x2 = inr bxb0)); vauto.
          rewrite ren_ext_it_neq in H; [|by ii; vauto].
          destruct x2; vauto.
          exploit (EX t); s in H; i; des; vauto.
          apply in_consD in IN; des; vauto.
        } 
        clear IN; i; rewrite ren_ext_it_neq; [|by ii; vauto].
        apply RV in IN; des; esplits; eauto.
        eapply vclos_mon; eauto; i; eapply gk_mon; eauto.
      * i; eapply foo'; i; clarify; eauto.
        eapply RK in IN0; des; esplits; eauto.
        eapply ksim_mon; eauto.
  - s; noemp; rr; eauto 10.
Qed.

Lemma mm_compat_cont': forall lw G bx ea eb
    (IH: mm_psim' lw (inl (inr bx) :: G) ea eb),
  mm_ksim' lw G (IEcont bx ea) (IEcont bx eb).
Proof.
  i; eapply mm_ksim_old; eapply mm_compat_cont; eauto
  ; try by eapply extendable_t_same; i; rewrite ren_var_id; eauto.
  rewrite ?ren_itvar_id ?ren_id_self_t ?ren_id_self_c; auto. by eapply mm_psim_old.
Qed.


Lemma mm_compat_letcont: forall lw G ra ea ea' bka bkb eb eb'
    (EX: extendable_c G ra bka bkb)
    (IH: mm_ksim lw G ra ea eb)
    (IH': mm_psim lw (inr bkb :: G) ra[cbka/bkb] ea' eb'),
  mm_psim lw G ra (IElet ea (inr bka) ea') (IElet eb (inr bkb) eb').
Proof.
  i; rr; splits; vauto; []; exists 0; i; s.
  rrdes IH IH'; pfold; econs; i.
  mm_destruct XTRa XTRb CNF.
  
  
  
  
  
  
  
  
  
  
  destruct (LEM (forall evt hb'' envb' e',
                   ~ reduce_ok evt ((hb' ## x3) ## x0) envb (IElet eb (inr bkb) eb')
                                   hb'' envb' e')).
  { eapply esim_err.
      by econs; eapply IRerr; eauto.
    rr; i; inv FH; des; clarify; rrdes PV; vauto. } 
  repeat (eapply not_all_ex_not in f; des); clear f; inv n3.
  hexploit IH; eauto; i; des; clarify.
  esplits; eauto; eapply esim_step; eauto; i.
  invc ST; [invc STEP|by exfalso; eapply STK; eauto].
  do 7 eexists; exists false; splits; first (by eauto); last (by s; eauto)
  ; last (econs; [by eauto|by eauto| |]); last first; first 1 last.
  - inst_pairs; s; noemp; rr; eauto 10.
  - left; econs; eauto.
    + left; inst_pairs; eapply IRok; econs; eauto.
    + eapply gres_incl_id; left.
      rewrite mid_csplit; instantiate (1:=_#_).
      clear CO CO0.
      eapply pesim_release; eauto; try (by rr; eauto).
      inst_pairs; eapply IH'1; eauto using extpub_trans.
      * instantiate (1 := (inr bkb, _) :: envb).
        i; eapply foo'; i; clarify; eauto.
      * i; eapply foo'; eauto.
        { i; clarify; simpl rename_contvar; s; tmext_simpl; eauto. }
        { i; clarify.
          destruct (LEM (x2 = bkb)); eauto.
          rewrite tmext_neq in H; [|by ii; vauto].
          exploit (EX x2); s in H; i; des; vauto.
          apply in_consD in IN; des; vauto.
        } 
        clear IN; i.
        rewrite ren_ext_c_neq'; [|by ii; vauto].
        eapply RK in IN; des; esplits; eauto.
  - s; noemp; rr; eauto 10.
Qed.

Lemma extendable_c_same: forall G ren x
    (E: forall y (IN: In y G) (E: rename_var ren y = inr x), rename_var ren y = y),
  extendable_c G ren x x.
Proof.
  s; ii; clarify; hexploit (E (inr xb')); eauto.
  s; i; clarify; rewrite H in E0; clarify.
Qed.

Lemma mm_compat_letcont': forall lw G ea ea' bk eb eb'
    (IH: mm_ksim' lw G ea eb)
    (IH': mm_psim' lw (inr bk :: G) ea' eb'),
  mm_psim' lw G (IElet ea (inr bk) ea') (IElet eb (inr bk) eb').
Proof.
  i; eapply mm_psim_old; eapply mm_compat_letcont; eauto
  ; try by eapply extendable_t_same; i; rewrite ren_var_id; eauto.
  - rewrite ?ren_itvar_id ?ren_id_self_t ?ren_id_self_c; auto.
    eapply extendable_c_same; i; rewrite ren_var_id; eauto.
  - rewrite ?ren_itvar_id ?ren_id_self_t ?ren_id_self_c; auto. by eapply mm_ksim_old.
  - rewrite ren_id_self_c. by eapply mm_psim_old.
Qed.


Hint Resolve 
  mm_compat_input mm_compat_output mm_compat_app mm_compat_read
  mm_compat_write mm_compat_new mm_compat_ret mm_compat_ifnz
  mm_compat_case mm_compat_inst mm_compat_bop mm_compat_refeq
  mm_compat_fst mm_compat_snd mm_compat_unit mm_compat_nat
  mm_compat_pair mm_compat_inl mm_compat_inr mm_compat_fix
  mm_compat_gen mm_compat_let mm_compat_cont mm_compat_letcont: mm_compat. 

Hint Resolve 
  mm_compat_input' mm_compat_output' mm_compat_app' mm_compat_read'
  mm_compat_write' mm_compat_new' mm_compat_ret' mm_compat_ifnz'
  mm_compat_case' mm_compat_inst' mm_compat_bop' mm_compat_refeq'
  mm_compat_fst' mm_compat_snd' mm_compat_unit' mm_compat_nat'
  mm_compat_pair' mm_compat_inl' mm_compat_inr' mm_compat_fix'
  mm_compat_gen' mm_compat_let' mm_compat_cont' mm_compat_letcont': mm_compat'. 






Lemma mm_asim_rename_: forall lw e G ren
    (WF: wfatom G e)
    (RN: forall x (IN: In x G) (BV: In (rename_var ren x) (bvars e)),
           rename_var ren x = x)
    (ST: stable lw)
    (IH: forall e', size e' < size e -> forall G' ren', wfexp G' e' -> 
             (forall x (IN: In x G') (BV: In (rename_var ren' x) (bvars e')),
                rename_var ren' x = x) -> stable lw ->
             mm_psim lw G' ren' (rename ren' e') e'),
  mm_asim lw G ren (rename ren e) e.
Proof.
  i; destruct e; vauto; ss; des; eauto with mm_compat.
  - destruct x, y; eauto using mm_compat_bop.
  - eapply mm_compat_fix; eauto using extendable_t_same.
    eapply IH; eauto.
    { i; revert BV.
      apply in_consD in IN; des; subst; [by s; tmext_simpl; auto|].
      rewrite ren_var_ext_neq; auto.
      apply in_consD in IN; des; subst; [by s; tmext_simpl; auto|].
      rewrite ren_ext_neq; auto.
      apply in_consD in IN; des; subst; [by s; tmext_simpl; auto|].
      rewrite ren_ext_neq; auto 10 using purge_inD.
    }
  - eapply mm_compat_gen; eauto.
    eapply IH; eauto.
    { i; revert BV.
      apply in_consD in IN; des; subst; [by s; tmext_simpl; auto|].
      rewrite ren_var_ext_neq; auto 10 using purge_inD.
    }
Qed.


Lemma mm_ksim_rename_: forall lw e G ren
    (WF: wfatom G e)
    (RN: forall x (IN: In x G) (BV: In (rename_var ren x) (bvars e)),
           rename_var ren x = x)
    (ST: stable lw)
    (IH: forall e', size e' < size e -> forall G' ren', wfexp G' e' -> 
             (forall x (IN: In x G') (BV: In (rename_var ren' x) (bvars e')),
                        rename_var ren' x = x) -> stable lw ->
             mm_psim lw G' ren' (rename ren' e') e'),
  mm_ksim lw G ren (rename ren e) e.
Proof.
  ii; destruct e; vauto.
  hexploit (IH eb'); try exact WF; eauto.
  - instantiate (1:=ren[inr bxb/bxb]).
    i; revert BV.
    apply in_consD in IN; des; subst; [by s; tmext_simpl; auto|].
    s in RN; rewrite ren_ext_neq; auto.
  - esplits; [by auto|].
    hexploit mm_compat_cont; eauto; last first.
    + intro SIM; unfold mm_ksim in SIM; hexploit SIM; eauto; i; des; eauto.
    + s in RN; eapply extendable_t_same; auto.
Qed.


Theorem mm_psim_rename: forall lw e G ren
    (WF: wfexp G e)
    (RN: forall x (IN: In x G) (BV: In (rename_var ren x) (bvars e)),
           rename_var ren x = x)
    (STA: stable lw),
  mm_psim lw G ren (rename ren e) e.
Proof.
  intros ? ?.
  induction e as [? IH] using (well_founded_ind size_lt_well_founded).
  destruct e; vauto; i; s in WF; des
  ; eauto using mm_compat_write, mm_compat_ret, mm_compat_output,
                mm_compat_app, mm_compat_inst.
  - simpl rename; s in RN.
    eapply mm_compat_new; eauto using extendable_t_same.
    eapply IH; eauto.
    i; revert BV.
    apply in_consD in IN; des; subst; [by s; tmext_simpl; auto|].
    s in RN; rewrite ren_ext_neq; auto.
  - simpl rename; s in RN.
    eapply mm_compat_read; eauto using extendable_t_same.
    eapply IH; eauto.
    i; revert BV.
    apply in_consD in IN; des; subst; [by s; tmext_simpl; auto|].
    s in RN; rewrite ren_ext_neq; auto.
  - simpl rename; s in RN.
    eapply mm_compat_input; eauto using extendable_t_same.
    eapply IH; eauto.
    i; revert BV.
    apply in_consD in IN; des; subst; [by s; tmext_simpl; auto|].
    s in RN; rewrite ren_ext_neq; auto.
  - simpl rename; s in RN.
    eapply mm_compat_ifnz; eauto using extendable_t_same
    ; (eapply IH; eauto using in_or_app; s; omega).
  - simpl rename; s in RN.
    eapply mm_compat_case; eauto using extendable_t_same.
    + eapply IH; eauto; [s; omega|].
      i; revert BV.
      apply in_consD in IN; des; subst; [by s; tmext_simpl; auto|].
      s in RN; rewrite ren_ext_neq; auto using in_or_app.
    + eapply IH; eauto; [s; omega|].
      i; revert BV.
      apply in_consD in IN; des; subst; [by s; tmext_simpl; auto|].
      s in RN; rewrite ren_ext_neq; auto using in_or_app.
  - destruct bx; des; i.
    + eapply mm_asim_rename_ in WF; eauto.
      * simpl rename; s in RN.
        eapply mm_compat_let; eauto using extendable_t_same.
        eapply IH; eauto; [s; omega|].
        i; revert BV.
        apply in_consD in IN; des; subst; [by s; tmext_simpl; auto|].
        s in RN; rewrite ren_ext_neq; auto using in_or_app.
      * i; s in RN; eauto using in_or_app.
      * i; eapply IH; s; eauto; omega.
    + eapply (mm_ksim_rename_ _ ren) in WF; eauto.
      * simpl rename; s in RN.
        eapply mm_compat_letcont; eauto using extendable_c_same.
        eapply IH; eauto; [s; omega|].
        i; revert BV.
        apply in_consD in IN; des; subst; [by s; tmext_simpl; auto|].
        s in RN; rewrite ren_c_ext_neq; auto using in_or_app.
      * i; s in RN; eauto using in_or_app.
      * i; eapply IH; s; eauto; omega.
Qed.


Theorem mm_asim_rename: forall lw e G ren
    (WF: wfatom G e)
    (RN: forall x (IN: In x G) (BV: In (rename_var ren x) (bvars e)),
           rename_var ren x = x)
    (ST: stable lw),
  mm_asim lw G ren (rename ren e) e.
Proof.
  i; eapply mm_asim_rename_; eauto.
  i; eapply mm_psim_rename; eauto.
Qed.


Theorem mm_ksim_rename: forall lw e G ren
    (WF: wfatom G e)
    (RN: forall x (IN: In x G) (BV: In (rename_var ren x) (bvars e)),
           rename_var ren x = x)
    (ST: stable lw),
  mm_ksim lw G ren (rename ren e) e.
Proof.
  i; eapply mm_ksim_rename_; eauto.
  i; eapply mm_psim_rename; eauto.
Qed.


Theorem mm_psim_refl: forall lw e G
    (WF: wfexp G e)
    (STA: stable lw),
  mm_psim lw G ren_id e e.
Proof.
  i; rewrite <-(rename_id e) at 1.
  eapply mm_psim_rename; eauto.
  i; rewrite ren_var_id; eauto.
Qed.

Theorem mm_asim_refl: forall lw e G
    (WF: wfatom G e)
    (ST: stable lw),
  mm_asim lw G ren_id e e.
Proof.
  i; rewrite <-(rename_id e) at 1.
  eapply mm_asim_rename; eauto.
  i; rewrite ren_var_id; eauto.
Qed.

Theorem mm_ksim_refl: forall lw e G
    (WF: wfatom G e)
    (ST: stable lw),
  mm_ksim lw G ren_id e e.
Proof.
  i; rewrite <-(rename_id e) at 1.
  eapply mm_ksim_rename; eauto.
  i; rewrite ren_var_id; eauto.
Qed.




(** * Reflexivity (putting things together). *)

















Inductive mm_tlsim lw FI : exp -> exp -> Prop :=
  | mm_tlsim_fix bf bx bk ea eb
      (SIM: mm_psim lw (inr bk::inl (inr bx)::inl (inr bf)::map (inl<*>inl) FI)
                    ren_id ea eb)
    : mm_tlsim lw FI (IEfix bf bx bk ea) (IEfix bf bx bk eb)
  | mm_tlsim_gen bk ea eb
      (SIM: mm_psim lw (inr bk::map (inl<*>inl) FI)
                    ren_id ea eb)
    : mm_tlsim lw FI (IEgen bk ea) (IEgen bk eb)
.


Lemma mm_tlsim_refl: forall lw FI e
    (GT: is_good_thunk e)
    (WF: wfatom (map (fun l => inl (inl l)) FI) e)
    (ST: stable lw),
  mm_tlsim lw FI e e.
Proof.
  i; destruct e; vauto
  ; (s in WF; rewrite map_comp purge_map_inl -map_comp in WF
     ; econs; eauto using mm_psim_refl).
Qed.


Lemma mm_msim_nil: forall FI,
  msim mm_gw FI nil nil nil.
Proof.
  i; exists (lwemp mm_gw); i; exists (tt, tt, refst_emp, tt).
  invc LMa LMb; clean.
  esplits; eauto using lwemp_stable.
  - i; rr; repeat ex @cemp; splits.
    + split; rr; auto.
    + apply mm_refconfrel_emp; eauto.
    + s; noemp; auto.
    + s; noemp; auto.
  - i; rr; eauto.
  - i; rrdes FH; inv REG.
  - i; vauto.
Qed.


Lemma mm_tlsim_tlsim: forall lw suba subb s ea eb
    (VSIM: mm_tlsim lw (map fst suba) ea eb)
    (DOM: map fst suba = map fst subb),
  @tlsim _ _ mm_gw lw (map fst suba) suba subb s
         (IVclo (map_pair (inl <*> inl) id suba) ea)
         (IVclo (map_pair (inl <*> inl) id subb) eb).
Proof.
  i; invc VSIM.
  - econs 1; s; eauto.
    dup SIM; rrdes SIM0.
    rewrite map_comp -purge_map_inl -map_comp in SIM.
    eapply mm_psim_old in SIM.
    eapply mm_compat_fix' in SIM; eauto.
    
    eapply gk_lsim; eauto.
    econs; [eapply mm_good_fix; eauto|].
    { ii; inv FH; eapply GEa; eauto. }
    { ii; inv FH; eapply GEb; eauto. }
    i; invc STK; [|by invc RF VR INSTa CA].
    invc RK RF VR; invc APPa APPb CA CB.
    eapply pesim_gknow; i.
    eapply SIM0; eauto.
    + i; apply in_consD in IN; des; [by vauto|].
      rewrite (list_assoc_cons_neq _ ka); [|by vauto].
      rewrite (list_assoc_cons_neq _ kb); [|by vauto].
      apply in_consD in IN; des.
      { clarify; rewrite 2!list_assoc_cons_eq; esplits; eauto. }
      rewrite ren_itvar_id.
      rewrite (list_assoc_cons_neq _ va); [|by ii; vauto].
      rewrite (list_assoc_cons_neq _ vb); [|by ii; vauto].
      apply in_consD in IN; des.
      { clarify; rewrite 2!list_assoc_cons_eq; esplits; eauto.
        exploit SIM; try by s; eauto.
        - instantiate (3 := map_pair (inl <*> inl) id subb).
          instantiate (3 := map_pair (inl <*> inl) id suba).
          i; apply in_map_iff in IN; des; s in IN; clarify.
          dup IN2; dup IN2; rewrite DOM in IN2.
          eapply list_assoc_map in IN2; eapply list_assoc_map in IN3; des.
          eapply RS in IN4; eauto; esplits.
          + eapply list_assoc_map_pair; eauto; i; ss; clarify.
          + eapply list_assoc_map_pair; eauto; i; ss; clarify.
          + econs; eapply vclgen_thunk, vclos_thunk_mon; eauto.
            i; eapply LE0, gk_mon; eauto.
        - i; des; s in x0; clarify.
      }
      rewrite list_assoc_cons_neq; [|by ii; vauto].
      rewrite list_assoc_cons_neq; [|by ii; vauto].
      eapply in_map_iff in IN; des; s in IN; clarify.
      hexploit list_assoc_map; eauto; []; i; des.
      rewrite DOM in IN3.
      hexploit list_assoc_map; eauto; []; i; des.
      rewrite -DOM in IN3.
      hexploit RS; eauto.
      i; esplits.
      * eapply list_assoc_map_pair; eauto; i; ss; clarify.
      * eapply list_assoc_map_pair; eauto; i; ss; clarify.
      * econs; eapply vclgen_thunk, vclos_thunk_mon; eauto.
        i; eapply LE0, gk_mon; eauto.
    + i; apply in_consD in IN; des.
      { clarify; rewrite 2!list_assoc_cons_eq; esplits; eauto. }
      rewrite (list_assoc_cons_neq _ ka); [|by ii; vauto].
      rewrite (list_assoc_cons_neq _ kb); [|by ii; vauto].
      apply in_consD in IN; des; [by vauto|].
      rewrite (list_assoc_cons_neq _ va); [|by ii; vauto].
      rewrite (list_assoc_cons_neq _ vb); [|by ii; vauto].
      apply in_consD in IN; des; [by vauto|].
      exfalso; apply in_map_iff in IN; des; vauto.
  - 
    econs 2; s; eauto.
    dup SIM; rrdes SIM0.
    
    eapply gk_lsim; eauto.
    econs; [eapply mm_good_gen; eauto|].
    { ii; inv FH; eapply GEa; eauto. }
    { ii; inv FH; eapply GEb; eauto. }
    i; invc STK; [by invc RF VR APPa CA|].
    invc RK RF VR; invc INSTa INSTb CA CB.
    eapply pesim_gknow; i.
    eapply SIM0; eauto.
    + i; apply in_consD in IN; des; [by vauto|].
      rewrite (list_assoc_cons_neq _ ka); [|by vauto].
      rewrite (list_assoc_cons_neq _ kb); [|by vauto].
      eapply in_map_iff in IN; des; s in IN; clarify.
      hexploit list_assoc_map; eauto; []; i; des.
      rewrite DOM in IN1.
      hexploit list_assoc_map; eauto; []; i; des.
      rewrite -DOM in IN1.
      hexploit RS; eauto.
      i; esplits.
      * eapply list_assoc_map_pair; eauto; i; ss; clarify.
      * eapply list_assoc_map_pair; eauto; i; ss; clarify.
      * econs; eapply vclgen_thunk, vclos_thunk_mon; eauto.
        i; eapply LE0, gk_mon; eauto.
    + i; apply in_consD in IN; des.
      { clarify; rewrite 2!list_assoc_cons_eq; esplits; eauto. }
      rewrite (list_assoc_cons_neq _ ka); [|by ii; vauto].
      rewrite (list_assoc_cons_neq _ kb); [|by ii; vauto].
      exfalso; apply in_map_iff in IN; des; vauto.
Grab Existential Variables.
  exact 0. exact 0.
Qed.
            

Lemma mm_msim_cons: forall lw FI ea eb l FO ma mb sil
    (VSIM: mm_tlsim lw FI ea eb)
    (MSIM: msim mm_gw (FI++[l]) FO ma mb)
    (ST: stable lw)
    (INIl: forall g (GK: gknow (wlift lw) g),
             lw.(confrel) g (tt, tt, refst_emp, sil) cemp cemp),
  msim mm_gw FI (l::FO) ((l,ea)::ma) ((l,eb)::mb).
Proof.
  i; rrdes MSIM.
  exists (wprod lw lw0); i.
  set (va := IVclo (map_pair (inl <*> inl) id suba) ea).
  set (vb := IVclo (map_pair (inl <*> inl) id subb) eb).
  hexploit MSIM; rr; eauto.
  { instantiate (1 := suba ++ [(l, va)]).
    rewrite map_app SUBa; auto.
  }
  { instantiate (1 := subb ++ [(l, vb)]).
    rewrite map_app SUBb; auto.
  }
  i; des.
  exists (fst si, (sil, snd si)).
  rrdes LMa LMb; subst.
  assert (fst si = (tt, tt, refst_emp)).
  { destruct si as [[[[][]]]]; s; repeat f_equal.
    hexploit INIg; eauto using gkmin_bot_gknow; []; i.
    rrdes X WHX.
    destruct hax, hay, hbx, hby; s in HPA; s in HPB; clarify.
    symmetry in H1; apply pmemp_plus in H1.
    symmetry in H3; apply pmemp_plus in H3.
    des; subst; rrdes WHY; clarify.
    apply refst_extensionality; split; i; vauto.
    apply RELV in H; des; apply pmread_empD in RDa; vauto.
  }    
  destruct si; clean; subst.

  assert (SIM': @tlsim _ _ mm_gw lw (map fst suba) suba subb
                       (tt, tt, refst_emp, sil) va vb).
  { eapply mm_tlsim_tlsim; eauto. }

  splits; eauto.
  - eapply stable_prod; eauto.
  - i; hexploit gknow_prod; first 2 last; eauto; []; i; des.
    eapply confrel_mon; [by eapply INIg; eauto|].
    i; rrdes PR; inv PR; eapply gk_mon; s; eauto.
  - i; hexploit gknow_prod; first 2 last; eauto; []; i; des.
    move INIl at bottom; move INIl0 at bottom.
    rr; esplits; clean; eauto; noemp; eauto.
  - i; apply in_consD in IN; des; subst.
    + esplits; try (rr; s; des_if); vauto.
      ii; rrdes LE LE0; destruct s as [[[[]]]]; destruct s2; clean.
      hexploit gknow_prod; first 2 last; eauto; []; i; des.
      exploit SIM'.
      * inst_pairs; split; [exact LE|exact LE0].
      * eauto.
      * rrdes RLS RLS; rr; eauto.
      * i; exploit RS; eauto; clean; i.
        eapply vclos_thunk_lift, vclos_thunk_unlift, vclos_thunk_mon; eauto.
      * i; eapply vclos_thunk_lift, vclos_thunk_unlift, vclos_thunk_mon; eauto.
        i; invc PR; eapply GK; repeat (eauto; split; clean).
    + hexploit SIM; eauto; []; i; des.
      esplits.
      * rr; s; des_if; [by exfalso; vauto|].
        rr in Ea; eauto.
      * rr; s; des_if; [by exfalso; vauto|].
        rr in Eb; eauto.
      * { ii; rrdes LE LE0; destruct s as [[[]]]; destruct s3; clean.
          i; hexploit gknow_prod; first 2 last; eauto; []; i; des.
          exploit SIM0.
          - inst_pairs; split; [exact LE|exact LE1].
          - eauto.
          - rrdes RLS RLS; rr; eauto.
          - i; eapply (lookup2 _ _ _ _ _ _ _ SUBb); eauto; i.
            + exploit RS; eauto.
              * apply in_map_fst; eauto using list_assoc_someD.
              * clean; i.
                eapply vclos_thunk_lift; eapply vclos_thunk_unlift in x1.
                eapply vclos_thunk_mon; eauto.
            + sng; exploit SIM'.
              * inst_pairs; split; [exact LE|exact LE0].
              * eauto.
              * rrdes RLS RLS; rr; esplits; rr; clean; esplits; eauto.
              * i; exploit RS; eauto; []; i.
                eapply vclos_thunk_lift, vclos_thunk_unlift.
                eapply vclos_thunk_mon; eauto.
              * i; eapply vclos_thunk_lift; eapply vclos_thunk_unlift in x0.
                eapply vclos_thunk_mon; eauto.
                i; inv PR; econs.
                { eapply GK; repeat (eauto; split; clean). }
                eauto.
          - i; eapply vclos_thunk_lift, vclos_thunk_unlift.
            eapply vclos_thunk_mon; eauto.
            i; invc PR; eapply GK; repeat (eauto; split; clean).
        }
Grab Existential Variables.
  exact tt. exact tt.
Qed.



Lemma mm_msim_refl: forall FI m FO
    (WFM: mid_wfmod FI m FO),
  msim mm_gw FI FO m m.
Proof.
  i; ginduction m; i; inv WFM.
  - eauto using mm_msim_nil.
  - eapply mm_msim_cons; eauto.
    + eapply mm_tlsim_refl; eauto using lwemp_stable.
    + eauto using lwemp_stable.
    + i; rr; eauto.
Grab Existential Variables.
  exact tt.
Qed.












(** * "Compatibility" lemmas lifted to the reflexive transitive closure. *)

Require Import Relations.

Lemma mmrtc_compat_input: forall lw G bx ea eb
    (IH: rtc (mm_psim lw (inl (inr bx) :: G) ren_id) ea eb),
  rtc (mm_psim lw G ren_id) (IEinput bx ea) (IEinput bx eb).
Proof.
  i; apply clos_rt_rt1n_iff in IH; ginduction IH; i; eauto.
  econs 3; [econs 1; eapply mm_psim_old; eapply mm_compat_input'; eapply mm_psim_old|]; eauto.
Qed.

Lemma mmrtc_compat_output: forall lw G x ea eb
    (WF: In (inl x) G)
    (IH: rtc (mm_psim lw G ren_id) ea eb),
  rtc (mm_psim lw G ren_id) (IEoutput x ea) (IEoutput x eb).
Proof.
  i; apply clos_rt_rt1n_iff in IH; ginduction IH; i; eauto.
  econs 3; [econs 1; eapply mm_psim_old; eapply mm_compat_output'; [|eapply mm_psim_old]|]; eauto.
Qed.

Lemma mmrtc_compat_read: forall lw G x bx ea eb
    (WF: In (inl x) G)
    (IH: rtc (mm_psim lw (inl (inr bx) :: G) ren_id) ea eb),
  rtc (mm_psim lw G ren_id) (IEread x bx ea) (IEread x bx eb).
Proof.
  i; apply clos_rt_rt1n_iff in IH; ginduction IH; i; eauto.
  econs 3; [econs 1; eapply mm_psim_old; eapply mm_compat_read'; [|eapply mm_psim_old]|]; eauto.
Qed.

Lemma mmrtc_compat_write: forall lw G x y ea eb
    (WFx: In (inl x) G)
    (WFy: In (inl y) G)
    (IH: rtc (mm_psim lw G ren_id) ea eb),
  rtc (mm_psim lw G ren_id) (IEwrite x y ea) (IEwrite x y eb).
Proof.
  i; apply clos_rt_rt1n_iff in IH; ginduction IH; i; eauto.
  econs 3; [econs 1; eapply mm_psim_old; eapply mm_compat_write'; [| |eapply mm_psim_old]|]; eauto.
Qed.

Lemma mmrtc_compat_new: forall lw G x bx ea eb
    (IH: rtc (mm_psim lw (inl (inr bx) :: G) ren_id) ea eb)
    (WF: In (inl x) G)
    (STA: stable lw),
  rtc (mm_psim lw G ren_id) (IEnew x bx ea) (IEnew x bx eb).
Proof.
  i; apply clos_rt_rt1n_iff in IH; ginduction IH; i; eauto.
  econs 3; [econs 1; eapply mm_psim_old; eapply mm_compat_new'; [|eapply mm_psim_old|]|]; eauto.
Qed.

Lemma mmrtc_compat_ifnz: forall lw G x ea ea' eb eb'
    (WF: In (inl x) G)
    (IH: rtc (mm_psim lw G ren_id) ea eb)
    (WFIH: wfexp G eb)
    (IH': rtc (mm_psim lw G ren_id) ea' eb')
    (WFIH': wfexp G eb')
    (ST: stable lw),
  rtc (mm_psim lw G ren_id) (IEifnz x ea ea') (IEifnz x eb eb').
Proof.
  i; apply clos_rt_rt1n_iff in IH; apply clos_rt_rt1n_iff in IH'.
  ginduction IH; i; eauto.
  - ginduction IH'; i; eauto.
    econs 3; [econs 1; eapply mm_psim_old; eapply mm_compat_ifnz'; [|eapply mm_psim_old|eapply mm_psim_old]|]; 
    eauto using mm_psim_refl.
  - inv IH'
    ; (econs 3; [econs 1; eapply mm_psim_old; eapply mm_compat_ifnz'; [|eapply mm_psim_old|eapply mm_psim_old]|]; 
       eauto using mm_psim_refl).
Qed.

Lemma mmrtc_compat_case: forall lw G x bx ea ea' eb eb'
    (WF: In (inl x) G)
    (IH: rtc (mm_psim lw (inl (inr bx) :: G) ren_id) ea eb)
    (WFIH: wfexp (inl (inr bx) :: G) eb)
    (IH': rtc (mm_psim lw (inl (inr bx) :: G) ren_id) ea' eb')
    (WFIH': wfexp (inl (inr bx) :: G) eb')
    (ST: stable lw),
  rtc (mm_psim lw G ren_id) (IEcase x bx ea ea') (IEcase x bx eb eb').
Proof.
  i; apply clos_rt_rt1n_iff in IH; apply clos_rt_rt1n_iff in IH'.
  ginduction IH; i; eauto.
  - ginduction IH'; i; eauto.
    econs 3; [econs 1; eapply mm_psim_old; eapply mm_compat_case'; [|eapply mm_psim_old|eapply mm_psim_old]|]; eauto using mm_psim_refl.
  - inv IH'
    ; (econs 3; [econs 1; eapply mm_psim_old; eapply mm_compat_case'; [|eapply mm_psim_old|eapply mm_psim_old]|]; eauto using mm_psim_refl).
Qed.

Lemma mmrtc_compat_fix: forall lw G bf bx bk ea eb
    (IH: rtc (mm_psim lw (inr bk::inl (inr bx)::inl (inr bf)::purge G) ren_id) ea eb),
  rtc (mm_asim lw G ren_id) (IEfix bf bx bk ea) (IEfix bf bx bk eb).
Proof.
  i; apply clos_rt_rt1n_iff in IH; ginduction IH; i; eauto.
  econs 3; [econs 1; eapply mm_asim_old; eapply mm_compat_fix'|]; eauto; by eapply mm_psim_old.
Qed.

Lemma mmrtc_compat_gen: forall lw G bk ea eb
    (IH: rtc (mm_psim lw (inr bk :: purge G) ren_id) ea eb),
  rtc (mm_asim lw G ren_id) (IEgen bk ea) (IEgen bk eb).
Proof.
  i; apply clos_rt_rt1n_iff in IH; ginduction IH; i; eauto.
  econs 3; [econs 1; eapply mm_asim_old; eapply mm_compat_gen'|]; eauto; by eapply mm_psim_old.
Qed.

Lemma mmrtc_compat_let: forall lw G ea ea' bx eb eb'
    (IH: rtc (mm_asim lw G ren_id) ea eb)
    (WFIH: wfatom G eb)
    (IH': rtc (mm_psim lw (inl (inr bx) :: G) ren_id) ea' eb')
    (WFIH': wfexp (inl (inr bx) :: G) eb')
    (ST: stable lw),
  rtc (mm_psim lw G ren_id) (IElet ea (inl bx) ea') (IElet eb (inl bx) eb').
Proof.
  i; apply clos_rt_rt1n_iff in IH; apply clos_rt_rt1n_iff in IH'.
  ginduction IH; i; eauto.
  - ginduction IH'; i; eauto.
    econs 3; [econs 1; eapply mm_psim_old; eapply mm_compat_let'; [eapply mm_asim_old|eapply mm_psim_old]|]; 
    eauto using mm_asim_refl.
  - inv IH'
    ; (econs 3; [econs 1; eapply mm_psim_old; eapply mm_compat_let'; [eapply mm_asim_old|eapply mm_psim_old]|]; 
       eauto using mm_psim_refl).
Qed.

Lemma mmrtc_compat_cont: forall lw G bx ea eb
    (IH: rtc (mm_psim lw (inl (inr bx) :: G) ren_id) ea eb),
  rtc (mm_ksim lw G ren_id) (IEcont bx ea) (IEcont bx eb).
Proof.
  i; apply clos_rt_rt1n_iff in IH; ginduction IH; i; eauto.
  econs 3; [econs 1; eapply mm_ksim_old; eapply mm_compat_cont'|]; eauto. by eapply mm_psim_old.
Qed.

Lemma mmrtc_compat_letcont: forall lw G ea ea' bxk eb eb'
    (IH: rtc (mm_ksim lw G ren_id) ea eb)
    (WFIH': wfatom G eb)
    (IH': rtc (mm_psim lw (inr bxk :: G) ren_id) ea' eb')
    (WFIH': wfexp (inr bxk :: G) eb')
    (ST: stable lw),
  rtc (mm_psim lw G ren_id) (IElet ea (inr bxk) ea') (IElet eb (inr bxk) eb').
Proof.
  i; apply clos_rt_rt1n_iff in IH; apply clos_rt_rt1n_iff in IH'.
  ginduction IH; i; eauto.
  - ginduction IH'; i; eauto.
    econs 3; [econs 1; eapply mm_psim_old; eapply mm_compat_letcont'; [eapply mm_ksim_old|eapply mm_psim_old]|]; eauto using mm_ksim_refl.
  - inv IH'
    ; (econs 3; [econs 1; eapply mm_psim_old; eapply mm_compat_letcont'; [eapply mm_ksim_old|eapply mm_psim_old]|]; eauto using mm_psim_refl).
Qed.


Lemma mmrtc_tlsim_fix: forall lw FI bf bx bk ea eb
    (SIM: rtc (mm_psim lw (inr bk::inl (inr bx)::inl (inr bf):: map (inl<*>inl) FI) ren_id)
              ea eb),
    rtc (mm_tlsim lw FI)
        (lang_mid.IEfix bf bx bk ea) (lang_mid.IEfix bf bx bk eb).
Proof.
  i; ginduction SIM; [econs 1|econs 2|econs 3]; eauto; econs 1; eauto.
Qed.

Lemma mmrtc_tlsim_gen: forall lw FI bk ea eb
    (SIM: rtc (mm_psim lw (inr bk :: map (inl<*>inl) FI) ren_id) ea eb),
    rtc (mm_tlsim lw FI) (lang_mid.IEgen bk ea) (lang_mid.IEgen bk eb).
Proof.
  i; ginduction SIM; [econs 1|econs 2|econs 3]; eauto; econs 2; eauto.
Qed.


Definition modmap (f: exp -> exp) : module -> module :=
  map_pair id f.



Lemma modmap_wfmod: forall f FI FO m
    (WF: mid_wfmod FI m FO)
    (UV: uniqvar m)
    (X: forall xs F e
            (GT: is_good_thunk e)
            (WF: wfatom (map (inl<*>inl) F) e)
            (UN: uniq (bvars e ++ xs)),
          uniq (bvars (f e) ++ xs) /\
          is_good_thunk (f e) /\
          wfatom (map (inl<*>inl) F) (f e)),
  mid_wfmod FI (modmap f m) FO /\
  uniqvar (modmap f m).
Proof.
  i.
  match goal with |- ?P =>
    cut (P /\ forall xs (UN: uniq (xs ++ bvars_mod m)),
                uniq (xs ++ bvars_mod (modmap f m)))
  end; [by i; des; eauto|].
  i; depgen UV; induction WF; s; vauto; i.
  r in UV; s in UV; dup UV; apply uniq_appD in UV; des.
  hexploit IHWF; auto; []; i; des.
  hexploit X; eauto; []; i; des.
  splits; eauto; [by r; s; eauto|]; clear X4.
  intro; rewrite 2!app_assoc; i.
  hexploit (X (xs ++ bvars_mod m)); eauto; rewrite app_assoc
  ; [|i; des; apply X1]
  ; (eapply uniq_perm; eauto using Permutation_app, Permutation_app_comm).
Qed.


Lemma mm_rtcmsim_cons: forall lw FI ea eb l FO ma mb sil
    (VSIM: rtc (mm_tlsim lw FI) ea eb)
    (MSIM: rtc (msim mm_gw (FI++[l]) FO) ma mb)
    (WFM: mid_wfmod (FI) ((l,eb)::mb) (l::FO))
    (ST: stable lw)
    (INIl: forall g (GK: gknow (wlift lw) g),
             lw.(confrel) g (tt, tt, refst_emp, sil) cemp cemp),
  rtc (msim mm_gw FI (l::FO)) ((l,ea)::ma) ((l,eb)::mb).
Proof.
  i; apply clos_rt_rt1n_iff in VSIM; apply clos_rt_rt1n_iff in MSIM.
  invc WFM; ginduction VSIM; i.
  - ginduction MSIM; i; eauto.
    econs 3; last eapply IHMSIM; eauto.
    econs 1; eapply mm_msim_cons; eauto using mm_tlsim_refl.
  - inv MSIM.
    + hexploit IHVSIM; eauto; []; i.
      econs 3; last by eauto.
      econs 1; eapply mm_msim_cons; eauto using mm_msim_refl.
    + hexploit IHVSIM; eauto; []; i.
      econs 3; last by eauto.
      econs 1; eapply mm_msim_cons; eauto.
Qed.






Inductive mm_tlsimB lw FI : exp -> exp -> Prop :=
  | mm_tlsimB_fix bfa bfb bxa bxb bka bkb ea eb
      (SIM: mm_psim lw (inr bkb::inl (inr bxb)::inl (inr bfb)::map (inl<*>inl) FI)
                    (ren_id)[inr bfa/bfb][inr bxa/bxb][cbka/bkb] ea eb)
      (E: bfa = bxa -> bfb = bxb)
    : mm_tlsimB lw FI (IEfix bfa bxa bka ea) (IEfix bfb bxb bkb eb)
  | mm_tlsimB_gen bka bkb ea eb
      (SIM: mm_psim lw (inr bkb::map (inl<*>inl) FI)
                    (ren_id)[cbka/bkb] ea eb)
    : mm_tlsimB lw FI (IEgen bka ea) (IEgen bkb eb)
.


Lemma mm_tlsimB_tlsim: forall lw suba subb s ea eb
    (VSIM: mm_tlsimB lw (map fst suba) ea eb)
    (DOM: map fst suba = map fst subb),
  @tlsim _ _ mm_gw lw (map fst suba) suba subb s
         (IVclo (map_pair (inl <*> inl) id suba) ea)
         (IVclo (map_pair (inl <*> inl) id subb) eb).
Proof.
  i; invc VSIM.
  - econs 1; s; eauto.
    dup SIM; rrdes SIM0.
    rewrite map_comp -purge_map_inl -map_comp in SIM.

    eapply mm_compat_fix in SIM; eauto
    ; try (by ii; apply in_map_iff in IN; des; vauto); [].

    
    eapply gk_lsim; eauto.
    econs; [eapply mm_good_fix; eauto|].
    { ii; inv FH; eapply GEa; eauto. }
    { ii; inv FH; eapply GEb; eauto. }
    i; invc STK; [|by invc RF VR INSTa CA].
    invc RK RF VR; invc APPa APPb CA CB.
    eapply pesim_gknow; i.
    eapply SIM0; eauto.
    + i; eapply foo'; vauto; []; clear IN; i.
      rewrite ren_ext_it_neq_c.
      i; eapply foo'; eauto; vauto.
      { i; clarify; s; tmext_simpl; eauto. }
      { i; clarify.
        destruct (LEM (x0 = inr bxb)); vauto.
        rewrite ren_ext_it_neq in H; [|by ii; vauto].
        exfalso.
        destruct x0; vauto.
        apply in_consD in IN; des; vauto.
        destruct (LEM (t = bfb)).
        { subst; s in H; tmext_simplH H; clarify; specialize (E eq_refl); vauto. }
        rewrite ren_ext_it_neq in H; [|by ii; vauto].
        apply in_consD in IN; des; vauto.
        apply in_map_iff in IN; des; vauto.
      }
      clear IN; i; rewrite ren_ext_it_neq; [eauto|by ii; vauto].
      eapply foo'; eauto; vauto.
      { i; clarify; s; tmext_simpl; splits; eauto.
        exploit SIM; try by s; eauto.
        - instantiate (3 := map_pair (inl <*> inl) id subb).
          instantiate (3 := map_pair (inl <*> inl) id suba).
          clear IN; i; apply in_map_iff in IN; des; s in IN; clarify.
          dup IN0; dup IN0; rewrite DOM in IN0.
          eapply list_assoc_map in IN0; eapply list_assoc_map in IN2; des.
          eapply RS in IN1; eauto; esplits.
          + eapply list_assoc_map_pair; eauto; i; ss; clarify.
          + eapply list_assoc_map_pair; eauto; i; ss; clarify.
          + econs; eapply vclgen_thunk, vclos_thunk_mon; eauto.
            i; eapply LE0, gk_mon; eauto.
        - i; des; s in x2; clarify.
      }
      { i; clarify.
        destruct (LEM (x1 = inr bfb)); vauto.
        rewrite ren_ext_it_neq in H; [|by ii; vauto].
        exfalso.
        destruct x1; vauto.
        apply in_consD in IN; des; vauto.
        apply in_map_iff in IN; des; s in IN; clarify.
      }
      clear IN; i; rewrite ren_ext_it_neq; [eauto|by ii; vauto].
      rewrite ren_itvar_id.
      i; apply in_map_iff in IN; des; s in IN; clarify.
      dup IN0; dup IN0; rewrite DOM in IN0.
      eapply list_assoc_map in IN0; eapply list_assoc_map in IN2; des.
      eapply RS in IN1; eauto; esplits.
      * eapply list_assoc_map_pair; eauto; i; ss; clarify.
      * eapply list_assoc_map_pair; eauto; i; ss; clarify.
      * econs; eapply vclgen_thunk, vclos_thunk_mon; eauto.
        i; eapply LE0, gk_mon; eauto.
    + i; eapply foo'; vauto.
      { i; clarify; s; tmext_simpl; eauto. }
      { i; clarify.
        destruct (LEM (x = bkb)); eauto.
        exfalso.
        do 3 (apply in_consD in IN; des; vauto).
        apply in_map_iff in IN; des; vauto.
      }
      { clear IN; i; do 2 (apply in_consD in IN; des; vauto).
        exfalso; apply in_map_iff in IN; des; vauto.
      }
  - 
    econs 2; s; eauto.
    dup SIM; rrdes SIM0.
    rewrite map_comp -purge_map_inl -map_comp in SIM.

    eapply mm_compat_gen in SIM; eauto
    ; try (by ii; apply in_map_iff in IN; des; vauto); [].
    eapply gk_lsim; eauto.
    econs; [eapply mm_good_gen; eauto|].
    { ii; inv FH; eapply GEa; eauto. }
    { ii; inv FH; eapply GEb; eauto. }
    i; invc STK; [by invc RF VR APPa CA|].
    invc RK RF VR; invc INSTa INSTb CA CB.
    eapply pesim_gknow; i.
    eapply SIM0; eauto.
    + i; eapply foo'; vauto; []; clear IN; i.
      rewrite ren_ext_it_neq_c ren_itvar_id.
      apply in_map_iff in IN; des; s in IN; clarify.
      dup IN0; dup IN0; rewrite DOM in IN0.
      eapply list_assoc_map in IN0; eapply list_assoc_map in IN2; des.
      eapply RS in IN1; eauto; esplits.
      * eapply list_assoc_map_pair; eauto; i; ss; clarify.
      * eapply list_assoc_map_pair; eauto; i; ss; clarify.
      * econs; eapply vclgen_thunk, vclos_thunk_mon; eauto.
        i; eapply LE0, gk_mon; eauto.
    + i; eapply foo'; vauto.
      { i; clarify; s; tmext_simpl; eauto. }
      { i; clarify.
        destruct (LEM (x = bkb)); eauto.
        exfalso.
        apply in_consD in IN; des; vauto.
        apply in_map_iff in IN; des; vauto.
      }
      { clear IN; i.
        exfalso; apply in_map_iff in IN; des; vauto.
      }
Grab Existential Variables.
  exact 0. exact 0.
Qed.
            




Lemma mm_msim_consB: forall lw FI ea eb l FO ma mb sil
    (VSIM: mm_tlsimB lw FI ea eb)
    (MSIM: msim mm_gw (FI++[l]) FO ma mb)
    (ST: stable lw)
    (INIl: forall g (GK: gknow (wlift lw) g),
             lw.(confrel) g (tt, tt, refst_emp, sil) cemp cemp),
  msim mm_gw FI (l::FO) ((l,ea)::ma) ((l,eb)::mb).
Proof.
  i; rrdes MSIM.
  exists (wprod lw lw0); i.
  set (va := IVclo (map_pair (inl <*> inl) id suba) ea).
  set (vb := IVclo (map_pair (inl <*> inl) id subb) eb).
  hexploit MSIM; rr; eauto.
  { instantiate (1 := suba ++ [(l, va)]).
    rewrite map_app SUBa; auto.
  }
  { instantiate (1 := subb ++ [(l, vb)]).
    rewrite map_app SUBb; auto.
  }
  i; des.
  exists (fst si, (sil, snd si)).
  rrdes LMa LMb; subst.
  assert (fst si = (tt, tt, refst_emp)).
  { destruct si as [[[[][]]]]; s; repeat f_equal.
    hexploit INIg; eauto using gkmin_bot_gknow; []; i.
    rrdes X WHX.
    destruct hax, hay, hbx, hby; s in HPA; s in HPB; clarify.
    symmetry in H1; apply pmemp_plus in H1.
    symmetry in H3; apply pmemp_plus in H3.
    des; subst; rrdes WHY; clarify.
    apply refst_extensionality; split; i; vauto.
    apply RELV in H; des; apply pmread_empD in RDa; vauto.
  }    
  destruct si; clean; subst.

  assert (SIM': @tlsim _ _ mm_gw lw (map fst suba) suba subb
                       (tt, tt, refst_emp, sil) va vb).
  { eapply mm_tlsimB_tlsim; eauto. }

  splits; eauto.
  - eapply stable_prod; eauto.
  - i; hexploit gknow_prod; first 2 last; eauto; []; i; des.
    eapply confrel_mon; [by eapply INIg; eauto|].
    i; rrdes PR; inv PR; eapply gk_mon; s; eauto.
  - i; hexploit gknow_prod; first 2 last; eauto; []; i; des.
    move INIl at bottom; move INIl0 at bottom.
    rr; esplits; clean; eauto; noemp; eauto.
  - i; apply in_consD in IN; des; subst.
    + esplits; try (rr; s; des_if); vauto.
      ii; rrdes LE LE0; destruct s as [[[[]]]]; destruct s2; clean.
      hexploit gknow_prod; first 2 last; eauto; []; i; des.
      exploit SIM'.
      * inst_pairs; split; [exact LE|exact LE0].
      * eauto.
      * rrdes RLS RLS; rr; eauto.
      * i; exploit RS; eauto; clean; i.
        eapply vclos_thunk_lift, vclos_thunk_unlift, vclos_thunk_mon; eauto.
      * i; eapply vclos_thunk_lift, vclos_thunk_unlift, vclos_thunk_mon; eauto.
        i; invc PR; eapply GK; repeat (eauto; split; clean).
    + hexploit SIM; eauto; []; i; des.
      esplits.
      * rr; s; des_if; [by exfalso; vauto|].
        rr in Ea; eauto.
      * rr; s; des_if; [by exfalso; vauto|].
        rr in Eb; eauto.
      * { ii; rrdes LE LE0; destruct s as [[[]]]; destruct s3; clean.
          i; hexploit gknow_prod; first 2 last; eauto; []; i; des.
          exploit SIM0.
          - inst_pairs; split; [exact LE|exact LE1].
          - eauto.
          - rrdes RLS RLS; rr; eauto.
          - i; eapply (lookup2 _ _ _ _ _ _ _ SUBb); eauto; i.
            + exploit RS; eauto.
              * apply in_map_fst; eauto using list_assoc_someD.
              * clean; i.
                eapply vclos_thunk_lift; eapply vclos_thunk_unlift in x1.
                eapply vclos_thunk_mon; eauto.
            + sng; exploit SIM'.
              * inst_pairs; split; [exact LE|exact LE0].
              * eauto.
              * rrdes RLS RLS; rr; esplits; rr; clean; esplits; eauto.
              * i; exploit RS; eauto; []; i.
                eapply vclos_thunk_lift, vclos_thunk_unlift.
                eapply vclos_thunk_mon; eauto.
              * i; eapply vclos_thunk_lift; eapply vclos_thunk_unlift in x0.
                eapply vclos_thunk_mon; eauto.
                i; inv PR; econs.
                { eapply GK; repeat (eauto; split; clean). }
                eauto.
          - i; eapply vclos_thunk_lift, vclos_thunk_unlift.
            eapply vclos_thunk_mon; eauto.
            i; invc PR; eapply GK; repeat (eauto; split; clean).
        }
Grab Existential Variables.
  exact tt. exact tt.
Qed.




















