(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List.
Require Import vgtac basis myclassical.
Require Import lang_common lang_spec lang_src_lsi lang_tgt.
Require Import model_common model model_properties.
Require Import gw_common gw ssr util ts.
Require Import src2tgt src2tgt_correct_aux.

Unset Implicit Arguments. 
Global Set Bullet Behavior "Strict Subproofs".



Lemma tr_exp_correct_nat: forall F G m,
  ts_psim F G Tnat (fun n => map enc (tr_exp (Enat m) n (map fst F) (map fst G))) (Enat m).
Proof.
  i; econs; r; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.
  ts_prepare.

  
  hexploit (tgt_stack CPa 1); i; des; s in FST; subst freest.
  des; repeat (destruct vs; [by vauto|]); destruct vs; [|by vauto].
  rewrite <-!pmplus_assoc in *; clear NST.

  eapply pesim_dance
  ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
  ; clear RLa0 SV HV FIN.
  - dance_init.
    { eapply Sok; [by vauto|eapply Sopr].
      - readpc' mpc0 Eladdr.
        unfold push; auto.
      - eauto.
      - eauto.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Sld].
      - readpc' mpc0 Eladdr.
      - eauto.
    }
    { depmcplus0. }

    dance_init.
    { eapply Sok; [by vauto|eapply Ssto].
      - readpc' mpc0 Eladdr.
      - unfold stack_top; rewrite -pmcells_sng.
        prep_store; eauto.
    }
    { s; tmext_simpl; depmcplus0.
      match goal with |- context[?a + 1 - 1] =>
        replace (a + 1 - 1) with a by omega
      end.
      rewrite pmcells_sng.
      eauto using pmvalid_plus_cells.
    }

    eapply dance_O.
    instantiate (1 := fun _ _ => depmach0); s; eauto.

  - i; rrdes XTRb RLb0; clarify; s in Ec; clarify.
    esplits; first 2 last; last first.
    + right; splits; eauto using wfo_le_n_add.
    + i; frames_nodep RLa cfas.
      destruct ea'; clear FIN; destruct_tgt_real RLa.
      clear RLa0.
      setcrf mrf'.
      set (s4' := ({mrf r_sp |-> m} ## fst s4, snd s4)).
      set (s' := (mrf', P0, env, s3, s4')).
      assert (EXT': ext (wlift mylw) s' (mrf, P0, env, s3, s4)).
      { unfold s'; repeat (split; auto); unfold mrf'; s; tmext_simpl; auto.
        - split_pmvalid HV; auto.
        - rrdes IN; eauto.
      }
      esplits; eauto.

      hexploit KSIM; clear KSIM.
      * eauto.
      * eauto.
      * exact EXT'.
      * s; unfold mrf'; tmext_simpl; auto.
      * s; unfold mrf'; tmext_simpl; auto.
      * s; unfold mrf'; tmext_simpl; auto.
      * s; unfold mrf'; tmext_simpl.
        match goal with |- context[?a + 1 - 1] =>
          replace (a + 1 - 1) with a by omega
        end.
        auto.
      * s; econs; econs; rr; eauto.
      * intro X; simpl tr_exp in X.
        exploit X; clear X.
        { unfold s'; rr; esplits; eauto; [|by rr; esplits; eauto].
          rr; esplits; eauto.
          - rr; splits; [|by rr; esplits; eauto]; s.
            eapply (tgt_confpred_push mrf'); eauto.
            unfold mrf'; s; tmext_simpl; auto.
          - dup EXT'; rrdes EXT'; simpl_proj.
            eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
            + s; i; eapply gk_mon; eauto; split; eauto.
            + s; i; eapply gk_mon; eauto; split; eauto.
        }
        simpl cplus; noemp; intro ES.
        eapply pesim_rewrite; eauto.
        match goal with |- context[?a + 1 - 1] =>
          replace (a + 1 - 1) with a by omega
        end.
        rewrite pmcells_sng pmplus_swap; f_equal.
        rewrite -pmplus_assoc; auto.
    + rewrite -plus_n_O; s; noemp; rr; esplits; eauto.
Qed.


Lemma tr_exp_correct_unit: forall F G,
  ts_psim F G Tunit (fun n => map enc (tr_exp Eunit n (map fst F) (map fst G))) Eunit.
Proof.
  i; econs; r; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.
  ts_prepare.

  
  hexploit (tgt_stack CPa 1); i; des; s in FST; subst freest.
  des; repeat (destruct vs; [by vauto|]); destruct vs; [|by vauto].
  rewrite <-!pmplus_assoc in *; clear NST.

  eapply pesim_dance
  ; first (by instantiate (1 := mkmach _ _ _ _); rr; esplits; s; eauto)
  ; clear RLa0 SV HV FIN.
  - dance_init.
    { eapply Sok; [s; i; omega|eapply Sopr].
      - readpc' mpc0 Eladdr.
        unfold push; auto.
      - eauto.
      - eauto.
      - eauto.
    }
    { depmcplus0. }

    eapply dance_O.
    instantiate (1 := depmach0); s; eauto.

  - i; rrdes XTRb RLb0; clarify; s in Ec; clarify.
    esplits; first 2 last; last first.
    + right; splits; eauto using wfo_le_n_add.
    + i; frames_nodep RLa cfas.
      destruct ea'; clear FIN; destruct_tgt_real RLa.
      clear RLa0.
      setcrf mrf'.
      set (s4' := ({mrf r_sp |-> n0} ## fst s4, snd s4)).
      set (s' := (mrf', P0, env, s3, s4')).
      assert (EXT': ext (wlift mylw) s' (mrf, P0, env, s3, s4)).
      { unfold s'; repeat (split; auto); unfold mrf'; s; tmext_simpl; auto.
        - split_pmvalid HV; auto.
        - rrdes IN; eauto.
      }
      esplits; eauto.

      hexploit KSIM; clear KSIM.
      * eauto.
      * eauto.
      * exact EXT'.
      * s; unfold mrf'; tmext_simpl; auto.
      * s; unfold mrf'; tmext_simpl; auto.
      * s; unfold mrf'; tmext_simpl; auto.
      * s; unfold mrf'; tmext_simpl.
        match goal with |- context[?a + 1 - 1] =>
          replace (a + 1 - 1) with a by omega
        end.
        auto.
      * s; econs; econs; rr; eauto.
      * intro X; simpl tr_exp in X.
        exploit X; clear X.
        { unfold s'; rr; esplits; eauto; [|by rr; esplits; eauto].
          rr; esplits; eauto.
          - rr; splits; [|by rr; esplits; eauto]; s.
            eapply (tgt_confpred_push mrf'); eauto.
            unfold mrf'; s; tmext_simpl; auto.
          - dup EXT'; rrdes EXT'; simpl_proj.
            eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
            + s; i; eapply gk_mon; eauto; split; eauto.
            + s; i; eapply gk_mon; eauto; split; eauto.
        }
        simpl cplus; noemp; intro ES.
        eapply pesim_rewrite; eauto.
        rewrite pmcells_sng pmplus_swap; f_equal.
        rewrite -pmplus_assoc; auto.
    + s; noemp; rr; esplits; eauto.
Qed.


Lemma tr_exp_correct_var: forall F G ty x
    (IN: lookup x G = Some ty),
  ts_psim F G ty (fun n => map enc (tr_exp (Evar x) n (map fst F) (map fst G))) (Evar x).
Proof.
  i; econs; r; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.
  ts_prepare.

  
  hexploit (tgt_stack CPa 1); i; des; s in FST; subst freest.
  des; repeat (destruct vs; [by vauto|]); destruct vs; [|by vauto].
  rewrite <-!pmplus_assoc in *; clear NST.

  rewrite -pmcells_sng.
  apply RDV in IN; des.
  eapply pesim_getVariable; eauto.
  - eapply reprlist_mon; eauto.
    i; eapply pmread_plus; eauto; pmplus_eq'.
  - i; eapply pmread_plus; eauto.
    rhs rewrite -(pmplus_comm (snd s4)) -!(pmplus_swap (snd s4)).
    (rhs rewrite Eladdr -pmplus_assoc DORD); eauto.
  - setcrf mrf'.
    set (s' := (mrf', P0, s, s3, ({mrf r_sp |-> va} ## fst s4, snd s4))).
    assert (EXT': ext (wlift mylw) s' (mrf, P0, s, s3, s4)).
    { unfold s'; repeat (split; auto); unfold mrf'; s; tmext_simpl; auto.
      - split_pmvalid HV; auto.
      - rrdes IN1; eauto.
    }
    hexploit (KSIM g); clear KSIM.
    + eauto.
    + eauto.
    + exact EXT'.
    + s; unfold mrf'; tmext_simpl; auto.
    + s; unfold mrf'; tmext_simpl; eauto with arith.
    + s; unfold mrf'; tmext_simpl; auto.
    + s; unfold mrf'; tmext_simpl.
      match goal with |- context[1 + ?a - 1] =>
        replace (1 + a - 1) with a by omega
      end.
      auto.
    + eapply vclos_mon; eauto.
      i; eapply gk_mon; eauto.
    + intro X; simpl tr_exp in X.
      eapply pesim_anti with (s' := s'); eauto.
      exploit X; clear X.
      { unfold s'; rr; esplits; eauto; [|by rr; esplits; eauto].
        rr; esplits; eauto.
        - rr; splits; [|by rr; esplits; eauto]; s.
          eapply (tgt_confpred_push mrf'); eauto.
          unfold mrf'; s; tmext_simpl; omega.
        - dup EXT'; rrdes EXT'; simpl_proj.
          eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
          + s; i; eapply gk_mon; eauto; split; eauto.
          + s; i; eapply gk_mon; eauto; split; eauto.
      }
      simpl cplus; noemp; intro ES.
      eapply pesim_rewrite; eauto.
      s; rewrite pmplus_swap map_length DORD; f_equal.
      rewrite -pmplus_assoc; auto.
  - rewrite -DORD; apply wfo_le_n_add.
Qed.


Lemma tr_exp_correct_lbl: forall F G ty x
    (IN: lookup x F = Some ty),
  ts_psim F G ty (fun n => map enc (tr_exp (Elbl x) n (map fst F) (map fst G))) (Elbl x).
Proof.
  i; econs; r; i; des; s in Eladdr.
  destruct si as [[[[]]]], s as [[[[]]]]; clean.
  ts_prepare.

  
  hexploit (tgt_stack CPa 1); i; des; s in FST; subst freest.
  des; repeat (destruct vs; [by vauto|]); destruct vs; [|by vauto].
  rewrite <-!pmplus_assoc in *; clear NST.

  rewrite -pmcells_sng.
  apply RSV in IN; des.

  eapply pesim_stepb.
  i; rrdes XTR RLb; clarify.
  rrdes RL; s in Ec; noemp; clarify; esplits; [|right; esplits].
  { econs 1; inst_pairs; eapply Rok; eauto. }
  { inst_pairs; s; noemp; rr; esplits; eauto. }

  eapply pesim_getLabel; eauto.
  - i; eapply pmread_plus; eauto.
    (rhs rewrite Eladdr -!(pmplus_swap {mpc - n |=> _})); eauto.
  - i; eapply pmread_plus; eauto.
    rhs rewrite -(pmplus_comm (snd s4)) -!(pmplus_swap (snd s4)).
    (rhs rewrite Eladdr -pmplus_assoc SORD); eauto.
  - setcrf mrf'.
    set (s' := (mrf', P0, env0, s3, ({mrf r_sp |-> va} ## fst s4, snd s4))).
    assert (EXT': ext (wlift mylw) s' (mrf, P0, env0, s3, s4)).
    { unfold s'; repeat (split; auto); unfold mrf'; s; tmext_simpl; auto.
      - split_pmvalid HV; auto.
      - rrdes IN2; eauto.
    }
    hexploit (KSIM g); clear KSIM.
    + eauto.
    + eauto.
    + exact EXT'.
    + s; unfold mrf'; tmext_simpl; auto.
    + s; unfold mrf'; tmext_simpl; eauto with arith.
    + s; unfold mrf'; tmext_simpl; auto.
    + s; unfold mrf'; tmext_simpl.
      match goal with |- context[1 + ?a - 1] =>
        replace (1 + a - 1) with a by omega
      end.
      auto.
    + eapply vclos_mon; eauto.
      i; eapply gk_mon; eauto.
    + intro X; simpl tr_exp in X.
      eapply pesim_anti with (s' := s'); eauto.
      exploit X; clear X.
      { unfold s'; rr; esplits; eauto; [|by rr; esplits; eauto].
        rr; esplits; eauto.
        - rr; splits; [|by rr; esplits; eauto]; s.
          eapply (tgt_confpred_push mrf'); eauto.
          unfold mrf'; s; tmext_simpl; omega.
        - dup EXT'; rrdes EXT'; simpl_proj.
          eapply ts_refconfrel_mon; [eapply ts_refconfrel'_smon|]; eauto.
          + s; i; eapply gk_mon; eauto; split; eauto.
          + s; i; eapply gk_mon; eauto; split; eauto.
      }
      simpl cplus; noemp; intro ES.
      eapply pesim_rewrite; eauto.
      s; rewrite pmplus_swap; f_equal.
      rewrite -pmplus_assoc; auto.
  - rewrite -SORD; apply wfo_le_n_add.
Grab Existential Variables.
  exact (wfo_nat 42).
Qed.
