(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List Max Relations.
Require Import vgtac basis myclassical ssr.
Require Import lang_common lang_spec lang_mid lang_src_lsi.
Require Import model_common model_untyped model model_properties_untyped model_properties.
Require Import gw_common gw_untyped gw.

Set Implicit Arguments. 
Global Set Bullet Behavior "Strict Subproofs".


Notation cid := constructive_indefinite_description.



Module Export A := model_untyped.
Module Export B := model.
Module Export GA := gw_untyped.
Module Export GB := gw.
Module Export PA := model_properties_untyped.
Module Export PB := model_properties.

 
Notation "p <Ur= q" :=
  (forall  va vb (PR: p  va vb : Prop), 
          q  va vb : Prop)
  (at level 50, no associativity).
Notation "p <Ug= q" :=
  (forall s, (p s) <Ur= (q s))
  (at level 50, no associativity).
Notation "p \Ur/ q" :=
  (fun  va vb => 
     p  va vb \/ q  va vb)
  (at level 50, no associativity).
Notation "p \Ug/ q" :=
  (fun s => (p s) \Ur/ (q s))
  (at level 50, no associativity).

Definition refst_comp (sra: location -> location -> Prop)
                      (srb: typ -> location -> location -> Prop)
                      ty lt ls :=
  exists lm, sra lt lm /\ srb ty lm ls.

Program Definition refst_compP (sra: GA.refst) (srb: GB.refst) : refst :=
  Build_refst (refst_comp sra srb) _ _.
Next Obligation.
  i; unfold refst_comp in *; des; vauto.
  hexploit (gw_untyped.refst_bij IN IN'); eauto.
  hexploit (refst_bij IN0 IN'0); eauto.
  i; des; split; eauto.
  i; subst; specialize (H2 eq_refl); des; eauto.
Qed.
Next Obligation.
  hexploit (GA.refst_fin sra); []; i; des.
  hexploit (GB.refst_fin srb); []; i; des.
  exists (map (fun p => let '(lt, _, (ty, _, ls)) := p in (ty, lt, ls))
              (combine_full l l0)).
  ii; rrdes IN.
  apply FIN in IN; apply FIN0 in IN0.
  hexploit (combine_full_in _ _ _ _ IN IN0); []; i.
  apply in_map_iff; econs; inst_pairs; eauto.
Qed.

Lemma refst_compP_ext: forall
    sra sra' (EXTa: ext GA.ref_sts sra' sra)
    srb srb' (EXTb: ext GB.ref_sts srb' srb),
  ext ref_sts (refst_compP sra' srb') (refst_compP sra srb).
Proof.
  ii; rrdes PR; ss; r; eauto.
Qed.

Lemma refst_compP_extpub: forall
    sra sra' (EXTa: extpub GA.ref_sts sra' sra)
    srb srb' (EXTb: extpub GB.ref_sts srb' srb),
  extpub ref_sts (refst_compP sra' srb') (refst_compP sra srb).
Proof.
  ii; rrdes PR; ss; r; eauto.
Qed.


Program Definition refst_minusa (sra: GA.refst) (sr: refst) : GA.refst :=
  {| GA.refst_st lt lm := sra lt lm /\ forall ty ls, ~ sr ty lt ls |}.
Next Obligation.
  hexploit (GA.refst_bij H1 H); eauto.
Qed.
Next Obligation.
  hexploit (GA.refst_fin sra); []; i; des.
  exists l; ii; des; eauto.
Qed.

Program Definition refst_minusb (srb: GB.refst) (sr: refst) : GB.refst :=
  {| GB.refst_st ty lm ls := srb ty lm ls /\ forall lt, ~ sr ty lt ls |}.
Next Obligation.
  hexploit (refst_bij H1 H); eauto.
Qed.
Next Obligation.
  hexploit (GB.refst_fin srb); []; i; des.
  exists l; ii; des; eauto.
Qed.



Lemma freshloc: exists freshloc: list location -> typ -> location -> location -> location,
  <<freshloc_inj: forall ctx ty lt ls ty' lt' ls',
                    freshloc ctx ty lt ls = freshloc ctx ty' lt' ls' ->
                    ty = ty' /\ lt = lt' /\ ls = ls'>> /\
  <<freshloc_fresh: forall ctx ty lt ls,
                      ~ In (freshloc ctx ty lt ls) ctx>>.
Proof.
  assert (X := cntinf_prod typ_cntinf (cntinf_prod nat_cntinf nat_cntinf)).
  eapply ci_impl_ci2 in X; red in X; des.
  exists (fun ctx ty l1 l2 => (fold_right max 0 ctx) + 1 + f (ty,(l1,l2))).
  split; unnw; intros.
  - apply plus_reg_l in H; eapply X in H; inv H.
  - revert FH.
    assert (MAX : forall a, ~ In (max a (fold_right max 0 ctx) + 1 + f (ty, (lt, ls))) ctx).
    { induction ctx; auto; intros; simpls; destruct FH.
      - apply (lt_irrefl a); red; rewrite ->H at 2.
        eapply (le_trans _ (max a (fold_right max 0 ctx) + 1 + f (ty, (lt, ls)))).
        + rewrite <- plus_assoc, plus_comm, <- plus_assoc; apply le_n_S; rewrite plus_comm.
          apply le_plus_trans, le_max_l.
        + do 2 apply plus_le_compat_r; apply le_max_r.
      - eapply IHctx; rewrite max_assoc in H; eauto.
    }
    intros; eapply MAX with 0; auto.
Qed.
